classdef GtThreshold < handle
    properties
        conf;

        % Reconstruction parameters from parameters.mat
        param;
    end

    properties (Constant)
        backspace = '\b';
    end

    methods
        function obj = GtThreshold(varargin)
            obj.conf = [];
            obj.conf.verbose = true;

            obj.param = load('parameters.mat');
            obj.param = obj.param.parameters;

            par_thr = obj.getThrParams();
            [par_thr, rej_params] = parse_pv_pairs(par_thr, varargin);
            if (~isfield(par_thr, 'do_morph_recon'))
                par_thr.do_morph_recon = false;
            end
            obj.param.rec.thresholding = par_thr;

            obj.conf = parse_pv_pairs(obj.conf, rej_params);
        end

        function [defaults, sample] = generateDefaults(obj, fileTable)
            samplefile  = fullfile(obj.param.acq.dir, '4_grains', 'sample.mat');
            if (exist('fileTable', 'var'))
                sample = GtSample.loadFromLockedFile(fileTable, samplefile);
            else
                sample = GtSample.loadFromFile(samplefile);
            end

            defaults = [];
            defaults.percentile = obj.param.rec.thresholding.percentile;

            defaults.phases = {};
            for phaseID = 1:length(sample.phases)
                defaults.phases{phaseID}.min_id = 1;
                defaults.phases{phaseID}.max_id = length(sample.phases{phaseID}.grains);
            end
        end

        function [varargout] = batchThreshold(obj, phaseID, firstID, lastID, percentile)
        % GTTHRESHOLD/BATCHTHRESHOLD
        % [varargout] = batchThreshold(obj, phaseID, firstID, lastID, percentile)

            global GT_DB
            gtDBConnect(GT_DB.host, GT_DB.user, GT_DB.password, GT_DB.name);
            filetable = [obj.param.acq.name '_filetable'];

            [defaults, sample] = obj.generateDefaults(filetable);

            if (exist('percentile', 'var'))
                obj.param.rec.thresholding.percentile = percentile;

                parameters = obj.param;
                save('parameters.mat', 'parameters', '-v7.3');
            end

            if ischar(phaseID)
                phaseID = str2double(phaseID);
            end

            % Here is for the lazy who wants to launch from CMD line OR for OAR.
            if (ischar(firstID))
                if (strcmpi(firstID, 'first'))
                    firstID = defaults.phases{phaseID}.min_id;
                else
                    firstID = str2double(firstID);
                end
            end
            if (ischar(lastID))
                if (strcmpi(lastID, 'last'))
                    lastID = defaults.phases{phaseID}.max_id;
                else
                    lastID = str2double(lastID);
                end
            end

            if ((firstID < defaults.phases{phaseID}.min_id) ...
                    || (firstID > lastID) ...
                    || (lastID > defaults.phases{phaseID}.max_id))
                error('SEGMENT:wrong_argument', ...
                      ['The range indexes are not correct: (%d, %d).\n' ...
                      'Largest bounds (%d, %d)'], ...
                      firstID, lastID, defaults.phases{phaseID}.min_id, ...
                      defaults.phases{phaseID}.max_id);
            end

            gauge = GtGauge([firstID lastID], ...
                sprintf('Phase %02d: ', phaseID), ...
                'verbose', obj.conf.verbose);
            for grainID = firstID : lastID
                gauge.incrementAndDisplay();

                try
                    grain = obj.loadGrain(phaseID, grainID);

                    % Actually threshold now!
                    if (obj.checkIsIterative())
                        % iterative search
                        grain = obj.thresholdAutoSingleGrainIter(grain);
                    else
                        % normal way
                        grain = obj.thresholdAutoSingleGrain(grain);
                    end

                    % Check size and update metadata
                    if isempty(grain.seg)
                        fprintf('\n\nWarning! Empty volume in grain %d.\n\n',grainID);
                        gauge.rePrint();
                        grain.seg = 1;
                        grain.segbb = round([(size(grain.vol) -1), (size(grain.vol) +1)] /2);
                    end
                    sample.phases{phaseID}.setThreshold(grainID, grain.threshold);
                    sample.phases{phaseID}.setBoundingBox(grainID, grain.segbb);

                    % Write to file
                    obj.saveGrain(phaseID, grainID, grain);
                catch mexc
                    gtPrintException(mexc, ...
                        'Segmenting grain %d in phase %d failed', grainID, phaseID);
                    gauge.rePrint();
                end

            end
            gauge.delete();

            disp('Now writing to sample.mat')
            samplefile = fullfile(obj.param.acq.dir, '4_grains', 'sample.mat');
            sample.mergeToFile(filetable, samplefile, 'sample');

            if (nargout >= 1)
                varargout{1} = grain.seg;
            end
        end

        function grain = singleGrainThresholdInFile(obj, phaseID, grainID, threshold, center)
        % GTTHRESHOLD/singleGrainThresholdInFile
        % grain = singleGrainThresholdInFile(obj, phaseID, grainID, threshold, center)

            fprintf('Segmenting Grain %d, from phase %d.. ', grainID, phaseID);
            grain = obj.loadGrain(phaseID, grainID);

            if (exist('center', 'var'))
                do_morpho = true;
            else
                center = [];
                do_morpho = false;
                warning('THRESHOLD:less_capability', ...
                        'Morphological reconstruction is disabled')
            end

            if (obj.checkIsIterative())
                % iterative search
                grain = obj.thresholdAutoSingleGrainIter(grain, threshold, center);
            else
                % normal way
                grain = obj.segmentAndDoMorph(grain, threshold, do_morpho, center);
            end

            if isempty(grain.seg)
                fprintf('\n\nWarning! Empty volume in grain %d.\n\n', grainID);
                grain.seg = 1;
                grain.segbb = round([(size(grain.vol) -1), (size(grain.vol) +1)] /2);
            end

            obj.saveGrain(phaseID, grainID, grain);
            fprintf('Done.\n');
        end

        function grain = singleGrainAutoThreshold(obj, phaseID, grainID)
        % GTTHRESHOLD/singleGrainAutoThreshold
        % grain = singleGrainAutoThreshold(obj, phaseID, grainID)

            samplefile  = fullfile(obj.param.acq.dir, '4_grains', 'sample.mat');
            sample = GtSample.loadFromFile(samplefile);

            fprintf('Segmenting Grain %d, from phase %d..', grainID, phaseID);
            grain = obj.loadGrain(phaseID, grainID);

            if (obj.checkIsIterative())
                % iterative search
                grain = obj.thresholdAutoSingleGrainIter(grain);
            else
                % normal way
                grain = obj.thresholdAutoSingleGrain(grain);
            end

            if isempty(grain.seg)
                fprintf('\n\nWarning! Empty volume in grain %d.\n\n',grainID);
                grain.seg = 1;
                grain.segbb = round([(size(grain.vol) -1), (size(grain.vol) +1)] /2);
            end
            sample.phases{phaseID}.setThreshold(grainID, grain.threshold);
            sample.phases{phaseID}.setBoundingBox(grainID, grain.segbb);

            obj.saveGrain(phaseID, grainID, grain);
            fprintf('\b\b: Done.\n');

            fprintf('Writing to sample.mat..')
            save(samplefile, 'sample', '-v7.3');
            fprintf('\b\b: Done.\n');
        end

        function segvol = volumeThreshold(obj, volume, threshold, center)
        % GTTHRESHOLD/volumeThreshold
        % segvol = volumeThreshold(obj, volume, threshold, center)

            if (isempty(volume))
                gtError('GtThreshold:wrong_argument', ...
                    'Volume to segment is empty!')
            end

            if (exist('center', 'var') && ~isempty(center) && all(center > 0))
                do_morpho = true;
            else
                center = [];
                do_morpho = false;
                warning('THRESHOLD:less_capability', ...
                        'Morphological reconstruction is disabled')
            end

            mess = 'Error while segmenting the volume';
            segvol = obj.segmentVol(volume, threshold, do_morpho, mess, center);
        end
    end

    methods (Access = protected)
        function grain = loadGrain(obj, phaseID, grainID)
        % GTTHRESHOLD/LOADGRAIN
        % grain = loadGrain(obj, phaseID, grainID)

            phase_dir = fullfile(obj.param.acq.dir, '4_grains', ...
                sprintf('phase_%02d', phaseID));
            filePath = fullfile(phase_dir, sprintf('grain_%04d.mat', grainID));
            grain = load(filePath);

            if (isfield(obj.param.rec, 'grains') ...
                    && ismember(upper(obj.param.rec.grains.algorithm), {'6DL1', '6DTV', '6DTVL1'}))
                file_details_path = fullfile(phase_dir, ...
                    sprintf('grain_details_%04d.mat', grainID));

                grain_details = load(file_details_path, 'ODF6D');

                grain.vol = grain_details.ODF6D.intensity;
                grain.proj.shift = grain_details.ODF6D.shift;
            end

            if (~isfield(grain, 'vol') || isempty(grain.vol))
                gtError('GtThreshold:wrong_argument', ...
                    'Volume to segment is empty!')
            end
        end

        function saveGrain(obj, phaseID, grainID, grain)
        % GTTHRESHOLD/SAVEGRAIN
        % saveGrain(obj, phaseID, grainID, grain)

            phase_dir = fullfile(obj.param.acq.dir, '4_grains', ...
                sprintf('phase_%02d', phaseID));
            filePath = fullfile(phase_dir, sprintf('grain_%04d.mat', grainID));

            if (verLessThan('matlab', '7.14'))
                seg        = grain.seg;         %#ok<NASGU>
                segbb      = grain.segbb;       %#ok<NASGU>
                threshold  = grain.threshold;   %#ok<NASGU>
                Area       = grain.Area;        %#ok<NASGU>

                save(filePath, '-append', 'seg', 'segbb', 'threshold', 'Area');
            else
                matobj = matfile(filePath, 'Writable', true);
                matobj.seg = grain.seg;
                matobj.segbb = grain.segbb;
                matobj.threshold = grain.threshold;
                matobj.Area = grain.Area;
            end
        end

        function is_iterative = checkIsIterative(obj)
            thr_params = obj.getThrParams();
            is_iterative = thr_params.num_iter > 0;
        end

        function thr_params = getThrParams(obj)
        % GTTHRESHOLD/getThrParams
        %   FUNCTION that deals with parameters versions

            if (isfield(obj.param.rec, 'thresholding'))
                thr_params = obj.param.rec.thresholding;
            else
                warning('GtThreshold:old_params', ...
                    'Old "rec" parameters. You should produce a new version.')
                thr_params = struct( ...
                    'percentile', obj.param.rec.percentile2, ...
                    'do_morph_recon', obj.param.rec.do_morph_recon, ...
                    'num_iter', 0, 'iter_factor', 1);
            end
        end

        function grain = thresholdAutoSingleGrain(obj, grain)
        % GTTHRESHOLD/THRESHOLDAUTOSINGLEGRAIN Segment 3D grain volume
        %
        % function grain = thresh.thresholdAutoSingleGrain(grain, rec)
        %
        % INPUTS:  grain.vol          graylevel reconstruction of grain volume
        %          rec                contents of parameters.rec structure:
        %             .percentile    adaptive threshold see below
        %             .do_morph_recon remove not connected areas by morphological
        %                             reconstruction (<logical> {true})
        % OUTPUTS:
        %          grain.thresh Threshold value
        %          grain.seg    Segmented grain volume
        %          grain.Area   Area of grain
        %          grain.segbb  BoundingBox (gt convention) for insertion of seg
        %                       volume into sample volume
        %
        % explanation - we expect a histogram with a peak corresponding to the
        % grain, and a peak a low values corresponding to the background of zeros.
        % We want a threshold value below the peak corresponding to the grain, but
        % above the background peak.

            vol = grain.vol;

            par_thr = obj.getThrParams();

            % calulate intensity in the central region of the grain
            centerx = round(size(vol, 1) * 0.45) : round(size(vol, 1) * 0.55);
            centery = round(size(vol, 2) * 0.45) : round(size(vol, 2) * 0.55);
            centerz = round(size(vol, 3) * 0.45) : round(size(vol, 3) * 0.55);

            background = gtImgMeanValue(vol(centerx, centery, centerz)) / 10;
            % Small performance boost
            indexes = vol > background;
            % define threshold as percentile/100 x standard dev below the mean,
            % excluding low (10 percentile background) values
            stdvol  = std( vol(indexes) );
            meanvol = mean( vol(indexes) );

            threshold_val = meanvol - par_thr.percentile/100 * stdvol;

            center = round([size(vol, 1), size(vol, 2), size(vol, 3)] / 2);

            grain = obj.segmentAndDoMorph( grain, threshold_val, ...
                                            par_thr.do_morph_recon, ...
                                            center);
        end

        function grain = thresholdAutoSingleGrainIter(obj, grain, threshold_val, center)
        % GTTHRESHOLD/THRESHOLDAUTOSINGLEGRAINITER Segment 3D grain volume
        % iteratively
        %
        % grain = thresh.thresholdAutoSingleGrainIter(grain, threshold_val, center)
        %
        % INPUTS:  grain.vol          graylevel reconstruction of grain volume
        %
        % OUTPUTS:
        %          grain.thresh Threshold value
        %          grain.seg    Segmented grain volume
        %          grain.Area   Area of grain
        %          grain.segbb  BoundingBox (gt convention) for insertion of seg
        %                       volume into sample volume
        %
        % explanation - we want to end up with a grain that has a volume
        % which is consistent with the difspot sizes.  Adjust the threshold
        % until we achieve this.
        % We need other fields:
        %   parameters.rec.thresholding.iter_factor
        %   parameters.rec.thresholding.num_iter

            vol = grain.vol;

            par_thr = obj.getThrParams();

            % slow and stupid for the moment

            % get bounding box sizes from pair table
            query = sprintf([ ...
                'SELECT avbbXsize, avbbYsize ' ...
                '   FROM %s ' ...
                '   WHERE grainid = %d' ], ...
                obj.param.acq.pair_tablename, grain.id);
            [bbx, bby] = mym(query);
            bbx = median(bbx);
            bby = median(bby);
            target = bbx * bbx * bby * par_thr.iter_factor;

            if (~exist('threshold_val', 'var') || isempty(threshold_val))
                % calulate intensity in the central region of the grain
                centerx = round(size(vol, 1) * 0.45) : round(size(vol, 1) * 0.55);
                centery = round(size(vol, 2) * 0.45) : round(size(vol, 2) * 0.55);
                centerz = round(size(vol, 3) * 0.45) : round(size(vol, 3) * 0.55);
                % use this as a first guess
                threshold_val = gtImgMeanValue(vol(centerx, centery, centerz)) ;
            end
            if (~exist('center', 'var') || isempty(center) || any(center <= 0))
                % centre point for morphological
                center = round([size(vol, 1), size(vol, 2), size(vol, 3)] / 2);
            end

            happy = false; %:-(
            count = 0;
            while (~happy && count < par_thr.num_iter)
                % segment, test size
                grain = obj.segmentAndDoMorph( grain, threshold_val, ...
                    par_thr.do_morph_recon, center);

                test = sum(grain.seg(:));
                delta = 100 * (test - target) / target;

                if (abs(delta) > 1)
                    delta_sign = sign(delta);
                    delta_mag = min(threshold_val * abs(delta) / 100, 0.2);

                    threshold_val = threshold_val + (delta_sign * delta_mag);
                else
                    happy = true; %:-)
                end
                count = count + 1;
            end
            fprintf('finished grain %d after %d iterations, error %0.2f\n', ...
                grain.id, count, delta)
        end

        function grain = segmentAndDoMorph(obj, grain, threshold, do_morpho, center)
        % GTTHRESHOLD/SEGMENTANDDOMORPH

            message = sprintf(['morphological reconstruction failed for ' ...
                               'grain %d - continue without...'], ...
                              grain.id);
            segvol = obj.segmentVol(grain.vol, threshold, do_morpho, message, ...
                                    center);

            regprop = regionprops(segvol, 'BoundingBox', 'Area');
            if (length(regprop) > 1)
                % let's take the biggest! (usually happens without morpho)
                ind     = find([regprop.Area] == max([regprop.Area]), 1);
                regprop = regprop(ind);
            end

            if isempty(regprop)
                grain.seg   = [];
                grain.Area  = [];
                grain.segbb = [];
            else
                segbb = regprop.BoundingBox([2 1 3 5 4 6]);

                % ceil because Matlab regionprops yields "0.5" (instead of 1)
                % for bb start values
                grain.seg       = gtCrop(segvol, ceil(segbb));
                grain.Area      = regprop.Area;
                segbb(1:3)      = grain.proj.shift + floor(segbb(1:3));
                grain.segbb     = segbb;
            end
            grain.threshold = threshold;
        end

        function segvol = segmentVol(~, vol, threshold, do_morpho, errorMsg, center)
        % GTTHRESHOLD/SEGMENTVOL

            segvol = vol > threshold;

            % morphological reconstruction removes not connected areas
            if (do_morpho && ~isempty(center))
                try
                    marker = false(size(vol));
                    marker(center(1), center(2), center(3)) = true;
                    marker = marker & segvol;
                    segvol = double(imreconstruct(marker, segvol));
                catch mexc
                    gtPrintException(mexc, errorMsg);
                end
            end
        end
    end
end
