function grain = gtCalculateGrain(grain, parameters, varargin)
% GTCALCULATEGRAIN  Calulates diffraction vectors and detector intersection points
%
%     grain = gtCalculateGrain(grain, [parameters], varargin)
%     -----------------------------------------------------------------
%
%     INPUT:
%       grain      = <struct>   grain of interest
%       parameters = <struct>   parameters file {parameters.mat}
%
%     OPTIONAL INPUT (p/v pairs):
%       showfigure = <logical>  show or not the figure {false}
%       color      = <logical>  show or not the prediction spot positions coloured as
%                               omega {false}
%       axis       = <logical>  plot axis, image reference and beam direction in the
%                               image {false}
%       overlay    = <double>   difspots image output from "gtPlaceDifspotinFull" (MxM)
%                               M is the size of the overlay image (typically 2048)
%       markersize = <double>   marker size {10}
%       clims      = <double>   scale of grays {[-300 500]}
%       usestrain  = <logical>  use strain info
%       variants   = <logical>  Use twin variants simulated orientation
%                               stored in 'g_twins', if existing {false}
%
%     OUTPUT:
%       grain      = <struct>   grain of interest. Added fields:
%            .allblobs(fsimID)  = <struct>   Simulated reflections data
%            .used_fam          = <double>   List of used unique HKL families
%            .used_famsp        = <double>   List of all the used HKL families
%
%     Simplified version of gtFedGenerateRandomGrain   15-03-2012  - WLudwig


if (nargin < 1)
    help(mfile)
    error('gtCalculateGrain:wrong_arguments', 'Not enough arguments')
end
if (~exist('parameters', 'var') || isempty(parameters))
    parameters = load('parameters.mat');
    parameters = parameters.parameters;
end

% set default values for optional arguments
app.color      = false;
app.showfigure = false;
app.axis       = false;
app.markersize = 25;
app.overlay    = []; % zeros(parameters.acq.ydet, parameters.acq.xdet);
app.variants   = false;
app.usestrain  = false;
if (isfield(parameters.fsim, 'clims'))
    app.clims = parameters.fsim.clims;
else
    app.clims = parameters.fsim.Grayscale;
end

[app, rej_pars] = parse_pv_pairs(app, varargin);
if (app.clims == 0)
    app.clims = [];
end

if ~isfield(grain, 'full') || isempty(grain.full)
    grain.full = app.overlay;
end

% If strain data is not available, disable strain
if ~isfield(grain, 'strain') || ~isfield(grain.strain, 'strainT') || ...
    isnan(grain.strain.strainT(1, 1))

    app.usestrain = false;
end

% Rodrigues vector (grain orientation)
R_vec = grain.R_vector;
gcsam = grain.center;

cryst  = parameters.cryst(grain.phaseid);
labgeo = parameters.labgeo;
samgeo = parameters.samgeo;

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Detector geometry

% Scaling for Lab -> Detector transformation (mm -> pixels)
detscaleu = 1 / labgeo.pixelsizeu;
detscalev = 1 / labgeo.pixelsizev;

% Detector projection matrix
Qdet = gtFedDetectorProjectionTensor(labgeo.detdiru', labgeo.detdirv', ...
                                     detscaleu, detscalev);
if (~isfield(labgeo, 'detnorm'))
    labgeo.detnorm = gtMathsCross(labgeo.detdiru, labgeo.detdirv);
end

uvorig = [labgeo.detrefu, labgeo.detrefv]';
omstep = 180 / parameters.acq.nproj;


%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Crystallography from parameters file
lambda = gtConvEnergyToWavelength(parameters.acq.energy);

%disp('Translating Miller indices into normalized cartesian coordinates for plane normals')
Bmat = gtCrystHKL2CartesianMatrix(cryst.latticepar);

% D-spacings in the reference crystal for the given reflections
dsp_ref = gtCrystDSpacing(double(cryst.hkl(:, cryst.thetatypesp)), Bmat);

% Plane normals in Cartesian CRYSTAL coordinates
pl_cry = gtCrystHKL2Cartesian(double(cryst.hklsp), Bmat)';

% Compute all orientation matrices g (Reminder: Vc = g . Vs)
g = gtMathsRod2OriMat(R_vec');

% Express plane normals in cartesian SAMPLE CS
pl_sam = gtVectorCryst2Lab(pl_cry, g)';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute diffraction angles and detector intersection points
rotcomp = gtMathsRotationMatrixComp(labgeo.rotdir', 'col');

% Impose strain on grain
if (app.usestrain)
    % Deformation tensor (use strain tensor for now)
    defT = grain.strain.strainT;

    % New deformed plane normal and relative elongation (relative d-spacing)
    [pl_samd, drel] = gtStrainPlaneNormals(pl_sam, defT); % unit vector
else
    pl_samd = pl_sam;
    num_plane_normals = size(pl_sam, 2);
    drel = ones(1, num_plane_normals);
end

% d-spacing after applying the strain tensor
dsp_d = drel .* dsp_ref;

% Bragg angles theta in the deformed state
sinth = 0.5 * lambda ./ dsp_d;

% New sin(theta) in deformed state is inversely proportional to elongation
theta = asind(sinth);

% The plane normals need to be brought in the Lab reference where the
% beam direction and rotation axis are defined.
% Use the Sample -> Lab orientation transformation assuming omega=0;
% (vector length preserved for free vectors)
pl_labd = gtGeoSam2Lab(pl_samd', eye(3), labgeo, samgeo, true, false)';

% Four omegas of the plane normal (1-3 and 2-4 are the two Friedel pairs):
% pls is plus or minus pl, and St is the rotation tensor per each omega
[om4, pllab4, pls4, rot_l2s_4, omind4] = gtFedPredictOmegaMultiple( ...
    pl_labd, sinth, labgeo.beamdir', labgeo.rotdir', rotcomp, []);

% Delete those where no reflection occurs
todel               = (isnan(om4(1, :)) | isnan(om4(3, :)));
om4(:, todel)       = [];
omind4(:, todel)    = [];
sinth(todel)        = [];
theta(todel)        = [];
pllab4(:, todel, :) = [];
pls4(:, todel, :)   = [];
rot_l2s_4(:, :, todel, :) = [];
pl_sam(:, todel)    = [];

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Propagating the same common values for all the 4 omegas coming from a
% % single plane normal.
plorig4 = pl_sam(:, :, [1 1 1 1]);

if (size(cryst.hkl, 1) == 3 || size(cryst.hkl, 1) == 4)
    hkl = cryst.hkl;
    hklsp = cryst.hklsp;
else
    hkl = cryst.hkl';
    hklsp = cryst.hklsp';
end
num_hkl_indxs = size(hkl, 1);

thetatypesp = cryst.thetatypesp(~todel);
thetatypesp = reshape(thetatypesp, [], 1);

hkl = hkl(:, thetatypesp)';
hklsp = hklsp(:, ~todel)';

thetatypesp4 = thetatypesp(:, [1 1 1 1]);
hkl4 = hkl(:, :, [1 1 1 1]);
hklsp4 = hklsp(:, :, [1 1 1 1]);
hklsp4(:, :, [3 4]) = -hklsp4(:, :, [3 4]);
sinth4 = sinth([1 1 1 1], :);
theta4 = theta([1 1 1 1], :);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Following the convention of the matching output, the omega value
% % smaller than 180deg (spot "a") is used in the pair data.

chind4 = om4(1, :) > om4(3, :);

pllab4(:, chind4, [1, 3]) = pllab4(:, chind4, [3, 1]);
pls4(:, chind4, [1, 3])   = pls4(:, chind4, [3, 1]);
om4([1, 3], chind4)       = om4([3, 1], chind4);
omind4([1, 3], chind4)    = omind4([3, 1], chind4);
rot_l2s_4(:, :, chind4, [1, 3]) = rot_l2s_4(:, :, chind4, [3, 1]);

chind4 = om4(2, :) > om4(4, :);

pllab4(:, chind4, [2, 4]) = pllab4(:, chind4, [4, 2]);
pls4(:, chind4, [2, 4])   = pls4(:, chind4, [4, 2]);
om4([2, 4], chind4)       = om4([4, 2], chind4);
omind4([2, 4], chind4)    = omind4([4, 2], chind4);
rot_l2s_4(:, :, chind4, [2, 4]) = rot_l2s_4(:, :, chind4, [4, 2]);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Four omegas of the plane normal (1-3 and 2-4 are the two Friedel pairs):
% % But the two Friedel pairs are going to be 1a-1b and 2a-2b.

pllab4(:, :, [2, 3]) = pllab4(:, :, [3, 2]);
pls4(:, :, [2, 3])   = pls4(:, :, [3, 2]);
om4([2, 3], :)       = om4([3, 2], :);
omind4([2, 3], :)    = omind4([3, 2], :);
rot_l2s_4(:, :, :, [2, 3]) = rot_l2s_4(:, :, :, [3, 2]);
hklsp4(:, :, [2, 3]) = hklsp4(:, :, [3, 2]);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Unfolding the 4-fold data structures

om = reshape(om4, [], 1);
omind = reshape(omind4, [], 1);
sinth = reshape(sinth4, [], 1);
theta = reshape(theta4, [], 1);

pllab = reshape(permute(pllab4, [1 3 2]), 3, []);
pls = reshape(permute(pls4, [1 3 2]), 3, []);
plorig = reshape(permute(plorig4, [1 3 2]), 3, []);

rot_l2s = reshape(permute(rot_l2s_4, [1 2 4 3]), 3, 3, []);

thetatypesp = reshape(thetatypesp4', [], 1);
hkl = reshape(permute(hkl4, [3 1 2]), [], num_hkl_indxs);
hklsp = reshape(permute(hklsp4, [3 1 2]), [], num_hkl_indxs);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Computing the remaining physical quantities

% Diffraction vectors % takes coloumn vectors
dveclab = gtFedPredictDiffVecMultiple(pllab, labgeo.beamdir');

% Normalize vectors
dvec_norms = sqrt(sum(dveclab .* dveclab, 1));
dveclab = (dveclab ./ dvec_norms([1 1 1], :))';

% Eta angles % takes row vector
eta = gtGeoEtaFromDiffVec(dveclab, labgeo);

% Diffraction vector in Sample reference
dvecsam = gtGeoLab2Sam(dveclab, om, labgeo, samgeo, true);

% Let's convert grain center Sam -> Lab
rot_s2l = permute(rot_l2s, [2 1 3]);
gcsam_v = gcsam(ones(1, size(rot_l2s, 3)), :);
gclab_v = gtGeoSam2Lab(gcsam_v, rot_s2l, labgeo, samgeo, false);

% u,v,w coordinates on the detector
uvw = gtFedPredictUVWMultiple([], dveclab', gclab_v', labgeo.detrefpos', ...
    labgeo.detnorm', Qdet, uvorig, om', omstep)';

% Lorentz factor for the given projections as of eq. 4.1 from page 36 in
% Henning's book
Lfactor = 1 ./ (sinth .* abs(sind(eta)));

% Precompute which reflections will actually fall into the detector
ondet = ...
    (uvw(:, 1) >= 1) & (labgeo.detsizeu >= uvw(:, 1)) & ...
    (uvw(:, 2) >= 1) & (labgeo.detsizev >= uvw(:, 2));

% Initialse output variables
grain.allblobs.uvw = uvw;
grain.allblobs.uv = uvw(:, 1:2);
grain.allblobs.plorig = plorig';
grain.allblobs.pl = pls';
grain.allblobs.pllab = pllab';
grain.allblobs.hkl = hkl;
grain.allblobs.hklsp = hklsp;
grain.allblobs.sintheta = sinth;
grain.allblobs.theta = theta;
grain.allblobs.eta = eta;
grain.allblobs.thetatype = thetatypesp;
grain.allblobs.lorentzfac = Lfactor;
grain.allblobs.omega = om;
grain.allblobs.omind = omind;
grain.allblobs.dvec = dveclab;
grain.allblobs.dvecsam = dvecsam;
grain.allblobs.srot = rot_l2s;
grain.allblobs.ondet = ondet;

grain.full = [];

% Visual Feedback
if (app.showfigure)
    sfDisplayFigure(parameters, app, grain);
end

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sub-functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function sfDisplayFigure(parameters, app, grain)

    detusize = parameters.acq.xdet;
    detvsize = parameters.acq.ydet;

    if (isempty(app.overlay))
        app.overlay = zeros(parameters.acq.ydet, parameters.acq.xdet);
    end

    hf = figure();
    ax = axes('Parent', hf);
    imshow(app.overlay, app.clims, 'Parent', ax);
    hold(ax, 'On');

    if (app.axis)
        % Detector geometry
        beamdir = parameters.labgeo.beamdir';
        rotdir  = parameters.labgeo.rotdir';
        detdiru = parameters.labgeo.detdiru';
        detdirv = parameters.labgeo.detdirv';
        Qdet    = gtFedDetectorProjectionTensor(detdiru, detdirv, 1, 1);
        uvorig  = [parameters.labgeo.detrefu, parameters.labgeo.detrefv]';

        % Image frame
        plot([0 detusize+1], [0 0], 'k')
        plot([0 detusize+1], [detvsize+1 detvsize+1], 'k')
        plot([0 0], [0 detvsize+1], 'k')
        plot([detusize+1 detusize+1], [0 detvsize+1], 'k')

        % Midlines
        plot([0 detusize+1], [detvsize/2+0.5 detvsize/2+0.5], '-.k')
        plot([detusize/2+0.5 detusize/2+0.5], [0 detvsize+1], '-.k')

        % Set figure propeties
        axis equal
%         set(gca, 'YDir', 'reverse') % ?????
        set(ax, 'Position', [0.05 0.05 0.95 0.9])
%         xlim([-detusize*0.1 detusize*1.1])
%         ylim([-detvsize*0.1 detvsize*1.1])
        xlabel('U direction')
        ylabel('V direction')

        % Beam direction in image
        beamuv = Qdet * beamdir;
        % Arrow in figure indicating beam direction
        quiver(uvorig(1), uvorig(2), beamuv(1), beamuv(2), 1000, '-k', 'Linewidth', 3)

        % Rotation axis direction in image
        rotuv = Qdet*rotdir;
        % Arrow in figure indicating rotation axis direction
        quiver(uvorig(1), uvorig(2), rotuv(1), rotuv(2), 1000, '-.k', 'Linewidth', 3)
    end

    cmap = jet(1001);
    omtoRGB = @(om) cmap(round((om)/(360)*1000+1), :);

    if (app.color)
        freezeColors(gca)
        colormap(cmap)
        colorbar
        caxis([0 360])
    else
        colormap(gray);
    end

    uvw = grain.allblobs.uvw;
    if (app.color)
        om = grain.allblobs.omega;
        h = scatter3(uvw(:, 1), uvw(:, 2), uvw(:, 3), app.markersize, 'fill', 'o', 'CData', omtoRGB(om));
    else
        h = scatter3(uvw(:, 1), uvw(:, 2), uvw(:, 3), app.markersize, 'fill', 'o');
    end
    set(h, 'MarkerEdgeColor', [0 0 0])
    view(0, 90)
    xlim([0 detusize])
    ylim([0 detvsize])

    hold(ax, 'Off');

    ax_title = sprintf('Fsim for grain %d in phase %d', grain.id, grain.phaseid);
    title(ax, ax_title, 'FontWeight', 'bold');
end
