function grain = gtCalculateGrain_p(grain, parameters, varargin)
% GTCALCULATEGRAIN_P  Calulates diffraction vectors and detector intersection points
%
%     grain = gtCalculateGrain_p(grain, parameters, varargin)
%     -----------------------------------------------------
%       Parallel version of gtCalculateGrain
%
%     INPUT:
%       grain      = <struct>/<cell <struct> > grain(s) of interest
%       parameters = <struct>   parameters file
%
%     OPTIONAL INPUT (as a list of pairs):
%       showfigure = <logical>  show or not the figure {false}
%       color      = <logical>  show or not the prediction spot positions coloured as
%                               omega {false}
%       axis       = <logical>  plot axis, image reference and beam direction in the
%                               image {false}
%       overlay    = <double>   difspots image output from "gtPlaceDifspotinFull" (MxM)
%                               M is the size of the overlay image (typically 2048)
%       markersize = <double>   marker size {10}
%       clims      = <double>   scale of grays {[-300 500]}
%       usestrain  = <logical>  use strain info
%
%     OUTPUT:
%       grain      = <struct>/<cell <struct> > grain(s) of interest
%                               Added fields:
%            .pllab     = <double>
%            .hklsp     = <double>
%            .dvec      = <double>
%            .dvecsam   = <double>
%            .allblobs  = <struct>
%                               Updated fields:
%            .thetatype = <double>
%            .hkl       = <double>
%            .pl        = <double>
%            .theta     = <double>
%            .eta       = <double>
%            .omega     = <double>
%
% Simplified version of gtFedGenerateRandomGrain   15-03-2012  - WLudwig

if nargin < 1
    disp('Usage:  grain =  gtCalculateGrain(grain, parameters, varargin)');
    return
end
if ~exist('parameters','var') || isempty(parameters)
    parameters = load('parameters.mat');
    parameters = parameters.parameters;
end

% set default values for optional arguments
app.color      = false;
app.showfigure = false;
app.axis       = false;
app.markersize = 10;
app.clims      = [-300 500];
app.overlay    = []; % zeros(parameters.acq.ydet, parameters.acq.xdet);
app.usestrain  = false;

app = parse_pv_pairs(app,varargin);

% If strain data is not available, disable strain
if ~isfield(grain,'strain') || ~isfield(grain.strain,'strainT') || ...
    isnan(grain.strain.strainT(1,1))

    app.usestrain = false;
end

if (iscell(grain))
    phase = unique(cellfun(@(x)x.phaseid, grain));
    if (numel(phase) > 1)
        error('gtCalculateGrain:wrong_argument', ...
            'Cannot handle multiple phases!');
    end

    R_vec = cellfun(@(x){x.R_vector}, grain);
    R_vec = cat(1, R_vec{:});

    gcsam = cellfun(@(x){x.center}, grain);
    gcsam = cat(3, gcsam{:});

    num_grains = numel(grain);
else
    % Rodrigues vector (grain orientation)
    R_vec = grain.R_vector;
    phase = grain.phaseid;
    gcsam = grain.center;

    num_grains = 1;
end


cryst  = parameters.cryst(phase);
labgeo = parameters.labgeo;
samgeo = parameters.samgeo;

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Detector geometry

% Scaling for Lab -> Detector transformation (mm -> pixels)
detscaleu = 1 / labgeo.pixelsizeu;
detscalev = 1 / labgeo.pixelsizev;

% Detector projection matrix
Qdet = gtFedDetectorProjectionTensor(labgeo.detdiru', labgeo.detdirv', ...
                                     detscaleu, detscalev);
if (~isfield(labgeo, 'detnorm'))
    labgeo.detnorm = gtMathsCross(labgeo.detdiru, labgeo.detdirv);
end

uvorig = [labgeo.detrefu, labgeo.detrefv]';
omstep = 180 / parameters.acq.nproj;

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Crystallography from parameters file
lambda = gtConvEnergyToWavelength(parameters.acq.energy);

%disp('Translating Miller indices into normalized cartesian coordinates for plane normals')
Bmat = gtCrystHKL2CartesianMatrix(cryst.latticepar);

% D-spacings in the reference crystal for the given reflections
dsp_ref = gtCrystDSpacing(double(cryst.hkl(:, cryst.thetatypesp)), Bmat);

% Plane normals in Cartesian CRYSTAL coordinates
pl_cry = gtCrystHKL2Cartesian(double(cryst.hklsp), Bmat)';

% Compute all orientation matrices g (Reminder: Vc = g . Vs)
g = gtMathsRod2OriMat(R_vec');

% Express plane normals in cartesian SAMPLE CS
pl_sam = gtVectorCryst2Lab(pl_cry, g);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute diffraction angles and detector intersection points
rotcomp = gtMathsRotationMatrixComp(labgeo.rotdir', 'col');

% Let's merge multple grains information for vectorization purposes
pl_samd = permute(pl_sam, [1 3 2]);
pl_samd = reshape(pl_samd, [], 3);

dsp_ref = dsp_ref(1, :, ones(1, num_grains));
dsp_ref = reshape(dsp_ref, 1, []);

num_plane_normals = size(pl_sam, 1);

% Impose strain on grain
if (app.usestrain)
    % Deformation tensor (use strain tensor for now)
    defT = grain.strain.strainT;

    % New deformed plane normal and relative elongation (relative d-spacing)
    [pl_samd, drel] = gtStrainPlaneNormals(pl_samd', defT); % unit vector
else
    pl_samd = pl_samd';
    drel = ones(1, num_plane_normals * num_grains);
end

% d-spacing after applying the strain tensor
dsp_d = drel .* dsp_ref;

% Bragg angles theta in the deformed state
sinth = 0.5 * lambda ./ dsp_d;

% New sin(theta) in deformed state is inversely proportional to elongation
theta = asind(sinth);

% The plane normals need to be brought in the Lab reference where the
% beam direction and rotation axis are defined.
% Use the Sample -> Lab orientation transformation assuming omega=0;
% (vector length preserved for free vectors)
pl_labd = gtGeoSam2Lab(pl_samd', eye(3), labgeo, samgeo, true, false)';

% Four omegas of the plane normal (1-3 and 2-4 are the two Friedel pairs):
% pls is plus or minus pl, and St is the rotation tensor per each omega
[om4, pllab4, pls4, rot4, omind4] = gtFedPredictOmegaMultiple(pl_labd, ...
    sinth, labgeo.beamdir', labgeo.rotdir', rotcomp, []);

% Delete those where no reflection occurs
todel4 = (isnan(om4(1, :)) | isnan(om4(3, :)));
% todel4 is 4 x (num_pn * num_gr) already!
todel4 = todel4([1 1 1 1], :);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Propagating the same common values for all the 4 omegas coming from a
% % single plane normal.
plorig4 = pl_samd(:, :, [1 1 1 1]);

if (size(cryst.hkl, 1) == 3 || size(cryst.hkl, 1) == 4)
    hkl = cryst.hkl;
    hklsp = cryst.hklsp;
else
    hkl = cryst.hkl';
    hklsp = cryst.hklsp';
end
num_hkl_indxs = size(hkl, 1);

thetatypesp = reshape(cryst.thetatypesp, [], 1);
thetatypesp = thetatypesp(:, ones(1, num_grains));
thetatypesp = reshape(thetatypesp, [], 1);
hkl = hkl(:, thetatypesp)';

hklsp = hklsp(:, :, ones(1, num_grains));
hklsp = reshape(hklsp, num_hkl_indxs, []);
hklsp = hklsp';

thetatypesp4 = thetatypesp(:, [1 1 1 1]);
hkl4 = hkl(:, :, [1 1 1 1]);
hklsp4 = hklsp(:, :, [1 1 1 1]);
hklsp4(:, :, [3 4]) = -hklsp4(:, :, [3 4]);
sinth4 = sinth([1 1 1 1], :);
theta4 = theta([1 1 1 1], :);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Following the convention of the matching output, the omega value
% % smaller than 180deg (spot "a") is used in the pair data.

chind4 = om4(1, :) > om4(3, :);

pllab4(:, chind4, [1, 3]) = pllab4(:, chind4, [3, 1]);
pls4(:, chind4, [1, 3])   = pls4(:, chind4, [3, 1]);
om4([1, 3], chind4)       = om4([3, 1], chind4);
omind4([1, 3], chind4)    = omind4([3, 1], chind4);
rot4(:, :, chind4, [1, 3]) = rot4(:, :, chind4, [3, 1]);

chind4 = om4(2, :) > om4(4, :);

pllab4(:, chind4, [2, 4]) = pllab4(:, chind4, [4, 2]);
pls4(:, chind4, [2, 4])   = pls4(:, chind4, [4, 2]);
om4([2, 4], chind4)       = om4([4, 2], chind4);
omind4([2, 4], chind4)    = omind4([4, 2], chind4);
rot4(:, :, chind4, [2, 4]) = rot4(:, :, chind4, [4, 2]);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Four omegas of the plane normal (1-3 and 2-4 are the two Friedel pairs):
% % But the two Friedel pairs are going to be 1a-1b and 2a-2b.

pllab4(:, :, [2, 3]) = pllab4(:, :, [3, 2]);
pls4(:, :, [2, 3])   = pls4(:, :, [3, 2]);
om4([2, 3], :)       = om4([3, 2], :);
omind4([2, 3], :)    = omind4([3, 2], :);
rot4(:, :, :, [2, 3]) = rot4(:, :, :, [3, 2]);
hklsp4(:, :, [2, 3]) = hklsp4(:, :, [3, 2]);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Unfolding the 4-fold data structures

todel = reshape(todel4, [], 1);

om = reshape(om4, [], 1);
omind = reshape(omind4, [], 1);
sinth = reshape(sinth4, [], 1);
theta = reshape(theta4, [], 1);

% pllab4, pls4, and plorig4 are: 3 x (num_pn * num_gr) x 4
pllab = reshape(permute(pllab4, [1 3 2]), 3, []);
pls = reshape(permute(pls4, [1 3 2]), 3, []);
plorig = reshape(permute(plorig4, [1 3 2]), 3, []);

rot_l2s = reshape(permute(rot4, [1 2 4 3]), 3, 3, []);

thetatypesp = reshape(thetatypesp4', [], 1);
hkl = reshape(permute(hkl4, [3 1 2]), [], num_hkl_indxs);
hklsp = reshape(permute(hklsp4, [3 1 2]), [], num_hkl_indxs);

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % Computing the remaining physical quantities

% Diffraction vectors % takes coloumn vectors
dveclab = gtFedPredictDiffVecMultiple(pllab, labgeo.beamdir');

% Normalize vectors
dvec_norms = sqrt(sum(dveclab .* dveclab, 1));
dveclab = (dveclab ./ dvec_norms([1 1 1], :))';

% Eta angles % takes row vector
eta = gtGeoEtaFromDiffVec(dveclab, labgeo);

% Diffraction vector in Sample reference
dvecsam = gtGeoLab2Sam(dveclab, om, labgeo, samgeo, true);

% Let's convert grain center Sam -> Lab
rot_s2l = permute(rot_l2s, [2 1 3]);
gcsam_v = permute(gcsam, [1 3 2]);
gcsam_v = gcsam_v(ones(1, num_plane_normals*4), :, :);
gcsam_v = reshape(gcsam_v, [], 3);
gclab_v = gtGeoSam2Lab(gcsam_v, rot_s2l, labgeo, samgeo, false);

% u,v,w coordinates on the detector
uvw = gtFedPredictUVWMultiple([], dveclab', gclab_v', labgeo.detrefpos', ...
    labgeo.detnorm', Qdet, uvorig, om', omstep);

% Lorentz factor for the given projections as of eq. 4.1 from page 36 in
% Henning's book
Lfactor = 1 ./ (sinth .* abs(sind(eta)));

% Precompute which reflections will actually fall into the detector
ondet = ...
    (uvw(1, :) >= 1) & (labgeo.detsizeu >= uvw(1, :)) & ...
    (uvw(2, :) >= 1) & (labgeo.detsizev >= uvw(2, :));

% We could add the ondet, if we wish
keep = ~todel;

if (num_grains > 1)
    keep = reshape(keep, [], num_grains);

    uvw = reshape(uvw, 3, [], num_grains);
    uvw = permute(uvw, [2 1 3]);

    plorig = reshape(plorig, 3, [], num_grains);
    pls = reshape(pls, 3, [], num_grains);
    pllab = reshape(pllab, 3, [], num_grains);
    plorig = permute(plorig, [2 1 3]);
    pls = permute(pls, [2 1 3]);
    pllab = permute(pllab, [2 1 3]);

    hkl = reshape(hkl', size(cryst.hkl, 1), [], num_grains);
    hklsp = reshape(hklsp', size(cryst.hklsp, 1), [], num_grains);
    hkl = permute(hkl, [2 1 3]);
    hklsp = permute(hklsp, [2 1 3]);

    sinth = reshape(sinth, [], num_grains);
    theta = reshape(theta, [], num_grains);
    thetatypesp = reshape(thetatypesp, [], num_grains);

    eta = reshape(eta, [], num_grains);
    Lfactor = reshape(Lfactor, [], num_grains);

    om = reshape(om, [], num_grains);
    omind = reshape(omind, [], num_grains);

    dveclab = reshape(dveclab', 3, [], num_grains);
    dvecsam = reshape(dvecsam', 3, [], num_grains);
    dveclab = permute(dveclab, [2 1 3]);
    dvecsam = permute(dvecsam, [2 1 3]);

    rot_l2s = reshape(rot_l2s, 3, 3, [], num_grains);
    ondet = reshape(ondet, [], num_grains);

    for ii_gr = 1:num_grains
        % Initialse output variables
        gr_keep = keep(:, ii_gr);

        grain{ii_gr}.allblobs.uvw = uvw(gr_keep, :, ii_gr);
        grain{ii_gr}.allblobs.uv = uvw(gr_keep, 1:2, ii_gr);

        grain{ii_gr}.allblobs.plorig = plorig(gr_keep, :, ii_gr);
        grain{ii_gr}.allblobs.pl = pls(gr_keep, :, ii_gr);
        grain{ii_gr}.allblobs.pllab = pllab(gr_keep, :, ii_gr);

        grain{ii_gr}.allblobs.hkl = hkl(gr_keep, :, ii_gr);
        grain{ii_gr}.allblobs.hklsp = hklsp(gr_keep, :, ii_gr);

        grain{ii_gr}.allblobs.sintheta = sinth(gr_keep, ii_gr);
        grain{ii_gr}.allblobs.theta = theta(gr_keep, ii_gr);
        grain{ii_gr}.allblobs.thetatype = thetatypesp(gr_keep, ii_gr);
        grain{ii_gr}.allblobs.eta = eta(gr_keep, ii_gr);
        grain{ii_gr}.allblobs.lorentzfac = Lfactor(gr_keep, ii_gr);

        grain{ii_gr}.allblobs.omega = om(gr_keep, ii_gr);
        grain{ii_gr}.allblobs.omind = omind(gr_keep, ii_gr);

        grain{ii_gr}.allblobs.dvec = dveclab(gr_keep, :, ii_gr);
        grain{ii_gr}.allblobs.dvecsam = dvecsam(gr_keep, :, ii_gr);

        grain{ii_gr}.allblobs.srot = rot_l2s(:, :, gr_keep, ii_gr);
        grain{ii_gr}.allblobs.ondet = ondet(gr_keep, ii_gr);

        grain{ii_gr}.full = [];
    end
else
    % Initialse output variables
    grain.allblobs.uvw = uvw(:, keep)';
    grain.allblobs.uv = uvw(1:2, keep)';
    grain.allblobs.plorig = plorig(:, keep)';
    grain.allblobs.pl = pls(:, keep)';
    grain.allblobs.pllab = pllab(:, keep)';
    grain.allblobs.hkl = hkl(keep, :);
    grain.allblobs.hklsp = hklsp(keep, :);
    grain.allblobs.sintheta = sinth(keep);
    grain.allblobs.theta = theta(keep);
    grain.allblobs.eta = eta(keep);
    grain.allblobs.thetatype = thetatypesp(keep);
    grain.allblobs.lorentzfac = Lfactor(keep);
    grain.allblobs.omega = om(keep);
    grain.allblobs.omind = omind(keep);
    grain.allblobs.dvec = dveclab(keep, :);
    grain.allblobs.dvecsam = dvecsam(keep, :);
    grain.allblobs.srot = rot_l2s(:, :, keep);
    grain.allblobs.ondet = ondet(keep);

    grain.full = [];
end

% Visual Feedback
if (app.showfigure)

    if (isempty(app.overlay))
        app.overlay = zeros(parameters.acq.ydet, parameters.acq.xdet);
    end

    figure();
    imshow(app.overlay, app.Grayscale);
    hold on

    if (app.axis)
        detusize=parameters.acq.xdet;
        detvsize=parameters.acq.ydet;

        % Image frame
        plot([0 detusize+1],[0 0],'k')
        plot([0 detusize+1],[detvsize+1 detvsize+1],'k')
        plot([0 0],[0 detvsize+1],'k')
        plot([detusize+1 detusize+1],[0 detvsize+1],'k')

        % Midlines
        plot([0 detusize+1],[detvsize/2+0.5 detvsize/2+0.5],'-.k')
        plot([detusize/2+0.5 detusize/2+0.5],[0 detvsize+1],'-.k')

        % Set figure propeties
        axis equal
        set(gca,'YDir','reverse')
        set(gca,'Position',[0.05 0.05 0.95 0.9])
        xlim([-detusize*0.1 detusize*1.1])
        ylim([-detvsize*0.1 detvsize*1.1])
        xlabel('U direction')
        ylabel('V direction')

        % Beam direction in image
        beamuv = Qdet*labgeo.beamdir';
        % Arrow in figure indicating beam direction
        quiver(uvorig(1),uvorig(2),beamuv(1),beamuv(2),1000,'-k','Linewidth',3)

        % Rotation axis direction in image
        rotuv = Qdet*labgeo.rotdir';
        % Arrow in figure indicating rotation axis direction
        quiver(uvorig(1),uvorig(2),rotuv(1),rotuv(2),1000,'-.k','Linewidth',3)
    end

    impixelinfo
    om = grain.allblobs.omega;
    uv = grain.allblobs.uv;

    cmap = jet(1001);
    omtoRGB = @(om) cmap(round((om)/(360)*1000+1),:);

    if (app.color)
        colormap(cmap)
        colorbar
        caxis([0 360])
    else
        colormap(gray);
    end

    for ii = 1 : length(uv)
        if (~isempty(uv(ii,1)))  % there might be no intersection when using offaxis detector configurations (vertical)
            if (app.color)
                plot(uv(ii,1), uv(ii,2), 'or', 'MarkerSize', app.Markersize, ...
                    'MarkerEdgeColor', omtoRGB(om(ii)));
            else
                plot(uv(ii,1), uv(ii,2), 'ob', 'MarkerSize', app.Markersize);
            end
        end
    end
end % showfigure

end % end of function
