classdef GtAssembleVol3D < handle
% GtAssembleVol3D  A class to assemble volumes
%
% Can be initialized as:
% assem = GtAssembleVol3D(varargin);
%
%   OPTIONAL INPUT (parse_pv_pairs):
%     add_to_filename        =                 .edf volume containing a set of existing grains
%     display_error_messages = <logical>       show error messages (Used?)
%     cache                  = <GtSimpleCache> object, in case you have a filled one, 
%                                              so you can avoid reloading things from
%                                              disk
%     update_status_cb       = <Callback>      update the status of the computation 
%                                              (used for instance in a gui). The
%                                              said callback is executed every time the loop
%                                              terminates the processing of an entry
%     autosave               = <logical>       Saves automatically after assembling
%     overlaps               = <string>        How to handle overlapping grains:
%                                              ('zero' | 'summed' | {'conflict'} | 'adaptive' | 'assign', 'parent')
%     list                   = <double>        if list is 'all', or [], load all grains.  Otherwise, only the list
%     deal_with_twins        = <logical>       Handles twins  true | {false}
%     use_parent_mask        = <logical>       Use parent mask {false}
%     convention             = <string>        Hexagonal axes convention {'X'} | 'Y'
%     merge_angle            = <double>        Merge angle in degrees
%
%   Version 004 10-05-2014 by LNervo
%     Added static method to retrieve assemble output
%     i.e. foutput = GtAssembleVol3D.loadPhaseTwinOutput(workingdir, phaseID)
%
%   Version 003 18-04-2014 by LNervo
%     Added parent mask for twins
%
%   Version 002 09-10-2013 by LNervo
%     Implemented twin detection for hcp materials; added 'convention' parameter
%     to switch to available conventions for crystal reference system for hcp
%     materials

    properties
        parameters = [];
        localPar   = [];

        volumes = [];

        redo = []; % Keeps track of the grains that had an error

        twinInfo       = {};    % In case of twins
        sigmas         = [];    % In case of twins

        symm           = {};
        spacegroup     = {};
        crystal_system = {};

        conflictStrategies = {'summed', 'zero', 'conflict', 'adaptive', 'assign', 'parent'};
    end

    methods (Access = public)

        function obj = GtAssembleVol3D(varargin)
        % GTASSEMBLEVOL3D/GTASSEMBLEVOL3D  Constructor

            % Load Parameters
            obj.parameters = load('parameters.mat');
            obj.parameters = obj.parameters.parameters;
            
            for ii=1:obj.parameters.acq.nof_phases
                obj.crystal_system{ii} = obj.parameters.cryst(ii).crystal_system;
                obj.spacegroup{ii} = obj.parameters.cryst(ii).spacegroup;
                obj.symm{ii} = gtCrystGetSymmetryOperators(obj.crystal_system{ii}, obj.spacegroup{ii});
            end

            % Array of all the partial phase volumes
            obj.volumes.phases = {};
            % Complete volume: merge of all phases has two fields which really
            % hold the data:
            % - 'phases' which contains the information about the phaseIDs in
            %   all the voxels
            % - 'grains' which contains the information about the grainIDs in
            %   all the voxels
            obj.volumes.complete = [];
            obj.volumes.complete.phases = [];
            obj.volumes.complete.grains = [];

            % handle variables
            obj.localPar = GtAssembleVol3D.generateDefaults();

            obj.changeParams(varargin);
        end % end function GtAssembleVol3D

        function rejected = changeParams(obj, arguments)
        % GTASSEMBLEVOL3D/CHANGEPARAMS Changes the local parameters, validates
        % them, and prepares for the correct usage of them

            if (nargout > 0)
                [obj.localPar, rejected] = parse_pv_pairs(obj.localPar, arguments);
            else
                obj.localPar = parse_pv_pairs(obj.localPar, arguments);
            end

            % not that "zeros" is ambiguous in Matlab because it is a function,
            % and in this context we normally use it as a string.  Use "zero"
            % instead, small bodge here in case this is not done.
            if ((~ischar(obj.localPar.overlaps) && (obj.localPar.overlaps == 0)) ...
                    || (ischar(obj.localPar.overlaps) && strcmpi(obj.localPar.overlaps, 'zeros')) )
                obj.localPar.overlaps = 'zero';
                disp('guessing that overlaps should be set to zero')
            end

            if (~ischar(obj.localPar.overlaps) ...
                    || ~ismember(obj.localPar.overlaps, obj.conflictStrategies))
                strats = obj.conflictStrategies;
                strats(2,:) = {'" | "'};
                gtError('ASSEMBLE:wrong_parameter', ...
                      'You should specify a valid "overlap" behavior: "%s" is not among ( "%s" )', ...
                      obj.localPar.overlaps, [strats{1:end-1}] )
            end

            % Initializing cache
            if (isempty(obj.localPar.cache))
                obj.localPar.cache = GtSimpleCache();
            end
            if (~obj.localPar.cache.isModel('sample'))
                grainDirsModel = fullfile('4_grains', 'sample.mat');
                obj.localPar.cache.createModel(grainDirsModel, 'sample');
            end
            if (~obj.localPar.cache.isModel('grain'))
                grainDirsModel = fullfile('4_grains', 'phase_%02d', 'grain_%04d.mat');
                obj.localPar.cache.createModel(grainDirsModel, 'grain');
            end

            if (obj.localPar.deal_with_twins)
                disp('dealing with twins')
            else
                disp('not dealing with twins')
            end
        end % end function changeParams

        function assemblePhase(obj, phaseID)
        % GTASSEMBLEVOL3D/ASSEMBLEPHASE

            if (~exist('phaseID', 'var'))
                phaseID = 1;
            end

            [sampleInfo, list] = obj.initialiseAssembling(phaseID);

            if (obj.parameters.acq.no_direct_beam)
                disp('WARNING! No direct beam in the scan.')
                disp('  Revise sample bounding box before continuing!')
            end

            if (any(obj.parameters.acq.bb(3) == [0 1 NaN]) || ...
                any(obj.parameters.acq.bb(4) == [0 1 NaN]))
                disp('WARNING! Invalid value for ''parameters.acq.bb'' .')
            end

            % Let's initialize the volume
            volume = zeros(obj.parameters.acq.bb(3), ...
                           obj.parameters.acq.bb(3), ...
                           obj.parameters.acq.bb(4));

            if (obj.localPar.deal_with_twins)
                volume = obj.calculatePhaseVolumeTwins(volume, sampleInfo, phaseID, list);
                obj.savePhaseTwinOutput(phaseID);
            else
                volume = obj.doAssemblePhaseVolume(volume, sampleInfo, phaseID, list);
            end

            % store volume
            obj.volumes.phases{phaseID} = volume;

            if (obj.localPar.autosave)
                obj.savePhaseVolume(phaseID);
            end
        end % end function assemblePhase

        function assembleAllPhases(obj)
        % GTASSEMBLEVOL3D/ASSEMBLEALLPHASES

            % Now let's load metadata
            sampleInfo = obj.localPar.cache.get('sample', {}, 'sample');

            % For all phases:
            for phaseID = 1:length(sampleInfo.phases)
                obj.assemblePhase(phaseID);
            end
        end % end function assembleAllPhases

        function assembleInExistingVolume(obj, phaseID, filename)
        % GTASSEMBLEVOL3D/ASSEMBLEINEXISTINGVOLUME Discouraged function to write
        % a new grain map, to a previously existing grain map.

            if (~exist('phaseID', 'var'))
                phaseID = 1;
            end

            [sampleInfo, list] = obj.initialiseAssembling(phaseID);

            % look for the file
            if (~exist('filename', 'var'))
                filename = obj.localPar.add_to_filename;
            end
            absolute_file_path = fullfile(obj.parameters.acq.dir, ...
                                          '5_reconstruction', filename);

            if exist(filename, 'file')
                volume = edf_read(filename);
                % check if it's in the 5_reconstruction directory
            elseif exist( absolute_file_path, 'file')
                volume = edf_read(absolute_file_path);
                % otherwise, complain
            else
                error('ASSEMBLE:wrong_argument', ...
                      ['Cannot find the volume to add to: %s\ncheck the' ...
                       ' filename / path\n'], ...
                       filename)
            end

            % if we want to add grains to an existing volume, this is loaded
            % in. Then labels in the list are set to zero, before the grains
            % are "re-added" to the volume. I don't think this will deal
            % with overlapping grains correctly.
            if (max(volume(:)) > 0)
                for ii = 1:length(list)
                    volume(obj.volume == list(ii)) = 0;
                end
            end

            volume = obj.doAssemblePhaseVolume(volume, sampleInfo, phaseID, list);

            % Dumb saving.
            name = fullfile('5_reconstruction', filename);
            fprintf('writing volume %s\n', name);
            edf_write(volume, name, 'uint16');
        end % end function assembleInExistingVolume

        function assembleCompleteVolume(obj)
        % GTASSEMBLEVOL3D/ASSEMBLEFINALVOLUME

            % Let's load metadata
            sampleInfo = obj.localPar.cache.get('sample', {}, 'sample');

            % Let's initialize the phase IDs volumes
            volumePIDs = zeros( obj.parameters.acq.bb(3), ...
                                obj.parameters.acq.bb(3), ...
                                obj.parameters.acq.bb(4), 'int16' );
            % Let's initialize the grains IDs volumes
            volumeGIDs = zeros( obj.parameters.acq.bb(3), ...
                                obj.parameters.acq.bb(3), ...
                                obj.parameters.acq.bb(4), 'int16' );

            phasesNum = length(sampleInfo.phases);
            gauge = GtGauge(phasesNum, 'Processing phases: ');
            for ii = 1:phasesNum
                gauge.incrementAndDisplay();
                % no real need to cache phase volumes
                phaseVol = gtMATVolReader(sampleInfo.phases{ii}.volumeFile);
                grainMap = phaseVol;
                phaseVol = int16(phaseVol ~= 0) * ii;

                maskFile = fullfile('5_reconstruction', ...
                                    sprintf('phase_%02d_mask.mat', ii));
                if (exist(maskFile, 'file'))
                    maskVol = gtMATVolReader([maskFile ':mask']);
                    phaseVol = phaseVol .* int16(maskVol);
                    grainMap = grainMap .* int16(maskVol);
                end

                switch (obj.localPar.overlaps)
                    case {'zero', 'conflict', 'adaptive', 'parent'}
                        overlappingPoints = ((volumePIDs ~= 0) & (phaseVol ~= 0));
                        volumePIDs = volumePIDs + phaseVol;
                        volumePIDs(overlappingPoints) = -1;

                        volumeGIDs = volumeGIDs + grainMap;
                        volumeGIDs(overlappingPoints) = -1;
                    case {'summed'}
                        volumePIDs = volumePIDs + phaseVol;
                        volumeGIDs = volumeGIDs + grainMap;

                end
            end
            gauge.delete();

            % Shouldn't really happen, but it happens :)
            if (strcmpi(obj.localPar.overlaps, 'zero'))
                fprintf('Setting overlaps to 0..');
                numBefore = length(find(volumePIDs ~= 0));

                % set overlapping regions to zero - this could be another label colour
                volumePIDs(volumePIDs == -1) = 0;
                volumeGIDs(volumeGIDs == -1) = 0;

                numAfter = length(find(volumePIDs ~= 0));
                fprintf(' Done. (%d -> %d)\n', numBefore, numAfter);
            end

            obj.volumes.complete.phases = volumePIDs;
            obj.volumes.complete.grains = volumeGIDs;

            % Autosaving
            if (obj.localPar.autosave)
                obj.saveCompleteVolume();
            end
        end % end function assembleCompleteVolume

        function savePhaseVolume(obj, phaseID, fileName)
        % GTASSEMBLEVOL3D/SAVEPHASEVOLUME
        
            if (~exist('phaseID', 'var'))
                phaseID = 1;
            end
            if (~exist('fileName', 'var'))
                fileName = sprintf('phase_%02d_vol.mat', phaseID);
            end
            name = obj.checkAndSaveExistingFile(fileName);

            % Small trick to keep consistency in field name in the written file.
            outStruct = [];
            volField = 'vol';
            outStruct.(volField) = int16(obj.volumes.phases{phaseID});

            % We lock, to avoid other processes to write to this file:
            % Actually this is commented out because it should happen
            % through the DB, as long as the ESRF's filesystem is not good
            % enough
%             lock = GtLock(fullfile('4_grains', 'sample.mat'));
%             lock.acquire();

            % Now let's load metadata
            sampleInfo = obj.localPar.cache.get('sample', {}, 'sample');
            obj.checkSampleInfo(sampleInfo, phaseID);

            % Store info about volume in the metadata
            sampleInfo.phases{phaseID}.volumeFile = [name ':' volField];
            obj.localPar.cache.set('sample', {}, sampleInfo, 'sample');
            % Flush metadata
            fprintf('Flushing sample.mat\n');
            obj.localPar.cache.flushObjectField('sample', {}, 'sample');

            % Let's release the lock
%             lock.release();

            % Saving the data
            fprintf('Writing phase %02d volume to: %s\n', phaseID, name);
            save(name, '-struct', 'outStruct', volField, '-v7.3');

        end % end function savePhaseVolume

        function saveCompleteVolume(obj, fileName)
        % GTASSEMBLEVOL3D/SAVECOMPLETEVOLUME
        
            if (~exist('fileName', 'var'))
                fileName = 'volume.mat';
            end
            name = obj.checkAndSaveExistingFile(fileName);

            % We lock, to avoid other processes to write to this file:
            lock = GtLock(fullfile('4_grains', 'sample.mat'));
            lock.acquire();

            % Now let's load metadata
            sampleInfo = obj.localPar.cache.get('sample', {}, 'sample');

            % Store info about volume in the metadata
            sampleInfo.volumeFile = name;
            obj.localPar.cache.set('sample', {}, sampleInfo, 'sample');
            % Flush metadata
            fprintf('Flushing sample.mat\n');
            obj.localPar.cache.flushObjectField('sample', {}, 'sample');

            % Let's release the lock
            lock.release();

            % Saving the data
            fprintf('Writing complete volume to: %s\n', name);
            outStruct = obj.volumes.complete;
            save(name, '-struct', 'outStruct', 'phases', 'grains', '-v7.3');

        end % end function saveCompleteVolume

        function dilateCompleteVolume(obj, dilate_length, outFileName, inFileName)
        % GTASSEMBLEVOL3D/DILATECOMPLETEVOLUME

            if (~exist('outFileName', 'var'))
                outFileName = 'volume_dilated.mat';
            end
            name = obj.checkAndSaveExistingFile(outFileName);

            if (~exist('inFileName', 'var'))
                inFileName = fullfile('5_reconstruction', 'volume.mat');
            end

            compositeVol = load(inFileName);

            [uidVol, offset] = GtAssembleVol3D.convertCombinedVolumeToUIDs(compositeVol.phases, compositeVol.grains);

            % Try to see if there's a mask, and in case, use it
            maskFile = fullfile('5_reconstruction', 'volume_mask.mat');
            if (exist(maskFile, 'file'))
                maskVol = gtMATVolReader([maskFile ':vol']);

                fprintf('Doing %02d single dilation steps:\n', dilate_length)
                for step = 1:dilate_length
                    fprintf(' - %02d/%02d: ', step, dilate_length)
                    uidVol = gtDilateGrains(uidVol, 1, false);

                    fprintf('          Applying Absorption Mask.\n')
                    % Let's cut of the things out of the mask
                    newUidVol = zeros(size(uidVol), class(uidVol));
                    indexes = (maskVol ~= 0);
                    newUidVol(indexes) = uidVol(indexes);
                    uidVol = newUidVol;
                end
                fprintf('Done.\n')
            else
                uidVol = gtDilateGrains(uidVol, dilate_length, false);
            end

            [compositeVol.phases, compositeVol.grains] = GtAssembleVol3D.convertUIDsToCombinedVolume(uidVol, offset);

            save(name, '-struct', 'compositeVol', 'phases', 'grains', '-v7.3');

        end % end function dilateCompleteVolume

        function twinInfo = getTwinInfo(obj)
        % GTASSEMBLEVOL3D/GETTWININFO
             twinInfo = obj.twinInfo;
        end % end function getTwinInfo

        function redoIDs = getRedoList(obj)
            redoIDs = obj.redo;
        end

    end % end of public methods

    methods (Access = private)

        function checkSampleInfo(~, sampleInfo, phaseID)
        % GTASSEMBLEVOL3D/CHECKSAMPLEINFO

            if ((length(sampleInfo.phases) < phaseID) ...
                    || isempty(sampleInfo.phases{phaseID}))
                error('ASSEMBLE:wrong_argument', ...
                      'Phase: %02d, is not in the sample structure', phaseID);
            end
        end % end function checkSampleInfo

        function list = getGrainsList(obj, grainsNum)
        % GTASSEMBLEVOL3D/GETGRAINSLIST

            if isempty(obj.localPar.list) %|| strcmpi(obj.localPar.list, 'all')
                list = 1:grainsNum;
            else
                list = obj.localPar.list;
            end
        end % end function getGrainsList

        function finalName = checkAndSaveExistingFile(~, fileName)
        % GTASSEMBLEVOL3D/CHECKANDSAVEEXISTINGFILE  Checks if the file exists and
        % saves it to the archive directory.

            if (isempty(fileName))
                error('ASSEMBLE:wrong_argument', ...
                      'not saving data - to save, supply a filename')
            end
            finalName = fullfile('5_reconstruction', fileName);

            if (exist(finalName, 'file'))
                [~, tempFilename, tempExt] = fileparts(finalName);
                fileInfo = dir(finalName);
                destName = [tempFilename ...
                            datestr(fileInfo(1).datenum, '_yyyy-mm-dd_HH-MM-SS') ...
                            tempExt];
                if ~exist(fullfile('5_reconstruction', 'archive'),'dir')
                    [~,msg]=mkdir(fullfile('5_reconstruction', 'archive')); disp(msg)
                end
                copyfile(finalName, fullfile('5_reconstruction', 'archive', destName));
            end
        end % end function checkAndSaveExistingFile

        function [sampleInfo, list] = initialiseAssembling(obj, phaseID)
        % GTASSEMBLEVOL3D/INITIALISEASSEMBLING  Initialises the support structures 
        % needed for assembly.

            % Now let's load metadata
            sampleInfo = obj.localPar.cache.get('sample', {}, 'sample');
            % And check as possible
            obj.checkSampleInfo(sampleInfo, phaseID);
            % Grain Number
            grainsNum = length(sampleInfo.phases{phaseID}.grains);

            % handle list
            list = obj.getGrainsList(grainsNum);
        end % end function initialiseAssembling

        function volume = doAssemblePhaseVolume(obj, volume, sampleInfo, phaseID, list)
        % GTASSEMBLEVOL3D/DOASSEMBLEPHASEVOLUME  Function that effectively assembles phase volume
        
            grainsNum = length(list);
            gauge = GtGauge(grainsNum, sprintf('Phase %02d: ', phaseID));

            for ii = list;
                gauge.setExtraText(sprintf('(grain %04d)', ii));
                gauge.incrementAndDisplay();

                if (~sampleInfo.phases{phaseID}.getSelected(ii))
                    fprintf('\nSkipping deselected grain: %04d\n', ii);
                    gauge.rePrint();
                    continue;
                end

                try
                    vol = obj.localPar.cache.get('grain', {phaseID, ii}, 'seg');
                    volBBox = sampleInfo.phases{phaseID}.getBoundingBoxExtremes(ii);
                    volume = gtPlaceSubVolume(volume, vol, volBBox(1:3), ii, obj.localPar.overlaps);
                catch mexc
                    gtPrintException(mexc, sprintf('\nGrain %d failed!!\n', ii));
                    gauge.rePrint();
                    obj.redo = [obj.redo ii];
                end % end try catch loop

                % Trigger callback (usually used in graphical applications)
                gtTriggerUpdateCallback(obj.localPar.update_status_cb);
            end  % for loop through all grains

            if (strcmpi(obj.localPar.overlaps, 'zero'))
                % set overlapping regions to zero - this could be another label colour
                volume(volume == -1) = 0;
            end
            gauge.delete();
        end % end function doAssemblePhaseVolume

        function volumeUpdate = calculatePhaseVolumeTwins(obj, volume, sampleInfo, phaseID, list)
        % GTASSEMBLEVOL3D/CALCULATEPHASEVOLUMETWINS
        % obj has the sigma as a property - not sure if useful like this
        % volume is zeros to build the new volume in
        % sample info includes bboxs and R vectors
        % phaseID - do we need?
        % list - grains to treat
        %
        % concept - place grain vol in a sample sized box
        % set undisputed voxels to the grainID
        % get a list of which grains overlap - test these for being twins
        % if they are twins, set the overlap to the grainID (principle of
        % largest first)
            obj.sigmas{phaseID} = gtCrystGetSigmas(obj.parameters.cryst(phaseID).crystal_system, ...
                obj.parameters.cryst(phaseID).latticepar, ...
                obj.localPar.convention);

            % how many grains to treat?
            grainsNum = length(list);

            index = obj.getGrainsSortedBySize(phaseID,list,'descend');

            output     = cell(grainsNum,1);
            twin_list  = zeros(grainsNum, 0);
            check_list = zeros(grainsNum, 0);
            merge_list = zeros(grainsNum, 0);

            gauge = GtGauge(grainsNum, sprintf('Phase %02d: ', phaseID));
            foutput = cell(0,6);
            % loop through in descending size order
            for kk = 1:length(list)
                ii = index(kk);

                %fprintf('%04d/%04d (grain %04d)', kk, grainsNum, ii);
                gauge.setExtraText(sprintf('(grain %04d)', ii));
                gauge.incrementAndDisplay();

                % skip deselected grains
                if (~sampleInfo.phases{phaseID}.getSelected(ii))
                    foutput{end+1,5} = sprintf('\nDeselected grainID %4d...', ii);
                    continue;
                end

                % read the grain volume
                try
                    vol = obj.localPar.cache.get('grain', {phaseID, ii}, 'seg');
                    volBBox = sampleInfo.phases{phaseID}.getBoundingBoxExtremes(ii);

                    % Getting the intersecting volume of complete volume and
                    % segmented volume
                    if ~isempty(vol)
                        [limsVolume, limsGrain] = gtGetVolsIntersectLimits(size(volume), size(vol), volBBox(1:3));
                    else
                        foutput{end+1,2} = sprintf('\nEmpty segmented volume for grainID %4d... Please threshold it again!', ii);
                        continue
                    end
                    % Getting the subvolumes -> so we operate all the costy
                    % operations on smaller volumes
                    subVolume = volume( limsVolume(1, 1):limsVolume(2, 1), ...
                                        limsVolume(1, 2):limsVolume(2, 2), ...
                                        limsVolume(1, 3):limsVolume(2, 3) );
                    subGrain = vol( limsGrain(1, 1):limsGrain(2, 1), ...
                                    limsGrain(1, 2):limsGrain(2, 2), ...
                                    limsGrain(1, 3):limsGrain(2, 3) );

                    % what grains overlap?
                    disputes = subVolume(find(subGrain));
                    disputes = unique(disputes);
                    disputes(find(disputes == 0 | disputes == -1)) = [];

                    % first, assign the non disputed part
                    indexes_notshared = find((subVolume == 0) & subGrain);
                    
                    subVolume(indexes_notshared) = ii;
                    info  = {};
                    % now, go through the disputed grains checking orientations
                    if ~isempty(disputes)
                        refRvector = sampleInfo.phases{phaseID}.getR_vector(ii);
                        twincount = 0;
                        checkcount = 0;
                        for jj = 1:length(disputes)
                            % test for being a twin
                            testRvector = sampleInfo.phases{phaseID}.getR_vector(disputes(jj));
                            [info{jj}, ~] = gtCrystTwinTest(refRvector, testRvector, phaseID, ...
                                'symm', obj.symm{phaseID}, 'sigmas', obj.sigmas{phaseID}, ...
                                'merge_angle', obj.localPar.merge_angle, 'convention', obj.localPar.convention);
                            info{jj}.grainID_1 = ii;
                            info{jj}.grainID_2 = disputes(jj);

                            % Test if it is twin of another grain
                            indexes = find((subVolume == disputes(jj)) & subGrain);

                            info{jj}.indexes_shared = indexes;
                            info{jj}.size_vol_free  = length(indexes_notshared);
                            info{jj}.size_vol       = length(indexes_notshared) + length(indexes);
                            info{jj}.frac_shared    = length(indexes)/info{jj}.size_vol;

                            if info{jj}.sigmaAnd == -1
                                if info{jj}.frac_shared >= 0.5
                                    % merge the two grains together; keep the parent
                                    grain_label = disputes(jj);
                                    %subVolume(indexes) = grain_label;
                                    foutput{end+1,3} = sprintf(...
                                        '\nGrains need merge (?): grain %4d and %4d - mis_angle %0.2f degrees (shared volume fraction %0.2f)', ...
                                        ii, grain_label, info{jj}.mis_angle, info{jj}.frac_shared);
                                    % record this in the merge list
                                    merge_list(ii, 1) = ii;
                                    merge_list(ii, 2) = grain_label;
                                end
                            elseif ~isempty(info{jj}.sigmaAnd)
                                if info{jj}.frac_shared >= 0.5
                                    % this is a twin.  Allow the new grain (smaller)
                                    % to overwrite the older grain
                                    subVolume(indexes) = ii;
                                    if (obj.localPar.use_parent_mask)
                                        subVolume(indexes_notshared) = 0;
                                    end
                                    foutput{end+1,1} = sprintf(...
                                        '\nFound twin %4d for grainID %4d with mis_angle %0.2f degrees (shared volume fraction %0.2f)', ...
                                        ii, disputes(jj), info{jj}.mis_angle, info{jj}.frac_shared);
                                    twin_list(ii, 1) = ii;
                                    twin_list(ii, twincount+2) = disputes(jj);
                                    twincount=twincount+1;
                                else
                                    % MAYBE this is a twin.  Allow the new grain (smaller)
                                    % to overwrite the older grain: PROBLEM
                                    % WITH THE PARENT SEGMENTATION
                                    subVolume(indexes) = ii;
                                    if (obj.localPar.use_parent_mask)
                                        subVolume(indexes_notshared) = 0;
                                    end
                                    foutput{end+1,6} = sprintf(...
                                        '\nFound POSSIBLE twin %4d for grainID %4d with mis_angle %0.2f degrees (shared volume fraction %0.2f)', ...
                                        ii, disputes(jj), info{jj}.mis_angle, info{jj}.frac_shared);
                                    check_list(ii, 1) = ii;
                                    check_list(ii, checkcount+2) = disputes(jj);
                                    checkcount=checkcount+1;
                                end
                            end
                            if isempty(info{jj}.sigmaAnd)
                                if info{jj}.frac_shared < 0.5
                                    % this is not a twin, so it is disputed
                                    switch (obj.localPar.overlaps)
                                        case {'zero', 'conflict', 'adaptive'}
                                            subVolume(indexes) = -1;
                                        case 'summed'
                                            subVolume(indexes) = subVolume(indexes) + ii;
                                        case 'assign'
                                            subVolume(indexes) = ii;
                                        case 'parent'
                                            subVolume(indexes) = disputes(jj);
                                    end
                                end
                            end
                        end
                        % Now let's get back the volume in the complete one
                        volume = gtPlaceSubVolume(volume, subVolume, volBBox(1:3), ...
                                                    0, 'assign');
                    else
                        % Ok, no conflict or no twin case -> let's treat it as
                        % normal case
                        volume = gtPlaceSubVolume(volume, vol, volBBox(1:3), ...
                                                    ii, obj.localPar.overlaps);
                    end

                    output{kk} = info;
            
                catch mexc
                    gtPrintException(mexc, sprintf('\nGrain %d failed!!\n', ii));
                    foutput{end+1,4} = sprintf('\nFailed grainID %4d... Please check it!', ii);
                    gauge.rePrint();
                    obj.redo = [obj.redo ii];
                end % try catch loop
            end
            gauge.delete();
            if size(foutput(:,1),1) > 0
                disp([foutput{:,1}])
            end
            if size(foutput(:,2),1) > 0
                disp([foutput{:,2}])
            end
            if size(foutput(:,3),1) > 0
                disp([foutput{:,3}])
            end
            if size(foutput(:,4),1) > 0
                disp([foutput{:,4}])
            end
            if size(foutput(:,5),1) > 0
                disp([foutput{:,5}])
            end
            if size(foutput(:,6),1) > 0
                disp([foutput{:,6}])
            end
            obj.twinInfo{phaseID}.output     = output;
            obj.twinInfo{phaseID}.foutput    = foutput;
            obj.twinInfo{phaseID}.twin_list  = twin_list;
            obj.twinInfo{phaseID}.check_list = check_list;
            obj.twinInfo{phaseID}.merge_list = merge_list;
            obj.twinInfo{phaseID}.fail_list  = obj.redo;
            
            if (strcmpi(obj.localPar.overlaps, 'zero'))
                % set overlapping regions to zero - this could be another label colour
                volume(volume == -1) = 0;
            end

            if nargout > 0
                volumeUpdate = volume;
            end
        end % end function calculatePhaseVolumeTwins

        function savePhaseTwinOutput(obj, phaseID)
        % GTASSEMBLEVOL3D/SAVEPHASETWINOUTPUT  
        % saving twin info on disk
            foutput    = obj.twinInfo{phaseID}.foutput;
            twinOut    = obj.twinInfo{phaseID}.output;
            twinOut = [twinOut{:}];
            twin_list  = obj.twinInfo{phaseID}.twin_list;
            check_list = obj.twinInfo{phaseID}.check_list;
            merge_list = obj.twinInfo{phaseID}.merge_list;
            fail_list  = obj.twinInfo{phaseID}.fail_list;

            disp(' ')
            if ~isempty(foutput)
                save(fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'foutput.mat'), ...
                    'foutput', '-v7.3');
                disp(['Saved twin output in ' fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'foutput.mat')])
            end
            if ~isempty(twinOut)
                save(fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'twinOut.mat'), ...
                    'twinOut', '-v7.3');
                disp(['Saved twin output in ' fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'twinOut.mat')])
            end
            if ~isempty(twin_list)
                % remove third column
                if size(twin_list,2) == 3
                    tmp = twin_list(twin_list(:,3)~=0,:);
                    if ~isempty(tmp)
                        tmp(:,1) = tmp(:,3);
                        tmp(:,3) = [];
                        twin_list(:,3) = [];
                        twin_list(end+1:end+size(tmp,1),:) = tmp;
                    end
                end
                twin_list(twin_list(:,1)==0,:) = [];
                twin_list = unique(twin_list, 'rows');

                save(fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'twin_list.mat'), ...
                    'twin_list', '-v7.3');
                disp(['Saved twin list in ' fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'twin_list.mat')])
            end
            if ~isempty(check_list)
                % remove third column
                if size(check_list,2) == 3
                    tmp = check_list(check_list(:,3)~=0,:);
                    if ~isempty(tmp)
                        tmp(:,1) = tmp(:,3);
                        tmp(:,3) = [];
                        check_list(:,3) = [];
                        check_list(end+1:end+size(tmp,1),:) = tmp;
                    end
                end
                check_list(check_list(:,1)==0,:) = [];
                check_list = unique(twin_list, 'rows');

                save(fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'check_list.mat'), ...
                    'check_list', '-v7.3');
                disp(['Saved check list in ' fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'check_list.mat')])
            end
            if ~isempty(merge_list)
                merge_list(merge_list(:,1)==0,:) = [];

                save(fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'merge_list.mat'), ...
                    'merge_list', '-v7.3');
                disp(['Saved merge list in ' fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'merge_list.mat')])
            end
            if ~isempty(fail_list)
                fail_list(fail_list(:,1)==0,:) = [];

                save(fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'fail_list.mat'), ...
                    'fail_list', '-v7.3');
                disp(['Saved fail list in ' fullfile(obj.parameters.acq.dir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'fail_list.mat')])
            end
            
            
            obj.twinInfo{phaseID}.foutput    = foutput;
            obj.twinInfo{phaseID}.output     = twinOut;
            obj.twinInfo{phaseID}.twin_list  = twin_list;
            obj.twinInfo{phaseID}.check_list = check_list;
            obj.twinInfo{phaseID}.merge_list = merge_list;
            obj.twinInfo{phaseID}.fail_list  = fail_list;
            
            info = obj.twinInfo{phaseID};
            if ~isempty(info)
                save(fullfile(obj.parameters.acq.dir, ...
                    '4_grains',  sprintf('phase_%02d', phaseID), 'twinInfo.mat'), ...
                    'info', '-v7.3');
                disp(['Saved twin info in ' fullfile(obj.parameters.acq.dir, ...
                    '4_grains',  sprintf('phase_%02d', phaseID), 'twinInfo.mat')])
            end
            
            
        end % end function savePhaseTwinOutput

        function index = getGrainsSortedBySize(obj, phaseID, list, sortOrder)
        % GTASSEMBLEVOL3D/
            % want to work in reverse size order
            grainsNum = length(list);
            if (~exist('sortOrder','var'))
                sortOrder = 'descend';
            end

            grainsize = zeros(grainsNum, 1);
            for ii = 1:grainsNum
                try
                    vol = obj.localPar.cache.get('grain', {phaseID, list(ii)}, 'seg');
                    grainsize(ii) = length(find(vol(:)));
                catch mexc
                    % We ignore the error because it will come up again
                    % later
                    grainsize(ii) = 0;
                end
            end
            % sort - to do largest grain first
            [~, index] = sort(grainsize, sortOrder);
        end
        
        
    end % end of private methods

    methods (Static, Access = public)

        function conf = generateDefaults()
        % GTASSEMBLEVOL3D/GENERATEDEFAULTS  Generates default parameters

            % handle variables
            conf.overlaps = 'conflict';         % how to handle overlapping grains:
                                                %   ('zero' | 'summed' | {'conflict'} | 'adaptive' | 'assign' | 'parent')
            conf.add_to_filename = [];          % to add grains to a preexisting volume (saved as edf)
            conf.list = [];                     % if list is 'all', or [], load all grains.  Otherwise, only the list
            conf.display_error_messages = true; % show error messages {true} | false
            conf.update_status_cb = [];         % Callback to update the status of anything
            conf.cache = [];                    % Grain volumes cache (to not reload all of them each time)
            conf.autosave = true;               % Saves automatically the volumes after assembling them {true} | false
            conf.autoload = true;               % Loads automatically the assembled volume in the GM gui {true} | false
            conf.deal_with_twins = false;       % true | {false}
            conf.use_parent_mask = false;       % use a mask before dilating if it is a parent/twin case
            conf.convention = 'X';              % Hexagonal axes convention {'X'} | 'Y' for assembling volume with deal_with_twins=true
            conf.merge_angle = 5;               % misorientation angle for two grains to be merged {5}

        end % end function generateDefaults

        function [newVol, offset] = convertCombinedVolumeToUIDs(phaseIDs, grainIDs, offset)
        % GTASSEMBLEVOL3D/CONVERTCOMBINEDVOLUMETOUIDS  Converts from a twofold matrix, 
        % into a unique ID matrix.

            % The conflicts are automatically set to zero
            maxGrainID_1 = max(grainIDs(:)) +1;
            if (~exist('offset', 'var'))
                offset = maxGrainID_1;
            elseif (offset < maxGrainID_1)
                error('ASSEMBLE:wrong_argument', ...
                      'bad offset, there could be overlaps in IDs');
            end
            indexes = (grainIDs < 0);
            grainIDs(indexes) = 0;
            phaseIDs(indexes) = 0;
            newVol = grainIDs + (phaseIDs -1) * offset;
            % We need to kill the conflict regions if they exist, and the zeros.
            newVol(newVol < 0) = 0;
        end % end function convertCombinedVolumeToUIDs

        function [phaseIDs, grainIDs] = convertUIDsToCombinedVolume(uidVol, offset)
        % GTASSEMBLEVOL3D/CONVERTUIDSTOCOMBINEDVOLUME  Converts back to a twofold
        % matrix, from a unique IDs matrix. The offset parameters is needed in
        % order to distinguish the phases.

            grainIDs = mod(uidVol, offset);
            uidVol = uidVol - grainIDs;
            phaseIDs = (floor(uidVol / offset) +1);
            phaseIDs(grainIDs == 0) = 0;
        end % end function convertUIDsToCombinedVolume

        function foutput = loadPhaseTwinOutput(wdir, phaseID, debug)
        % GTASSEMBLEVOL3D/LOADPHASETWINOUTPUT  loading twin info output from disk 
            if ~exist('wdir','var') || isempty(wdir)
                wdir = pwd();
            end
            if ~exist('debug','var') || isempty(debug)
                debug = false;
            end

            foutput = [];
            if exist(fullfile(wdir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'foutput.mat'),'file')
                load(fullfile(wdir, ...
                    '4_grains', sprintf('phase_%02d', phaseID), 'foutput.mat'));
                if (debug)
                    if size(foutput(:,1),1) > 0
                        disp([foutput{:,1}])
                    end
                    if size(foutput(:,2),1) > 0
                        disp([foutput{:,2}])
                    end
                    if size(foutput(:,3),1) > 0
                        disp([foutput{:,3}])
                    end
                    if size(foutput(:,4),1) > 0
                        disp([foutput{:,4}])
                    end
                    if size(foutput(:,5),1) > 0
                        disp([foutput{:,5}])
                    end
                    if size(foutput(:,6),1) > 0
                        disp([foutput{:,6}])
                    end
                end
                if nargout > 0
                    if exist(fullfile(wdir, '4_grains',  sprintf('phase_%02d', phaseID), 'twinInfo.mat'),'file')
                        foutput = load(fullfile(wdir, '4_grains',  sprintf('phase_%02d', phaseID), 'twinInfo.mat'));
                        foutput = foutput.info;
                    else
                        foutput = cat(1,foutput(:));
                        foutput(cellfun(@isempty,foutput)) = [];
                    end
                end
            else
                disp('File does not exist...')
                return
            end
            
        end

    end % end of static, public methods

end % end of class
