function proj = gtAstraPrepareAbsorptionStack(parameters)
% GTASTRAPREPAREABSORPTIONSTACK  Assembles the 3D projection stack of absorption
%                                images and saves them to file absorption_stack.mat
%
%     proj = gtAstraPrepareAbsorptionStack(parameters)
%     ------------------------------------------------
%     INPUT:
%       parameters <structure>
%
%     OUTPUT:
%       proj <structure> with fields:
%         - stack    <double X,X,Z>  3D stack of absorption images
%         - geom     <double Z,12 >  Vector describing Astra preojection geometry
%         - Omega    <Double Z>      Rotation angles in degrees
%
%
%     Version 002.1 26-09-2012 by AKing
%       Change sign of omega angles (used by 2DFBP) when axis is RH [0 0 1]
%
%     Version 002 26-09-2012 by LNervo
%       Formatting, cleaning, commenting

acq = parameters.acq;

labgeo = parameters.labgeo;
samgeo = parameters.samgeo;
recgeo = parameters.recgeo;

if (isfield(parameters.rec, 'absorption'))
    rec = parameters.rec.absorption;
else
    % Old style parameters
    warning('Old "rec" parameters')
    rec = struct(... % can be 'SIRT' or '2DFBP'
        'algorithm', parameters.rec.method, ...
        'num_iter', parameters.rec.abs_num_iter, ...
        'interval', parameters.rec.absinterval, ...
        'options', []);
end

if strcmp(acq.type, '360degree')
    totproj   = 2 * acq.nproj;
    increment = 2 * pi / totproj;
    angle_rad = 0 : increment : 2*pi-increment;
elseif strcmp(acq.type, '180degree')
    totproj   = acq.nproj;
    increment = pi / totproj;
    angle_rad = pi/2 : increment: 3/2*pi-increment;
else
    error('Scan type not recognised!')
end

% projection geometry... Here we assume that the absorption images are
% shifted such that the axis of rotation is in the center of the images
% The origin of the sample coordinate system is on the rotation axis - in
% the middle of the illuminated area
fprintf('Building geometry..')
imgs_nums = 0 : rec.interval : totproj-1;
Omega_rad = angle_rad(imgs_nums + 1);
Omega_deg = Omega_rad * 180 / pi;

bbpos = repmat(acq.bb, numel(Omega_deg), []);
geom = gtGeoProjForReconstruction([], Omega_deg', [], bbpos, [], labgeo, ...
    samgeo, recgeo, 'ASTRA_absorption');

fprintf('\b\b: Done.\nBuilding sinogram: ');
c = tic();
abs_files_dir = fullfile(acq.dir, '1_preprocessing', 'abs');
info = [];
stack = zeros(acq.bb(3), totproj/rec.interval, acq.bb(4));

for jj = 1:numel(imgs_nums)
    num_chars = fprintf('%03d/%03d', jj, numel(imgs_nums));

    if (acq.interlaced_turns == 0)
        absname = fullfile(abs_files_dir, sprintf('abs%04d.edf', imgs_nums(jj)));
    else
        absname = fullfile(abs_files_dir, sprintf('abs_renumbered%04d.edf', imgs_nums(jj)));
    end
    [abs_img, info] = edf_read(absname, [], false, info);

    stack(:, jj, :)  = -log(max(0.01,abs_img))';

    fprintf(repmat('\b', 1, num_chars))
end
fprintf('Done in %f seconds.\n', toc(c));

% deal with the sense of the rotation
if all(labgeo.rotdir == [0 0 1])
    % rotation axis is anti-clockwise
    Omega_rad = -Omega_rad;
end

fprintf('...done!\n')

proj.stack      = stack;
proj.Omega      = Omega_rad;
proj.geom       = geom;
proj.num_rows   = acq.bb(4);
proj.num_cols   = acq.bb(3);
proj.num_iter   = rec.num_iter;
proj.vol_size_x = acq.bb(3);
proj.vol_size_y = acq.bb(3);
proj.vol_size_z = acq.bb(4);


name = fullfile(acq.dir, 'absorption_stack.mat');
disp(['Writing sinogram in ' name]);
save(name, '-struct', 'proj');

end % end of function
