function [vols, sampling, gr, algo, sampler] = gtReconstructGrainOrientation(grainID, phaseID, parameters, sample)
    if ~exist('parameters','var') || isempty(parameters)
        load('parameters.mat');
    end

    if (~exist('sample','var'))
        samplefile = fullfile(parameters.acq.dir, '4_grains', 'sample.mat');
        sample = GtSample.loadFromFile(samplefile);
    end

    if (phaseID < 1 || phaseID > numel(sample.phases))
        error('gtReconstructGrainOrientation:wrong_phase_number', ...
            'Number of phases should be between [1, %d], but got %d', ...
            numel(sample.phases), phaseID)
    end
    if (grainID < 1 || grainID > numel(sample.phases{phaseID}.grains))
        error('gtReconstructGrainOrientation:wrong_grain_number', ...
            'Number of grains for phase %d should be between [1, %d], but got %d', ...
            phaseID, numel(sample.phases{phaseID}.grains), grainID)
    end

    selected = logical(sample.phases{phaseID}.grains{grainID}.selectedDiffspots);

    grain_dir = fullfile(parameters.acq.dir, '4_grains', sprintf('phase_%02d', phaseID));
    grain_file = fullfile(grain_dir, sprintf('grain_%04d.mat', grainID));
    grain_details_file = fullfile(grain_dir, sprintf('grain_details_%04d.mat', grainID));

    fields_to_load = {'R_vector', 'id', 'phaseid', 'allblobs', 'ondet', ...
        'included', 'proj', 'center', 'centerpix', 'bl', 'spotid', ...
        'intensity', 'omega'};

    if isempty(selected)
        fields_to_load = [fields_to_load, 'selected'];
        gr = load(grain_file, fields_to_load{:});
    else
        gr = load(grain_file, fields_to_load{:});
        gr.selected = selected;
    end

    %, 'spotid'
%     gr.selected = gtGuiGrainMontage(gr, selected, true);
%     GtVolView(gr.proj.stack(:, gr.selected, :))
%     pause

    rec = parameters.rec;
    if (isfield(rec, 'grains'))
        num_iter = rec.grains.num_iter;
        algorithm = rec.grains.algorithm;
    else
        num_iter = rec.num_iter;
        algorithm = '6DL1';
    end
    if (isfield(rec, 'grains') && isfield(rec.grains, 'options') && ~isempty(rec.grains.options))
        grid_edge = rec.grains.options.grid_edge;
        num_interp = rec.grains.options.num_interp;
        lambda = rec.grains.options.lambda;
        ss_grid_edge = rec.grains.options.super_sampling;
    else
        grid_edge = 9;
        num_interp = 8;
        lambda = 1e-1;
        ss_grid_edge = 1;
    end

%     sampler = GtOrientationSampling(gr.bl, parameters, gr, 'verbose', true);
    sampler = GtOrientationSampling(gr.bl, parameters, gr);
    sampler.make_simple_grid_estim_ODF('cubic', grid_edge, false, 1);
%     sampler.make_simple_grid_estim_ODF('bcc', grid_edge);
    if (ss_grid_edge > 1)
        sampler.make_oversampling_simple_grid([1 2 3], ss_grid_edge);
    end

    rec_factory = Gt6DReconstructionAlgorithmFactory();
    if (num_interp > 1)
        [algo, good_gr] = rec_factory.getSubBlobReconstructionAlgo(gr.bl, gr.selected, sampler, '3D', num_interp);
    else
        [algo, good_gr] = rec_factory.getBlobReconstructionAlgo(gr.bl, gr.selected, sampler, '3D');
    end
    switch (upper(algorithm))
        case '6DL1'
            algo.cp(num_iter, lambda);
        case '6DTV'
            algo.cpTV(num_iter, lambda);
        case '6DTVL1'
            algo.cpTVl1(num_iter, lambda);
        otherwise
            error('gtReconstructGrainOrientation:wrong_parameter', ...
                'Unknown 6D algorithm: "%s"', algorithm)
    end

    vols = algo.getCurrentSolution();

    [avg_R_vecs, avg_R_vecs_int] = sampler.getAverageOrientations(vols, good_gr);
    avg_R_vec = sampler.getAverageOrientation(vols, good_gr);
    s_g_odf = sampler.getODF(vols, good_gr);

    vol_size = size(avg_R_vecs_int);
    shift = round((parameters.acq.bb([3 3 4]) - vol_size) / 2 + gr.proj.centerpix) + 1;

    sampling = sampler.get_orientations();

    ODF6D = struct( ...
        'voxels_avg_R_vectors', {avg_R_vecs}, ...
        'intensity', {avg_R_vecs_int}, ...
        'shift', {shift}, ...
        'R_vectors', {sampler.get_R_vectors()}, ...
        'single_grain_ODF', {s_g_odf}, ...
        'single_grain_avg_R_vector', {avg_R_vec});
    save(grain_details_file, 'ODF6D', '-v7.3');
end

