function doneLocaly = gtReconstructGrains(first, last, workingdirectory, phaseID, parameters, varargin)
% GTRECONSTRUCTGRAINS  Launch a series of 3D ART (ASTRA) reconstruction on a GPU machine
%     doneLocaly = gtReconstructGrains(first, last, workingdirectory, phaseID, [parameters], [varargin])
%     --------------------------------------------------------------------------------------------------
%
%     If submitted on a GPU machine it will work locally - otherwise it will
%     submit the job via OAR to a machine with GPU (needed for 3D ASTRA code)
%
%     INPUT:
%       first            = <int>     first grainid
%       last             = <int>     last grainid
%       workingdirectory = <string>  working directory
%       phaseID          = <int>     phase ID
%       parameters       = <struct>  {parameters.mat}
%       varargin         =           parameters.rec
%
%     OUTPUT:
%       doneLocaly       = <logical> Check GPU machine; if not a gpu, it
%                                    is submitted to OAR
%
%     Version 001

currentDir = pwd;
cd(workingdirectory);

if (isdeployed)
    global GT_DB
    global GT_MATLAB_HOME
    load('workspaceGlobal.mat');
    first   = str2double(first);
    last    = str2double(last);
    phaseID = str2double(phaseID);
end

if ~exist('parameters','var') || isempty(parameters)
    load('parameters.mat');
end

if (isfield(parameters.rec, 'grains'))
    rec = parameters.rec.grains;
else
    % Old style parameters
    warning('Old "rec" parameters')
    rec = parameters.rec;
    rec.algorithm = 'SIRT';
end
dir_acq = parameters.acq.dir;

if ~isempty(varargin)
    rec = parse_pv_pairs(rec,varargin);
    parameters.rec = rec; %added
    save('parameters.mat','parameters'); %added
end

OK = gtCheckGpu();

if (~OK)
    disp('ASTRA:no_GPU   gtAstra3D: You should submit this job on a GPU machine ! Exiting...');
end

samplefile = fullfile(parameters.acq.dir, '4_grains', 'sample.mat');
sample = GtSample.loadFromFile(samplefile);

if ~isempty(rec.list)
    if (OK)
        gauge = GtGauge(length(rec.list), 'Reconstructing grains: ');
        for ii = 1 : length(rec.list)
            gauge.incrementAndDisplay();
            switch(upper(rec.algorithm))
                case 'SIRT'
                    gtAstraReconstructGrain(rec.list(ii), phaseID, parameters, sample);
                case {'6DL1', '6DTV', '6DTVL1'}
                    gtReconstructGrainOrientation(rec.list(ii), phaseID, parameters, sample);
            end
        end
        gauge.delete();
    else
        disp('If you provide a list of grains, you need to start this job from a gpu machine, Exiting...');
        return;
    end
else
    if (OK)
        gauge = GtGauge([first last], 'Reconstructing grains: ');
        for ii = first : last
            gauge.incrementAndDisplay();
            switch(upper(rec.algorithm))
                case 'SIRT'
                    gtAstraReconstructGrain(ii, phaseID, parameters, sample);
                case {'6DL1', '6DTV', '6DTVL1'}
                    gtReconstructGrainOrientation(ii, phaseID, parameters, sample);
            end
        end
        gauge.delete();
    else
        OAR_make('gtReconstructGrains', first, last, 1, [pwd ' ' num2str(phaseID)], true, 'gpu', true);
    end
end

if (nargout > 0)
    doneLocaly = OK;
end
cd(currentDir);

end % end of function
