function [info, sigma] = gtCrystTwinTest(R1, R2, phaseID, varargin)
% GTCRYSTTWINTEST  Tests if the two orientations are from two twins
%     [info, sigma] = gtCrystTwinTest(R1, R2, [phaseID], varargin)
%     ------------------------------------------------------------
%     INPUT:
%       R1 and R2   = Rodriguez vectors of the two different orientations
%       phaseID     = current phase
%
%     OPTIONAL INPUT (parse by pairs):
%       symm        = symmetry operators from gtCrystGetSymmetryOperators {[]}
%       sigmas      = sigma values from gtCrystGetSigmas {[]}
%       merge_angle = a threshold to decide if the angle is small {5}
%       convention  = hcp crystal axes convention {'X'}
%
%     Version 001 13-11-2013 by LNervo

if ~exist('phaseID','var') || isempty(phaseID)
    phaseID = 1;
end

app.symm   = [];
app.sigmas = [];
app.merge_angle = 5;
app.convention = 'X';
app = parse_pv_pairs(app, varargin);

parameters = [];
load('parameters.mat');
crystal_system = parameters.cryst(phaseID).crystal_system;
latticepar = parameters.cryst(phaseID).latticepar;
spacegroup = parameters.cryst(phaseID).spacegroup;
if isempty(app.symm)
    app.symm = gtCrystGetSymmetryOperators(crystal_system, spacegroup);
end
if isempty(app.sigmas)
    app.sigmas = gtCrystGetSigmas(crystal_system, latticepar, app.convention);
end

sigmas = app.sigmas;

sigma   = []; % default sigma
if strcmp(crystal_system, 'cubic')

    [mis_angle, mis_axis, mis_info] = gtDisorientation(R1.', R2.', app.symm, ...
        'latticepar', latticepar, 'sort', 'descend', varargin{:});
    if mis_angle > 90 
         mis_angle = 180 - mis_angle;
    end 
   
    sigmaOr = [];
    if ~isempty(mis_axis)
        % determine sigma type
        test_angle = abs(mis_angle - sigmas(:, 2)) < sigmas(:, 6);
        test_axis  = abs(acosd(dot(repmat(mis_axis,size(sigmas,1),1), sigmas(:, 3:5), 2))) < sigmas(:, 6);
        sigmaOr = sigmas(find(test_angle | test_axis), :);
        sigma = sigmas(find(test_angle & test_axis), 1);
    end

elseif strcmp(crystal_system, 'hexagonal')

    % In the case of hexagonal crystals
    [mis_angle, mis_axis, mis_info] = gtDisorientation(R1.', R2.', app.symm, ...
        'latticepar', latticepar, 'convention', app.convention, varargin{:});
    if mis_angle > 90 
         mis_angle = 180 - mis_angle;
    end

    sigmaOr = [];
    if ~isempty(mis_axis)
        % determine sigma type
        test_angle = abs(mis_angle - sigmas(:, 2)) < sigmas(:, 7);
        test_axis  = abs(acosd(dot(repmat(mis_axis,size(sigmas,1),1), sigmas(:, 8:10), 2))) < sigmas(:, 7);
        sigmaOr = sigmas(find(test_angle | test_axis), :);
        sigma   = sigmas(find(test_angle & test_axis), :);
    end
else
    disp('Crystal system is not currently supported... Quitting')
    return
end

% add low angle (merge grains) criteria
if (mis_angle < app.merge_angle)
    sigma = -1;
end

info.R1           = R1;
info.R2           = R2;
info.mis_angle    = mis_angle;
info.mis_axis     = mis_axis;
info.sigmaOr      = sigmaOr;
info.sigmaAnd     = sigma;
info.sigmas       = app.sigmas;
info.merge_angle  = app.merge_angle;
info.convention   = app.convention;
info = gtAddMatFile(info, mis_info, true, true);

end % end function twinTest
