/*
 * gtMathsRotationTensor_c.cpp
 *
 *  Created on: Jul 7, 2014
 *      Author: vigano
 */

#include "internal_cell_defs.h"

#include <algorithm>

const char * errorID = "C_FUN:gtMathsRotationTensor_c:wrong_argument";

#define COMPUTE_PROD(index) \
  base_tensor[index] = c_const[index] + c_cos[index] * w_cos + c_sin[index] * w_sin

#define BODY_COMPUTE(shift) \
  {\
    const double_t w = ws[w_index + shift] / 180.0 * M_PI;\
    const double_t w_cos = cos(w);\
    const double_t w_sin = sin(w);\
    Type * base_tensor = tensors + 9 * (w_index + shift);\
    \
    COMPUTE_PROD(0);\
    COMPUTE_PROD(1);\
    COMPUTE_PROD(2);\
    COMPUTE_PROD(3);\
    COMPUTE_PROD(4);\
    COMPUTE_PROD(5);\
    COMPUTE_PROD(6);\
    COMPUTE_PROD(7);\
    COMPUTE_PROD(8);\
  }

template<typename Type>
mwSize compute_tensors(mxArray * const rot_tensors, const mxArray * const omegas,
    const mxArray * const rot_const, const mxArray * const rot_cos,
    const mxArray * const rot_sin, const mwSize chunck_size = 24)
{
  Type * const tensors = (Type *) mxGetData(rot_tensors);
  Type * const ws = (Type *) mxGetData(omegas);

  const mwSize num_omegas = mxGetNumberOfElements(omegas);

  const double_t * const c_const  = (const double_t *) mxGetData(rot_const);
  const double_t * const c_cos  = (const double_t *) mxGetData(rot_cos);
  const double_t * const c_sin  = (const double_t *) mxGetData(rot_sin);

  const mwSize num_threads = std::max(std::min(num_omegas / chunck_size, omp_get_max_threads()/4*3), 1);
  const mwSize num_unroll_omegas = num_omegas - (num_omegas % chunck_size);

#pragma omp parallel num_threads(num_threads)
  {
#pragma omp for nowait
  for (mwIndex w_index = 0; w_index < num_unroll_omegas; w_index += chunck_size)
  {
    for (mwIndex c_index = 0; c_index < chunck_size; c_index++)
    {
      BODY_COMPUTE(c_index);
    }
  }
#pragma omp for nowait
  for (mwIndex w_index = num_unroll_omegas; w_index < num_omegas; w_index++)
  {
    BODY_COMPUTE(0);
  }
  }
  return num_threads;
}

void mexFunction( int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[] )
{
  if (nrhs != 2)
  {
    mexErrMsgIdAndTxt(errorID,
        "Not enough arguments! you need to provide the omegas array, and the rotation components");
    return;
  }
  if (!mxIsNumeric(prhs[0]))
  {
    mexErrMsgIdAndTxt(errorID, "First argument should be an array");
    return;
  }
  if (!mxIsStruct(prhs[1]))
  {
    mexErrMsgIdAndTxt(errorID, "Second argument should be a structure");
    return;
  }
  const mxArray * const rot_const = mxGetField(prhs[1], 0, "const");
  const mxArray * const rot_cos = mxGetField(prhs[1], 0, "cos");
  const mxArray * const rot_sin = mxGetField(prhs[1], 0, "sin");
  if (!rot_const)
  {
    mexErrMsgIdAndTxt(errorID, "Missing field: 'const'");
    return;
  }
  else if (mxGetNumberOfElements(rot_const) != 9)
  {
    mexErrMsgIdAndTxt(errorID, "Field 'const' should be a 3x3 double precision matrix");
    return;
  }
  else if (mxGetClassID(rot_const) != mxDOUBLE_CLASS)
  {
    mexErrMsgIdAndTxt(errorID, "Field 'const' should be a 3x3 double precision matrix");
    return;
  }

  if (!rot_cos)
  {
    mexErrMsgIdAndTxt(errorID, "Missing field: 'cos'");
    return;
  }
  else if (mxGetNumberOfElements(rot_cos) != 9)
  {
    mexErrMsgIdAndTxt(errorID, "Field 'cos' should be a 3x3 double precision matrix");
    return;
  }
  else if (mxGetClassID(rot_cos) != mxDOUBLE_CLASS)
  {
    mexErrMsgIdAndTxt(errorID, "Field 'cos' should be a 3x3 double precision matrix");
    return;
  }

  if (!rot_sin)
  {
    mexErrMsgIdAndTxt(errorID, "Missing field: 'sin'");
    return;
  }
  else if (mxGetNumberOfElements(rot_sin) != 9)
  {
    mexErrMsgIdAndTxt(errorID, "Field 'sin' should be a 3x3 double precision matrix");
    return;
  }
  else if (mxGetClassID(rot_sin) != mxDOUBLE_CLASS)
  {
    mexErrMsgIdAndTxt(errorID, "Field 'sin' should be a 3x3 double precision matrix");
    return;
  }

  const mwSize num_omegas = mxGetNumberOfElements(prhs[0]);

  // Let's allocate without initialization
  const mxClassID type = mxGetClassID(prhs[0]);
  const mwSize zero_dims[2] = {0, 0};
  const mwSize dims[3] = {3, 3, num_omegas};
  mxArray * const rot_tensors = mxCreateNumericArray(2, zero_dims, type, mxREAL);
  plhs[0] = rot_tensors;

  switch(type)
  {
    case mxDOUBLE_CLASS:
    {
      mxSetDimensions(rot_tensors, dims, 3);
      mxSetData(rot_tensors, mxMalloc(9 * num_omegas * sizeof(double_t)));
      const mwSize num_threads = compute_tensors<double_t>(rot_tensors, prhs[0], rot_const, rot_cos, rot_sin);
      plhs[1] = mxCreateDoubleScalar(num_threads);
      break;
    }
    case mxSINGLE_CLASS:
    {
      mxSetDimensions(rot_tensors, dims, 3);
      mxSetData(rot_tensors, mxMalloc(9 * num_omegas * sizeof(float_t)));
      const mwSize num_threads = compute_tensors<float_t>(rot_tensors, prhs[0], rot_const, rot_cos, rot_sin);
      plhs[1] = mxCreateDoubleScalar(num_threads);
      break;
    }
    default:
    {
      mxDestroyArray(rot_tensors);
      mexErrMsgIdAndTxt(errorID, "Only single and double precision floating points allowed");
    }
  }
}

