classdef Gt6DBlobReconstructor < Gt6DVolumeToBlobProjector
    properties
        % Data
        blobs;
        moderated_blobs;

        % Solution
        currentSolution = {};
        currentResidual = {};

        % Variables to plot
        normInitialResidual;
        normResiduals;
        maxVals;
        minVals;

        verbose = false;
    end

    methods (Access = public)
        function self = Gt6DBlobReconstructor(volumes, blobs, varargin)
            fprintf('Initializing BLOB Recontruction:\n - Variables..');
            c = tic();
            blobs_depths = cellfun(@(x)size(x, 2), blobs);
            proj_size = Gt6DBlobReconstructor.getProjSize(blobs);

            self = self@Gt6DVolumeToBlobProjector(volumes, proj_size, blobs_depths, varargin{:});

            self.statistics.add_task('sirt_apply_sum', 'SIRT Apply time');
            self.statistics.add_task('sirt_soft_thresh', 'SIRT Soft-threholding time');
            self.statistics.add_task('sirt_fista_reg', 'SIRT Fista reg time');
            self.statistics.add_task('sirt_non_neg_sum', 'SIRT Non-Negative time');
            self.statistics.add_task('sirt_norm_residual', 'SIRT resiual norm time');

            self.blobs = blobs;
            self.currentSolution = volumes;

            self.initProjectionGeometry();
            self.initProjectionGeometrySS();

            fprintf('\b\b (%2.1f s)\n - Weights:\n', toc(c));
            c = tic();
            self.initializeWeights();
            fprintf('   Tot (%3.1f s)), Done.\n', toc(c));
        end

        function solution = getCurrentSolution(self)
            % solution: x = P^(1/2) u
            solution = internal_cell_prod_copy(self.currentSolution, self.backwardWeights);
        end

        function projection = getProjectionOfCurrentSolution(self)
            projection = self.projectVolume(self.getCurrentSolution());
        end

        function reInit(self, blobs, volumes)
            self.blobs = blobs;
            if (exist('volumes', 'var'))
                self.currentSolution = volumes;
            else
                self.currentSolution = gtMathsGetSameSizeZeros(self.currentSolution);
            end

            fprintf('Recomputing Weights..');
            c = tic();
            self.initializeWeights();
            fprintf('\b\b: (%3.1f s) Done.\n', toc(c));
        end

        function sirt(self, numIters, lambda)
            self.initializeVariables(numIters);
            t = 1;

            c = tic();
            % Initial setup of variables needed for the first iteration
            fprintf('Computing initial residual..')
            self.currentResidual = self.computeResidual(self.currentSolution);
            self.normInitialResidual = gtMathsNorm_l2(self.currentResidual);
            fprintf('\b\b (%3.1f s), Done.\n', toc(c))

            if (self.verbose)
                fprintf('Reconstruction: ')
                f1 = figure('name', sprintf('SIRT: Lambda(%e), Iterations(%d)', ...
                                            lambda, numIters));
                ax1 = subplot(2, 2, 1, 'Parent', f1);
                ax2 = subplot(2, 2, 2, 'Parent', f1);
                ax3 = subplot(2, 2, 3, 'Parent', f1);
                drawnow;
            end

            % Iterations loop
            for ii = 1:numIters
                numchars = fprintf('%03d/%03d (in %f s)', ii-1, numIters, toc(c));

                self.statistics.tic('sirt_apply_sum');
                % Application of correction
                nextSolution = internal_cell_sub_copy(self.currentSolution, self.currentResidual);
                self.statistics.toc('sirt_apply_sum');

                self.statistics.tic('sirt_soft_thresh');
                % FISTA
                %   For theoretical reasons, step is always 1, so threshold
                %   is lambda * 1
                nextSolution = Gt6DBlobReconstructor.softThreshold(nextSolution, lambda);
                self.statistics.toc('sirt_soft_thresh');

                self.statistics.tic('sirt_non_neg_sum');
                % Non-negative
%                 nextSolution = gtMathsNonNegative(nextSolution);
                nextSolution = internal_cell_non_neg_assign(nextSolution);
                self.statistics.toc('sirt_non_neg_sum');

                self.statistics.tic('sirt_fista_reg');
                [nextEnhancedSolution, t] ...
                    = Gt6DBlobReconstructor.fistaReg(nextSolution, self.currentSolution, t);
                self.statistics.toc('sirt_fista_reg');

                self.currentResidual = self.computeResidual(nextEnhancedSolution);
                self.currentSolution = nextSolution;

                % Indication parameters
                self.statistics.tic('sirt_norm_residual');
                self.normResiduals(ii) ...
                    = sqrt(internal_cell_square_norm(self.currentResidual)) / self.normInitialResidual;
                self.statistics.toc('sirt_norm_residual');

                self.maxVals(ii) = mean(cellfun(@(x)max(max(max(x))), nextSolution));
                self.minVals(ii) = mean(cellfun(@(x)min(min(min(x))), nextSolution));

                if (self.verbose)
                    self.visualize(ax1, ax2, ax3, ii);
                    drawnow;
                end

                fprintf('%s', repmat(sprintf('\b'), 1, numchars))
            end
            fprintf('(%04d) Done in %f seconds.\n', numIters, toc(c) )

            self.printStats();
        end

        function cp(self, numIters, lambda)
            self.initializeVariables(numIters);
            sample_rate = 5;

            [sigma1, sigma1_1, tau] = self.initializeCPWeights();
            theta = 1;

            p = gtMathsGetSameSizeZeros(self.blobs);
            q = gtMathsGetSameSizeZeros(self.currentSolution);

            nextEnhancedSolution = self.currentSolution;

            fprintf('Reconstruction: ')

            c = tic();
            switch (numel(numIters))
                case 1
                    iters = 1:numIters;
                case 2
                    iters = numIters(1):numIters(2);
                otherwise
                    iters = numIters;
            end
            tot_iters = numel(iters);
            for ii = iters
                numchars = fprintf('%03d/%03d (in %f s)', ii-min(iters), tot_iters, toc(c));

                % Computing update of dual detector
                l = self.projectVolume(nextEnhancedSolution);
                l = internal_cell_sub_assign(l, self.blobs);
                l = internal_cell_prod_assign(l, sigma1);
                p = internal_cell_sum_assign(p, l);
                p = internal_cell_prod_assign(p, sigma1_1);

                % Computing update dual l1
                q = internal_cell_sum_assign(q, nextEnhancedSolution);
                for n = 1:numel(q)
                    q{n} = lambda * q{n} ./ max(lambda, abs(q{n}));
                end

                % Computing update primal
                nextSolution = self.backprojectVolume(p);
                nextSolution = internal_cell_sum_assign(nextSolution, q);
                nextSolution = internal_cell_prod_assign(nextSolution, tau);
                nextSolution = internal_cell_sum_assign(nextSolution, self.currentSolution);

                nextSolution = internal_cell_non_neg_assign(nextSolution);

                nextEnhancedSolution = internal_cell_sum_FISTA_scaled_copy(nextSolution, self.currentSolution, theta);

                self.currentSolution = nextSolution;

                fprintf('%s', repmat(sprintf('\b'), 1, numchars))

                if (self.verbose && (mod(ii, sample_rate) == 0))
                    l = self.projectVolume(self.currentSolution);
                    l = internal_cell_sub_assign(l, self.blobs);
                    self.normResiduals(ii / sample_rate) = ...
                        internal_cell_square_norm(l) / 2 ...
                        + lambda * gtMathsNorm_l1(self.currentSolution) ...
                        + internal_cell_square_norm(p) / 2 ...
                        + internal_cell_dot_product(p, self.blobs);

                    l = gtMathsSumCellVolumes(self.currentSolution);
                    ctr_slice = round(size(l, 3) / 2);
                    figure, imagesc(l(:, :, ctr_slice)), ...
                        title(sprintf('Slice: %d, Iteration: %d', ctr_slice, ii))
                end
                drawnow();
            end
            fprintf('(%04d) Done in %f seconds.\n', tot_iters, toc(c) )

            self.printStats();

            if (self.verbose)
                figure;
                subplot(1, 2, 1), plot(self.normResiduals(1:floor(max(iters) / sample_rate)));
                subplot(1, 2, 2), semilogy(abs(self.normResiduals));
            end
        end

        function cpTV(self, numIters, lambda)
            self.initializeVariables(numIters);
            sample_rate = 5;

            [sigma1, sigma1_1, tau] = self.initializeCPWeightsTV();
            theta = 1;

            p = gtMathsGetSameSizeZeros(self.blobs);
            q = {zeros(size(self.currentSolution{1}), class(self.currentSolution{1}))};
            q(2:3) = q(1);

            nextEnhancedSolution = self.currentSolution;

            fprintf('Reconstruction: ')

            c = tic();
            switch (numel(numIters))
                case 1
                    iters = 1:numIters;
                case 2
                    iters = numIters(1):numIters(2);
                otherwise
                    iters = numIters;
            end

            tot_iters = numel(iters);

            if (self.verbose)
                result_slice = zeros(size(self.currentSolution{1}, 1), ...
                    size(self.currentSolution{1}, 2), floor(tot_iters/sample_rate));
                vol_viewer = GtVolView(result_slice);
            end

            for ii = iters
                numchars = fprintf('%03d/%03d (in %f s)', ii-min(iters), tot_iters, toc(c));

                % Computing update of dual detector
                l = self.projectVolume(nextEnhancedSolution);
                l = internal_cell_sub_assign(l, self.blobs);
                l = internal_cell_prod_assign(l, sigma1);
                p = internal_cell_sum_assign(p, l);
                p = internal_cell_prod_assign(p, sigma1_1);

                % Computing update dual TV
                summedEnhancedSolution = gtMathsSumCellVolumes(nextEnhancedSolution) / numel(nextEnhancedSolution);
                [sES{1}, sES{2}, sES{3}] = gradient(summedEnhancedSolution);
                q = internal_cell_sum_assign(q, sES);
                for n = 1:numel(q)
                    q{n} = q{n} ./ max(1, abs(q{n}));
                end
                summedEnhancedSolution = divergence(q{:});
                summedEnhancedSolution = single(summedEnhancedSolution);

                % Computing update primal
                nextSolution = self.backprojectVolume(p);
                nextSolution = internal_cell_sub_assign(nextSolution, cellfun(@(x){summedEnhancedSolution}, nextSolution));
                nextSolution = internal_cell_prod_assign(nextSolution, tau);
                nextSolution = internal_cell_sum_assign(nextSolution, self.currentSolution);

                nextSolution = internal_cell_non_neg_assign(nextSolution);

                nextEnhancedSolution = internal_cell_sum_FISTA_scaled_copy(nextSolution, self.currentSolution, theta);

                self.currentSolution = nextSolution;

                fprintf('%s', repmat(sprintf('\b'), 1, numchars))

                if (self.verbose && (mod(ii, sample_rate) == 0))
                    l = gtMathsSumCellVolumes(self.currentSolution) / numel(nextEnhancedSolution);
                    [sES{1}, sES{2}, sES{3}] = gradient(l);
                    self.normResiduals(ii / sample_rate) = ...
                        + gtMathsNorm_l1(sES) ...
                        + internal_cell_dot_product(p, self.blobs);

                    half_slice = round(size(l, 3) / 2);
                    result_slice(:, :, ii / sample_rate) = l(:, :, half_slice);
                    vol_viewer.changeVol(result_slice);
                    vol_viewer.changeSlice(ii / sample_rate);
                end
                drawnow();
            end
            fprintf('(%04d) Done in %f seconds.\n', tot_iters, toc(c) )

            self.printStats();

            if (self.verbose)
                figure;
                subplot(1, 2, 1), plot(self.normResiduals(1:floor(max(iters) / sample_rate)));
                subplot(1, 2, 2), semilogy(abs(self.normResiduals));
            end
        end

        function cpTVl1(self, numIters, lambda)
            self.initializeVariables(numIters);
            sample_rate = 5;

            [sigma1, sigma1_1, tau] = self.initializeCPWeightsTV();
            theta = 1;

            p = gtMathsGetSameSizeZeros(self.blobs);
            q1 = {zeros(size(self.currentSolution{1}), class(self.currentSolution{1}))};
            q1(2:3) = q1(1);
            q2 = gtMathsGetSameSizeZeros(self.currentSolution);

            nextEnhancedSolution = self.currentSolution;

            fprintf('Reconstruction: ')

            c = tic();
            switch (numel(numIters))
                case 1
                    iters = 1:numIters;
                case 2
                    iters = numIters(1):numIters(2);
                otherwise
                    iters = numIters;
            end

            tot_iters = numel(iters);

            if (self.verbose)
                result_slice = zeros(size(self.currentSolution{1}, 1), ...
                    size(self.currentSolution{1}, 2), floor(tot_iters/sample_rate));
                vol_viewer = GtVolView(result_slice);
            end

            for ii = iters
                numchars = fprintf('%03d/%03d (in %f s)', ii-min(iters), tot_iters, toc(c));

                % Computing update of dual detector
                l = self.projectVolume(nextEnhancedSolution);
                l = internal_cell_sub_assign(l, self.blobs);
                l = internal_cell_prod_assign(l, sigma1);
                p = internal_cell_sum_assign(p, l);
                p = internal_cell_prod_assign(p, sigma1_1);

                % Computing update dual TV
                summedEnhancedSolution = gtMathsSumCellVolumes(nextEnhancedSolution) / numel(nextEnhancedSolution);
                [sES{1}, sES{2}, sES{3}] = gradient(summedEnhancedSolution);
                q1 = internal_cell_sum_assign(q1, sES);
                for n = 1:numel(q1)
                    q1{n} = q1{n} ./ max(1, abs(q1{n}));
                end
                summedEnhancedSolution = divergence(q1{:});
                summedEnhancedSolution = single(summedEnhancedSolution);

                % Computing update dual l1
                q2 = internal_cell_sum_assign(q2, nextEnhancedSolution);
                for n = 1:numel(q2)
                    q2{n} = lambda * q2{n} ./ max(lambda, abs(q2{n}));
                end

                % Computing update primal
                nextSolution = self.backprojectVolume(p);
                nextSolution = internal_cell_sub_assign(nextSolution, cellfun(@(x){summedEnhancedSolution}, nextSolution));
                nextSolution = internal_cell_sum_assign(nextSolution, q2);
                nextSolution = internal_cell_prod_assign(nextSolution, tau);
                nextSolution = internal_cell_sum_assign(nextSolution, self.currentSolution);

                nextSolution = internal_cell_non_neg_assign(nextSolution);

                nextEnhancedSolution = internal_cell_sum_FISTA_scaled_copy(nextSolution, self.currentSolution, theta);

                self.currentSolution = nextSolution;

                fprintf('%s', repmat(sprintf('\b'), 1, numchars))

                if (self.verbose && (mod(ii, sample_rate) == 0))
                    l = gtMathsSumCellVolumes(self.currentSolution) / numel(nextEnhancedSolution);
                    [sES{1}, sES{2}, sES{3}] = gradient(l);
                    self.normResiduals(ii / sample_rate) = ...
                        + gtMathsNorm_l1(sES) ...
                        + internal_cell_dot_product(p, self.blobs);

                    half_slice = round(size(l, 3) / 2);
                    result_slice(:, :, ii / sample_rate) = l(:, :, half_slice);
                    vol_viewer.changeVol(result_slice);
                    vol_viewer.changeSlice(ii / sample_rate);
                end
                drawnow();
            end
            fprintf('(%04d) Done in %f seconds.\n', tot_iters, toc(c) )

            self.printStats();

            if (self.verbose)
                figure;
                subplot(1, 2, 1), plot(self.normResiduals(1:floor(max(iters) / sample_rate)));
                subplot(1, 2, 2), semilogy(abs(self.normResiduals));
            end
        end
    end

    methods (Access = protected)
        function initializeWeights(self)
            c = tic();
            fprintf('   + Projecting ones vols..');
            self.forwardWeights = self.getRowsSum();
            fprintf('\b\b (%3.1f s), Fixing it..', toc(c));
            c = tic();
            numForward = length(self.forwardWeights);
            for ii = 1:numForward
                self.forwardWeights{ii} = 1 ./ sqrt(self.forwardWeights{ii} + (self.forwardWeights{ii} == 0));
            end
            fprintf('\b\b (%2.1f s)\n', toc(c));

            fprintf('   + Backrojecting ones sinos..');
            self.backwardWeights = self.getColumnsSum();
            fprintf('\b\b (%3.1f s), Fixing it..', toc(c));
            c = tic();
            numBackward = length(self.backwardWeights);
            for ii = 1:numBackward
                self.backwardWeights{ii} = 1 ./ sqrt(self.backwardWeights{ii} + (self.backwardWeights{ii} == 0));
            end
            fprintf('\b\b (%2.1f s)\n', toc(c));

            self.moderated_blobs = internal_cell_prod_copy(self.blobs, self.forwardWeights);
        end

        function [sigma1, sigma1_1, tau] = initializeCPWeights(self)
            sigma1 = self.forwardWeights;
            for n = 1:numel(sigma1)
                sigma1{n} = sigma1{n} .^ 2;
            end

            sigma1_1 = sigma1;
            for n = 1:numel(sigma1_1)
                sigma1_1{n} = 1 ./ (1 + sigma1_1{n});
            end

            tau = self.backwardWeights;
            for n = 1:numel(tau)
                tau{n} = - 1 ./ (1 ./ (tau{n} .^ 2) + 1);
            end
        end

        function [sigma1, sigma1_1, tau] = initializeCPWeightsTV(self)
            sigma1 = self.forwardWeights;
            for n = 1:numel(sigma1)
                sigma1{n} = sigma1{n} .^ 2;
            end

            sigma1_1 = sigma1;
            for n = 1:numel(sigma1_1)
                sigma1_1{n} = 1 ./ (1 + sigma1_1{n});
            end

            tau = self.backwardWeights;
            for n = 1:numel(tau)
                tau{n} = - 1 ./ (1 ./ (tau{n} .^ 2) + 1 / numel(tau));
            end
        end

        function initializeVariables(self, numIters)
            switch (numel(numIters))
                case 1
                    self.normResiduals = zeros(numIters, 1);
                    self.maxVals = zeros(numIters, 1);
                    self.minVals = zeros(numIters, 1);
                case 2
                    self.normResiduals(numIters(1):numIters(2)) = 0;
                    self.maxVals(numIters(1):numIters(2)) = 0;
                    self.minVals(numIters(1):numIters(2)) = 0;
                otherwise
                    self.normResiduals(numIters) = 0;
                    self.maxVals(numIters) = 0;
                    self.minVals(numIters) = 0;
            end
        end

        function residual = computeResidual(self, volumes)
            % l = W * u
            % Where: W = Q^(1/2) A P^(1/2) and u = P^(-1/2) x
            l = self.project(volumes);

            % l = W u - f
            % Where: f = Q^(1/2) b
            l = internal_cell_sub_assign(l, self.moderated_blobs);

            % residual = W^T (f - W u)
            residual = self.backproject(l);
        end

        function visualize(self, ax1, ax2, ax3, ii)
            if ((mod(ii, 10) == 0) || (ii < 25) ...
                    || ((ii < 50) && (mod(ii, 2) == 0)) ...
                    || ((ii < 80) && (mod(ii, 5) == 0)))
                semilogy(abs(self.normResiduals(1:ii)), 'Parent', ax1);
                plot(self.maxVals(1:ii), 'r', 'Parent', ax3);
                hold(ax3, 'on')
                plot(self.minVals(1:ii), 'g', 'Parent', ax3);
                hold(ax3, 'off')
            end
            if (mod(ii, 3) == 0)
%                 imagesc(self.currentSolution{1}(:,:,16), 'Parent', ax2)
                imagesc(self.currentSolution{1}(:,:,4), 'Parent', ax2)
%                 imagesc(self.currentSolution{1}(:,:,50), 'Parent', ax2)
            end
        end
    end

    methods (Static, Access = public)
        function proj_size = getProjSize(blobs)
            proj_size = [size(blobs{1}, 1), size(blobs{1}, 3)];
            for n = 2:numel(blobs)
                if (any(proj_size ~= [size(blobs{n}, 1), size(blobs{n}, 3)]))
                    error('Gt6DBlobReconstructor:wrong_argument', ...
                        'Blob: %d is malformed!', n)
                end
            end
        end

        function nextSolution = softThreshold(nextSolution, thresh)
            for ii = 1:numel(nextSolution)
                nextSolution{ii} = gtSoftThreshold2(nextSolution{ii}, thresh);
            end
%             waveletLevel = 1;
%             nextSolution = dwt_haar_soft_threshold(nextSolution, waveletLevel, thresh);
        end

        function t_1 = fistaTplus1(t)
            t_1 = (1 + sqrt(1+ 4 * t * t)) / 2;
        end

        function w = fistaWeight(t, t_1)
            w = (t - 1) / t_1;
        end

        function [nextEnhancedSolution, t] = fistaReg(nextSolution, currentSolution, t)
            t_1 = Gt6DBlobReconstructor.fistaTplus1(t);
            w = Gt6DBlobReconstructor.fistaWeight(t, t_1);
            t = t_1;

            nextEnhancedSolution = internal_cell_sum_FISTA_scaled_copy(nextSolution, currentSolution, w);
        end
    end
end