classdef Gt6DSpotReconstructor < Gt6DVolumeToSpotProjector
    properties
        spots;
        moderated_spots;

        currentSolution = {};

        % CG specific variables
        currentResidual;
        currentCondResidual;

        currentCorrection;
        currentProjCorrection;
        squareProjCorr;

        % FISTA params
        lambda = 1e-3;
        weightFISTA = 1;

        % NonNegative param
        non_negative = true;

        % Info parameters
        dotResiduals;
        detectorResiduals;
        maxVals;
        minVals;
        thresholds;
        alpha;
        beta;

        % Used to display stuff
        verbose = true;
    end

    methods (Access = public)
        function self = Gt6DSpotReconstructor(initialVols, spots, varargin)
            fprintf('Initializing SPOT Recontruction:\n - Variables..');
            c = tic();

            detectorSize = size(spots);
            self = self@Gt6DVolumeToSpotProjector(initialVols, ...
                [detectorSize(1) detectorSize(3)], varargin{:});

            self.statistics.add_task('sirt_apply_sum', 'SIRT Apply time');
            self.statistics.add_task('sirt_fista_sum', 'SIRT Fista time');
            self.statistics.add_task('sirt_bbox_sum', 'SIRT Bbox cut time');
            self.statistics.add_task('sirt_norm_residual', 'SIRT resiual norm time');

            self.initProjectionGeometry();
            self.initProjectionGeometrySS();

            self.spots = spots;

            self.currentSolution = initialVols;

            fprintf('\b\b (%2.1f s)\n - Weights:\n', toc(c));
            c = tic();
            self.initializeWeights();
            fprintf('   Tot (%3.1f s)), Done.\n', toc(c));
        end

        function cg(self, numIters, lambda)
            if (exist('lambda', 'var'))
                self.lambda = lambda;
            end
            fprintf('Initialization: ')
            self.initializeVariables(numIters);

            if (self.verbose)
                fprintf('(lambda: %e) Done.\nReconstruction: ', self.lambda)
                f1 = figure('name', sprintf('CG: Lambda(%e), Iterations(%d)', ...
                                            self.lambda, numIters));
                ax1 = subplot(2, 2, 1, 'Parent', f1);
                ax2 = subplot(2, 2, 2, 'Parent', f1);
                ax3 = subplot(2, 2, 3, 'Parent', f1);
                drawnow;
            end

            t = tic();
            % Initial setup of variables needed for the first iteration
            self.initializeCG();
            % Iterations loop
            for ii = 1:numIters
                fprintf('%05d/%05d', ii, numIters)

                try
                    self.makeCGIteration(ii, numIters);
                catch mexc
                    gtPrintException(mexc, 'Cleaning Gradient Memory');
                    self.initializeCG();
                end

                if (self.verbose)
                    self.visualize(ax1, ax2, ax3, ii);
                    drawnow;
                end

                fprintf('\b\b\b\b\b\b\b\b\b\b\b')
            end
            fprintf('(%04d) Done in %f seconds.\n', numIters, toc(t) )
            self.printStats();
        end

        function sirt(self, numIters, lambda)
            if (exist('lambda', 'var'))
                self.lambda = lambda;
            end
            fprintf('Initialization: ')
            self.initializeVariables(numIters);

            if (self.verbose)
                fprintf('(lambda: %e) Done.\nReconstruction: ', self.lambda)
                f1 = figure('name', sprintf('SIRT: Lambda(%e), Iterations(%d)', ...
                                            self.lambda, numIters));
                ax1 = subplot(2, 2, 1, 'Parent', f1);
                ax2 = subplot(2, 2, 2, 'Parent', f1);
                ax3 = subplot(2, 2, 3, 'Parent', f1);
                drawnow;
            end

            t = tic();
            elapsedTime = toc(t);
            % Initial setup of variables needed for the first iteration
            self.currentResidual = self.computeResiduals();
            % Iterations loop
            for ii = 1:numIters
                numchars = fprintf('%03d/%03d (in %f s)', ii, numIters, elapsedTime);

                self.makeSIRTIteration(ii);

                if (self.verbose)
                    self.visualize(ax1, ax2, ax3, ii);
                    drawnow;
                end

                fprintf('%s', repmat(sprintf('\b'), 1, numchars))
                elapsedTime = toc(t);
            end
            fprintf('(%04d) Done in %f seconds.\n', numIters, toc(t) )
            self.printStats();
        end

        function solution = getCurrentSolution(self)
            % solution: x = P^(1/2) u
            solution = internal_cell_div_copy(self.currentSolution, self.backwardWeights);
        end

        function finalVols = getMergeMultipleJunction(self, threshold)
            indices = cellfun(@(x){x > threshold}, self.currentSolution);
            finalVols = zeros(size(self.currentSolution{1}));

            numVols = length(self.currentSolution);
            for ii = 1:numVols
                first = true;
                for jj = 1:numVols
                    if ii ~= jj
                        if (first)
                            indices{ii} = indices{ii} & (self.currentSolution{ii} >= self.currentSolution{jj});
                        else
                            indices{ii} = indices{ii} & (self.currentSolution{ii} > self.currentSolution{jj});
                        end
                    end
                    first = false;
                end
                finalVols = finalVols + indices{ii} * ii;
            end
        end

        function finalVols = getMergeTwins(self, threshold)
            indices1 = (self.currentSolution{1} > self.currentSolution{2}) ...
                & (self.currentSolution{1} > threshold * max(self.currentSolution{1}(:)));
            indices2 = (self.currentSolution{2} >= self.currentSolution{1}) ...
                & (self.currentSolution{2} > threshold * max(self.currentSolution{2}(:)));
            finalVols = indices1 + 2 * indices2;
        end

        function cp(self, numIters, lambda)
            self.initializeVariables(numIters);
            sample_rate = 5;

            [sigma1, sigma1_1, tau] = self.initializeCPWeights();
            theta = 1;

            p = gtMathsGetSameSizeZeros(self.spots);
            q = gtMathsGetSameSizeZeros(self.currentSolution);

            nextEnhancedSolution = self.currentSolution;

            fprintf('Reconstruction: ')

            c = tic();
            switch (numel(numIters))
                case 1
                    iters = 1:numIters;
                case 2
                    iters = numIters(1):numIters(2);
                otherwise
                    iters = numIters;
            end
            tot_iters = numel(iters);
            for ii = iters
                numchars = fprintf('%03d/%03d (in %f s)', ii-min(iters), tot_iters, toc(c));

                l = self.projectVolume(nextEnhancedSolution);
                p = (p + (l - self.spots) .* sigma1) .* sigma1_1;

                q = internal_cell_sum_assign(q, nextEnhancedSolution);
                for n = 1:numel(q)
                    q{n} = lambda * q{n} ./ max(lambda, abs(q{n}));
                end

                nextSolution = self.backprojectVolume(p);
                nextSolution = internal_cell_sum_assign(nextSolution, q);
                nextSolution = internal_cell_prod_assign(nextSolution, tau);
                nextSolution = internal_cell_sum_assign(nextSolution, self.currentSolution);

                nextSolution = internal_cell_non_neg_assign(nextSolution);

                nextEnhancedSolution = internal_cell_sum_FISTA_scaled_copy(nextSolution, self.currentSolution, theta);

                self.currentSolution = nextSolution;

                fprintf('%s', repmat(sprintf('\b'), 1, numchars))

                if (mod(ii, sample_rate) == 0)
                    l = self.projectVolume(self.currentSolution);
                    l = l - self.spots;
                    self.dotResiduals(ii / sample_rate) = ...
                        gtMathsNorm_l2(l) / 2 ...
                        + lambda * gtMathsNorm_l1(self.currentSolution) ...
                        + gtMathsNorm_l2(p) / 2 ...
                        + gtMathsDotProduct(p, self.spots);
                end
            end
            fprintf('(%04d) Done in %f seconds.\n', tot_iters, toc(c) )

            self.printStats();

            figure;
            subplot(1, 2, 1), plot(self.dotResiduals(1:floor(max(iters) / sample_rate)));
            subplot(1, 2, 2), semilogy(abs(self.dotResiduals));
        end
    end

    methods (Access = protected)
        function initializeWeights(self)
            c = tic();
            fprintf('   + Projecting ones vols..');
            self.forwardWeights = self.getRowsSum();
            fprintf('\b\b (%3.1f s), Fixing it..', toc(c));
            c = tic();
            self.forwardWeights = 1 ./ sqrt(self.forwardWeights + (self.forwardWeights == 0));
            fprintf('\b\b (%2.1f s)\n', toc(c));

            fprintf('   + Backrojecting ones sinos..');
            self.backwardWeights = self.getColumnsSum();
            fprintf('\b\b (%3.1f s), Fixing it..', toc(c));
            c = tic();
            numGeoms = length(self.backwardWeights);
            for ii = 1:numGeoms
                self.backwardWeights{ii} = 1 ./ sqrt(self.backwardWeights{ii} + (self.backwardWeights{ii} == 0));
            end
            fprintf('\b\b (%2.1f s)\n', toc(c));

            self.moderated_spots = self.spots .* self.forwardWeights;
        end

        function initializeVariables(self, numIters)
            self.dotResiduals = zeros(numIters, 1);
            self.detectorResiduals = zeros(numIters, 1);
            self.squareProjCorr = zeros(numIters, 1);
            self.maxVals = zeros(numIters, 1);
            self.minVals = zeros(numIters, 1);
            self.alpha = zeros(numIters, 1);
            self.beta = zeros(numIters, 1);
        end

        function w = getFISTAWeight(self, t_1)
            w = (self.weightFISTA - 1) / t_1;
        end

        function initializeCG(self)
            self.currentResidual = self.computeResiduals(self.currentSolution);
            self.currentCondResidual = self.currentResidual;

            self.currentCorrection = internal_cell_prod_scalar_copy(self.currentResidual, -1);

            self.currentProjCorrection = self.project(self.currentCorrection);
        end

        function [sigma1, sigma1_1, tau] = initializeCPWeights(self)
            sigma1 = self.forwardWeights .^ 2;
            sigma1_1 = 1 ./ (1 + sigma1);

            tau = self.backwardWeights;
            for n = 1:numel(tau)
                tau{n} = - 1 ./ (1 ./ (tau{n} .^ 2) + 1);
            end
        end

        function [residual, detectorResidual] = computeResiduals(self, newVolumes)
            if (~exist('newVolumes', 'var'))
                newVolumes = self.currentSolution;
            end
            % projectedVols = W * u
            % Where: W = Q^(1/2) A P^(1/2) and u = P^(-1/2) x
            projectedVols = self.project(newVolumes);

            % detectorResidual = W u - f
            % Where: f = Q^(1/2) b
            detectorResidual = projectedVols - self.moderated_spots;

            % residual = W^T (f - W u)
            residual = self.backproject(detectorResidual);
        end

        function makeCGIteration(self, ii, totIters)
            % Square of projected correction -> ||W p_{k}||^2
            backprojCorr = self.backproject(self.currentProjCorrection);
            self.squareProjCorr(ii) = gtMathsDotProduct(self.currentCorrection, backprojCorr);
%             self.squareProjCorr(ii) = gtMathsSquareNorm(self.currentProjCorrection);

            % Dot product of current residual and corrected residual
            % -> r_{k}^{T} t_{k}
            self.dotResiduals(ii) = gtMathsSquareNorm(...
                                                        self.currentResidual);

            alphaPrefactor = length(self.currentSolution) * numel(self.currentSolution{1}) / totIters;
%             alphaPrefactor = 1;

            % Alpha_{k} - Quenched
            self.alpha(ii) = self.dotResiduals(ii) / self.squareProjCorr(ii) * alphaPrefactor;
            self.alpha(ii) = min(self.alpha(ii), 1);
            self.alpha(ii) = max(self.alpha(ii), 0);

            % x_{k+1} = r_{k} + alpha_{k} p_{k}
            nextSolution = internal_cell_sum_scaled_copy( ...
                self.currentSolution, self.currentCorrection, self.alpha(ii));

            % FISTA - regularization
            nextSolution = self.softThreshold(nextSolution, self.lambda);

            % NonNegative constrains
            nextSolution = gtMathsNonNegative(nextSolution);

            % FISTA and "non_negative" change the volumes, so we should update
            % the residuals (and corrections) as a consequence.
            nextResidual = self.computeResiduals(nextSolution);

            % squareNewRes = ||r_{k+1}||^2
            dotNextResisuals = gtMathsSquareNorm(nextResidual);

            % beta_{k+1} = ||r_{k+1}||^2 / ||r_{k}||^2
            self.beta(ii) = dotNextResisuals / self.dotResiduals(ii);
            self.beta(ii) = min(self.beta(ii), 1.0);
            self.beta(ii) = max(self.beta(ii), 0);

            nextCorrection = internal_cell_prod_scalar_copy(nextResidual, -1);
            nextCorrection = internal_cell_sum_scaled_assign(nextCorrection, self.currentCorrection, self.beta(ii));

            % W p_{k+1}
            self.currentProjCorrection = self.project(nextCorrection);

            % Next iteration assignments
            self.currentResidual = nextResidual;
            self.currentCorrection = nextCorrection;

            % Save Volumes x_{k}
            self.currentSolution = nextSolution;

            self.maxVals(ii) = mean(cellfun(@(x)max(x(:)), nextSolution));
            self.minVals(ii) = mean(cellfun(@(x)min(x(:)), nextSolution));
        end

        function visualize(self, ax1, ax2, ax3, ii)
            if ((mod(ii, 10) == 0) || (ii < 25) ...
                    || ((ii < 50) && (mod(ii, 2) == 0)) ...
                    || ((ii < 80) && (mod(ii, 5) == 0)))
                semilogy(abs(self.dotResiduals(1:ii)), 'Parent', ax1);
                hold(ax1, 'on')
                plot(abs(self.alpha(1:ii)), 'r', 'Parent', ax1);
                plot(self.squareProjCorr(1:ii), 'g', 'Parent', ax1);
                plot(abs(self.beta(1:ii)), 'y', 'Parent', ax1);
                hold(ax1, 'off')
                plot(self.maxVals(1:ii), 'r', 'Parent', ax3);
                hold(ax3, 'on')
                plot(self.minVals(1:ii), 'g', 'Parent', ax3);
                hold(ax3, 'off')
            end
            if (mod(ii, 3) == 0)
                slice_indx = round(size(self.currentSolution{1}, 3) / 2);
                imagesc(self.currentSolution{1}(:, :, slice_indx), 'Parent', ax2)
            end
        end

        function makeSIRTIteration(self, ii)
            numCells = numel(self.currentSolution);

            % Step length for FISTA
            self.statistics.tic('sirt_norm_residual');
            self.dotResiduals(ii) ...
                = internal_cell_square_norm(self.currentResidual);
            self.statistics.toc('sirt_norm_residual');

            self.statistics.tic('sirt_apply_sum');
            if (self.verbose)
                nextCorrection = internal_cell_prod_scalar_copy(self.currentResidual, -1);
                self.currentProjCorrection = self.project(nextCorrection);
                self.squareProjCorr(ii) = gtMathsSquareNorm(self.currentProjCorrection);
                nextSolution = internal_cell_sum_copy(self.currentSolution, nextCorrection);
            else
                nextSolution = internal_cell_sub_copy(self.currentSolution, self.currentResidual);
            end
            self.statistics.toc('sirt_apply_sum');

            % fista and nonnegative;
            % For theoretical reasons, step is always 1
            self.statistics.tic('sirt_fista_sum');
            nextSolution = self.applyFISTA(nextSolution, self.currentSolution);
            self.statistics.toc('sirt_fista_sum');

            self.statistics.tic('sirt_bbox_sum');
%             if (self.non_negative)
%                 nextSolution = gtMathsNonNegative(nextSolution);
%             end

            sizes = size(nextSolution{1});
            for n = 1:numCells
                nextSolution{n}([1:3 sizes(1)-2:sizes(1)], :, :) = 0;
                nextSolution{n}(:, [1:3 sizes(2)-2:sizes(2)], :) = 0;
                if (sizes(3) == 8)
                    nextSolution{n}(:, :, [1 2 3 6 7 8]) = 0;
                else
                    nextSolution{n}(:, :, [1 sizes(3)]) = 0;
                end
            end
            self.statistics.toc('sirt_bbox_sum');

            nextResidual = self.computeResiduals(nextSolution);

            self.currentSolution = nextSolution;
            self.currentResidual = nextResidual;

            self.maxVals(ii) = mean(cellfun(@(x)max(x(:)), nextSolution));
            self.minVals(ii) = mean(cellfun(@(x)min(x(:)), nextSolution));
        end

        function nextSolution = softThreshold(self, nextSolution, thresh)
%             waveletLevel = 1;
%             nextSolution = dwt_haar_soft_threshold(nextSolution, waveletLevel, thresh);
            for ii = 1:numel(nextSolution)
                nextSolution{ii} = gtSoftThreshold2(nextSolution{ii}, thresh);
            end
        end

        function nextSolution = applyFISTA(self, nextSolution, currentSolution)
            nextSolution = softThreshold(self, nextSolution, self.lambda);

            newWeightFISTA = Gt6DSpotReconstructor.getTplus1(self.weightFISTA);
            fistaCurrentWeight = self.getFISTAWeight(newWeightFISTA);
            self.weightFISTA = newWeightFISTA;
%             nextSolution = internal_cell_sum_FISTA_scaled_assign(...
%                         nextSolution, currentSolution, fistaCurrentWeight);
            nextSolution = internal_cell_sum_FISTA_scaled_non_neg_assign(...
                        nextSolution, currentSolution, fistaCurrentWeight);
        end
    end

    methods (Static, Access = public)
        function t_1 = getTplus1(t)
            t_1 = (1 + sqrt(1+ 4 * t * t)) / 2;
        end
    end
end
