classdef GtGrainODFuvwSolver < GtGrainODFwSolver
    properties
    end

    methods (Access = public)
        function self = GtGrainODFuvwSolver(parameters)
            self = self@GtGrainODFwSolver(parameters);
        end

        function vol = solve_synthetic(self, ref_gr, gvdm, algorithm, lambda)
            bls = ref_gr.bl(ref_gr.selected);

            self.build_sinogram(bls, 1);
            self.build_orientation_sampling_synthetic(ref_gr, gvdm, (self.get_num_ws()-2)/1.25, 1.2);
            self.build_projection_matrices(bls, ref_gr.selected);
            self.build_projection_weights();

            if (~exist('algorithm', 'var'))
                algorithm = 'cplsnn';
            end
            switch(lower(algorithm))
                case 'sirt'
                    self.solve_sirt();
                case 'cplsnn'
                    self.solve_cplsnn();
                case 'cplsl1nn'
                    self.solve_cplsl1nn(lambda);
                case 'cpl1nn'
                    self.solve_cpl1nn();
            end
            vol = self.get_volume();
        end

        function vol = solve_experimental(self, ref_gr, algorithm, lambda)
            bls = ref_gr.bl(ref_gr.selected);

            self.build_sinogram(bls, 1.2);
            self.build_orientation_sampling_experimental(ref_gr, floor(self.get_num_ws()/1.2*1.1), 1.1);
            self.build_projection_matrices(bls, ref_gr.selected);
            self.build_projection_weights();

            if (~exist('algorithm', 'var'))
                algorithm = 'cplsnn';
            end
            switch(lower(algorithm))
                case 'sirt'
                    self.solve_sirt();
                case 'cplsnn'
                    self.solve_cplsnn();
                case 'cplsl1nn'
                    self.solve_cplsl1nn(lambda);
                case 'cpl1nn'
                    self.solve_cpl1nn();
            end
            vol = self.get_volume();
        end
    end

    methods (Access = public) % Low Level API
        function build_sinogram(self, bls, oversize)
            num_blobs = numel(bls);
            blob_dephs = arrayfun(@(x)size(x.intm, 3), bls);
            blob_dephs = reshape(blob_dephs, [], 1);
            num_ws = max(blob_dephs) + 2;
            if (exist('oversize', 'var'))
                num_ws = round(num_ws * oversize);
            end

            self.size_sino = [size(bls(1).intm, 1), size(bls(1).intm, 2), num_ws, num_blobs];
            self.sino = zeros(self.size_sino);
            self.pre_paddings = floor((num_ws - blob_dephs) / 2) + 1;

            for ii_b = 1:num_blobs
                ints_interval = self.pre_paddings(ii_b):(self.pre_paddings(ii_b) + blob_dephs(ii_b) -1);
                masked_blob = bls(ii_b).intm;
                masked_blob(~bls(ii_b).mask) = 0;
                self.sino(:, :, ints_interval, ii_b) = masked_blob;
            end
            self.sino = reshape(self.sino, [], 1);
        end

        function deg_res = guess_resolution(self, ref_gr)
            deg_res_omega = guess_resolution@GtGrainODFwSolver(self, ref_gr);

            bls = ref_gr.bl(ref_gr.selected);

            bl_centers = [sum(cat(1, bls(:).bbuim), 2), sum(cat(1, bls(:).bbvim), 2)] / 2;
            gc_det_pos = gtGeoGrainCenterSam2Det(ref_gr.center, ref_gr.allblobs.omega, self.parameters);
            dists = sqrt(sum((bl_centers - gc_det_pos) .^ 2, 2));

            % Higest resolution in ETA
            deg_res_eta = min(2 .* atan2d(0.5, dists));

            % We ake the lowest resolution between the two
            deg_res = max(deg_res_eta, deg_res_omega);
        end

        function build_projection_matrices(self, bls, bl_selected)
            fprintf('Computing projection matrices..')
            c = tic();
            om_step = 180 / self.parameters.acq.nproj;
            num_ws = self.get_num_ws();

            num_orients = numel(self.grid_gr);

            bls_bbws = cat(1, bls(:).bbwim);
            min_w_conds = bls_bbws(:, 1) - self.pre_paddings + 1;
            max_w_conds = min_w_conds + num_ws - 1;
 
            bls_bbus = cat(1, bls(:).bbuim);
            bls_bbvs = cat(1, bls(:).bbvim);

            b_us = [];
            b_vs = [];
            b_ws = [];
            b_cs = [];
            b_is = [];
            b_os = [];

            for ii_o = 1:num_orients
                ab = self.grid_gr{ii_o}.allblobs;
                % W part
                ws = ab.omega(bl_selected) / om_step;

                min_ws = floor(ws);
                max_ws = min_ws + 1;

                max_w_cs = ws - min_ws;
                min_w_cs = 1 - max_w_cs;

                % UV part
                uv = ab.uv(find(bl_selected), :); % - [bls_bbus(:, 1), bls_bbvs(:, 1)] + 1;

                min_uvs = floor(uv);
                max_uvs = min_uvs + 1;

                max_uv_cs = uv - min_uvs;
                min_uv_cs = 1 - max_uv_cs;

                % Acceptance criteria
                ok_w_mins = (min_ws >= min_w_conds) & (min_ws <= max_w_conds);
                ok_w_maxs = (max_ws >= min_w_conds) & (max_ws <= max_w_conds) & (max_w_cs > eps('single'));

                ok_u_mins = (min_uvs(:, 1) >= bls_bbus(:, 1)) & (min_uvs(:, 1) <= bls_bbus(:, 2));
                ok_u_maxs = (max_uvs(:, 1) >= bls_bbus(:, 1)) & (max_uvs(:, 1) <= bls_bbus(:, 2)) & (max_uv_cs(:, 1) > eps('single'));

                ok_v_mins = (min_uvs(:, 2) >= bls_bbvs(:, 1)) & (min_uvs(:, 2) <= bls_bbvs(:, 2));
                ok_v_maxs = (max_uvs(:, 2) >= bls_bbvs(:, 1)) & (max_uvs(:, 2) <= bls_bbvs(:, 2)) & (max_uv_cs(:, 2) > eps('single'));

                u_oks = [ok_u_mins, ok_u_maxs];
                v_oks = [ok_v_mins, ok_v_maxs];
                w_oks = [ok_w_mins, ok_w_maxs];

                u_cs = [min_uv_cs(:, 1), max_uv_cs(:, 1)];
                v_cs = [min_uv_cs(:, 2), max_uv_cs(:, 2)];
                w_cs = [min_w_cs, max_w_cs];

                pos_us = [(min_uvs(:, 1) - bls_bbus(:, 1) + 1), (max_uvs(:, 1) - bls_bbus(:, 1) + 1)];
                pos_vs = [(min_uvs(:, 2) - bls_bbvs(:, 1) + 1), (max_uvs(:, 2) - bls_bbvs(:, 1) + 1)];
                pos_ws = [(min_ws - bls_bbws(:, 1) + self.pre_paddings), (max_ws - bls_bbws(:, 1) + self.pre_paddings)];

                for u_ii = 1:2
                    for v_ii = 1:2
                        for w_ii = 1:2
                            indx = find(u_oks(:, u_ii) & v_oks(:, v_ii) & w_oks(:, w_ii));
                            b_us = [b_us; pos_us(indx, u_ii)];
                            b_vs = [b_vs; pos_vs(indx, v_ii)];
                            b_ws = [b_ws; pos_ws(indx, w_ii)];

                            b_cs = [b_cs; u_cs(indx, u_ii) .* v_cs(indx, v_ii) .* w_cs(indx, w_ii)];

                            b_is = [b_is; indx];
                            b_os = [b_os; ii_o(ones(numel(indx), 1))];
                        end
                    end
                end
            end

            sino_indx = sub2ind(self.size_sino, b_us, b_vs, b_ws, b_is);

            self.S = sparse( ...
                    sino_indx, b_os, b_cs, ...
                    numel(self.sino), num_orients);

            self.St = self.S';
            fprintf('\b\b: Done in %f seconds.\n', toc(c));
        end
    end

    methods (Access = protected)
        function num_ws = get_num_ws(self)
            num_ws = self.size_sino(3);
        end
    end
end