classdef GtOrientationSampling < handle
    properties
        lattice = struct('gr', {});

        gr_ss = {};
        coeffs_gr_ss = {};

        stats;
        gvdm;

        fedpars = [];
        bl = [];
        parameters = [];
        ref_gr = [];

        verbose = false;
    end

    methods
        function self = GtOrientationSampling(bl, parameters, ref_gr, varargin)
            self.bl = bl;
            self.parameters = parameters;
            self.ref_gr = ref_gr;

            self = parse_pv_pairs(self, varargin);
        end

        function make_simple_grid_estim(self, type, edge_points, inv_oversize)
            if (~exist('inv_oversize', 'var'))
                inv_oversize = 1;
            end

            avg_R_vec = self.ref_gr.R_vector';

            omega_step = self.get_omega_step_rad();
            [raw_depths, weighted_depths] = self.get_blob_depths();

            [blob_depth, ii_bd_min] = min(weighted_depths);
            blob_depth = mean(weighted_depths);
            [blob_depth, ii_bd_max] = max(weighted_depths);
            fprintf('Blob Depths: selected %f, min %f (raw: %d), avg %f, max %f (raw: %d)\n', ...
                blob_depth, min(weighted_depths), raw_depths(ii_bd_min), ...
                mean(weighted_depths), max(weighted_depths), raw_depths(ii_bd_max));

            omega_spread = tan(blob_depth * omega_step / (2 * inv_oversize * sqrt(3)));
%             omega_spread = tan(blob_depth * omega_step / (2 * inv_oversize));
            gvdm_tmp = [];
            for cz = (-omega_spread/2):omega_spread:(omega_spread/2)
                for cy = (-omega_spread/2):omega_spread:(omega_spread/2)
                    for cx = (-omega_spread/2):omega_spread:(omega_spread/2)
                        gvdm_tmp = [gvdm_tmp, (avg_R_vec + [cx; cy; cz])];
                    end
                end
            end

            self.make_simple_grid(type, edge_points, gvdm_tmp);
        end

        function make_grid_estim_ODF_resoluion(self, type, even, oversize)
            if (~exist('oversize', 'var'))
                oversize = 1.05;
            end

            gvdm_tmp = self.guess_ODF_BB();
            size_bb = (max(gvdm_tmp, [], 2) - min(gvdm_tmp, [], 2))' * oversize;

            omega_step_rad = pi / self.parameters.acq.nproj;

            included = self.ref_gr.ondet(self.ref_gr.included);
            selected = included(self.ref_gr.selected);
            selected_etas = self.ref_gr.allblobs.eta(selected);
            lorentz = min(abs(sind(selected_etas)));

            rescaled_om_step = omega_step_rad .* lorentz;

            min_edge_points = min(ceil(size_bb ./ rescaled_om_step));

            if (exist('even', 'var') && even)
                if (~exist('oversize', 'var'))
                    self.make_even_simple_grid(type, min_edge_points, gvdm_tmp);
                else
                    self.make_even_simple_grid(type, min_edge_points, gvdm_tmp, oversize);
                end
            else
                self.make_simple_grid(type, min_edge_points, gvdm_tmp);
            end
        end

        function make_simple_grid_estim_ODF(self, type, edge_points, even, oversize)
            gvdm_tmp = self.guess_ODF_BB();

            if (exist('even', 'var') && even)
                if (~exist('oversize', 'var'))
                    self.make_even_simple_grid(type, edge_points, gvdm_tmp);
                else
                    self.make_even_simple_grid(type, edge_points, gvdm_tmp, oversize);
                end
            else
                self.make_simple_grid(type, edge_points, gvdm_tmp);
            end
        end

        function make_simple_grid(self, type, edge_points, gvdm, oversize)
            dmean = mean(gvdm, 2)';
            dstd  = std(gvdm, 0, 2)';
            dcenters = (max(gvdm, [], 2)' + min(gvdm, [], 2)') / 2;
            dranges = max(gvdm, [], 2)' - min(gvdm, [], 2)';
            fprintf('Mean orientation, std deviation, centers and ranges:\n')
            disp(dmean)
            disp(dstd)
            disp(dcenters)
            disp(dranges)

            if (~exist('oversize', 'var'))
                oversize = 1.1;
            end

            self.lattice = struct('gr', {});
            self.gvdm = {};

            type = lower(type);
            switch(type)
                case 'cubic'
                    sampling_range = GtOrientationSampling.get_cubic_range(edge_points, oversize);
                    self.produce_R_vectors_simple_grid(sampling_range, dcenters, dranges)
                case 'fcc'
                case 'bcc'
                    sampling_range = GtOrientationSampling.get_cubic_range(edge_points, oversize);
                    self.produce_R_vectors_simple_grid(sampling_range, dcenters, dranges, 1)

                    sampling_range = GtOrientationSampling.get_bcc_center_range(edge_points, oversize);
                    self.produce_R_vectors_simple_grid(sampling_range, dcenters, dranges, 2)
                otherwise
                    error('GtOrientationSampling:make_simple_grid', ...
                        'Unknown lattice structure: %s', type)
            end

            if (self.verbose && usejava('jvm'))
                self.plot_sampling(gvdm, dmean);
            end

            % Slow part which is based on gtCalculateGrain: could be improved!
            self.produce_geometry();

            self.produce_stats(type);
            if (self.verbose && usejava('jvm'))
                self.print_stats();
            end

            if (self.verbose && usejava('jvm'))
                self.plot_blob_coverage();
            end
        end

        function make_even_simple_grid(self, type, min_edge_points, gvdm, oversize)
            dmean = mean(gvdm, 2)';
            dstd  = std(gvdm, 0, 2)';
            dcenters = (max(gvdm, [], 2)' + min(gvdm, [], 2)') / 2;
            dranges = max(gvdm, [], 2)' - min(gvdm, [], 2)';
            fprintf('Mean orientation, std deviation, centers and ranges:\n')
            disp(dmean)
            disp(dstd)
            disp(dcenters)
            disp(dranges)

            if (~exist('oversize', 'var'))
                oversize = 1.1;
            end

            [~, min_ii] = min(dranges);
            other_iis = [1:min_ii-1 min_ii+1:3];
            min_spacing = dranges(min_ii) * oversize / min_edge_points;
            other_edge_points = ceil(dranges(other_iis) * oversize / min_spacing);
            edge_points(min_ii) = min_edge_points;
            edge_points(other_iis(1)) = other_edge_points(1);
            edge_points(other_iis(2)) = other_edge_points(2);

            self.lattice = struct('gr', {});
            self.gvdm = {};

            type = lower(type);
            switch(type)
                case 'cubic'
                    sampling_ranges{3} = GtOrientationSampling.get_cubic_range(edge_points(3), oversize);
                    sampling_ranges{2} = GtOrientationSampling.get_cubic_range(edge_points(2), oversize);
                    sampling_ranges{1} = GtOrientationSampling.get_cubic_range(edge_points(1), oversize);
                    self.produce_R_vectors_even_simple_grid(sampling_ranges, dcenters, dranges)
                case 'fcc'
                case 'bcc'
                otherwise
                    error('GtOrientationSampling:make_simple_grid', ...
                        'Unknown lattice structure: %s', type)
            end

            if (self.verbose && usejava('jvm'))
                self.plot_sampling(gvdm, dmean);
            end

            % Slow part which is based on gtCalculateGrain: could be improved!
            self.produce_geometry();

            self.produce_stats(type);
            if (self.verbose && usejava('jvm'))
                self.print_stats();
            end

            if (self.verbose && usejava('jvm'))
                self.plot_blob_coverage();
            end
        end

        function make_supersampling_simple_grid(self, directions, factor)
            size_gr = size(self.lattice.gr);
            ss_gvdm = reshape(self.gvdm{1}, [size_gr 3]);
            for dir_ii = directions
                switch(dir_ii)
                    case 1
                        tmp_gvdm = zeros((size(ss_gvdm) - 1) .* [factor 1 1 1] + 1);
                        tmp_gvdm(1:factor:end, :, :, :) = ss_gvdm;
                        [~, grad1, ~, ~] = gradient(ss_gvdm);
                        for ii = 2:factor
                            tmp_gvdm(ii:factor:end, :, :, :) = ...
                                ss_gvdm(1:end-1, :, :, :) + grad1(1:end-1, :, :, :) .* (ii-1) ./ factor;
                        end
                    case 2
                        tmp_gvdm = zeros((size(ss_gvdm) - 1) .* [1 factor 1 1] + 1);
                        tmp_gvdm(:, 1:factor:end, :, :) = ss_gvdm;
                        [grad2, ~, ~, ~] = gradient(ss_gvdm);
                        for ii = 2:factor
                            tmp_gvdm(:, ii:factor:end, :, :) = ...
                                ss_gvdm(:, 1:end-1, :, :) + grad2(:, 1:end-1, :, :) .* (ii-1) ./ factor;
                        end
                    case 3
                        tmp_gvdm = zeros((size(ss_gvdm) - 1) .* [1 1 factor 1] + 1);
                        tmp_gvdm(:, :, 1:factor:end, :) = ss_gvdm;
                        [~, ~, grad3, ~] = gradient(ss_gvdm);
                        for ii = 2:factor
                            tmp_gvdm(:, :, ii:factor:end, :) = ...
                                ss_gvdm(:, :, 1:end-1, :) + grad3(:, :, 1:end-1, :) .* (ii-1) ./ factor;
                        end
                end
                ss_gvdm = tmp_gvdm;
            end

            % Let's compute the position of the regular grid in the super
            % sampled, and remove those orientations
            if ismember(1, directions)
                range1 = 1:factor:size(ss_gvdm, 1);
            else
                range1 = 1:size(ss_gvdm, 1);
            end
            if ismember(2, directions)
                range2 = 1:factor:size(ss_gvdm, 2);
            else
                range2 = 1:size(ss_gvdm, 2);
            end
            if ismember(3, directions)
                range3 = 1:factor:size(ss_gvdm, 3);
            else
                range3 = 1:size(ss_gvdm, 3);
            end

            indx = sub2ind([size(ss_gvdm, 1), size(ss_gvdm, 2), size(ss_gvdm, 3)], ...
                repmat(range1', [1 numel(range2) numel(range3)]), ...
                repmat(range2, [numel(range1) 1 numel(range3)]), ...
                repmat(reshape(range3, [1 1 numel(range3)]), [numel(range1) numel(range2) 1]) );

            ss_gvdm = reshape(ss_gvdm, [(numel(ss_gvdm)/3) 3]);

            ss_gvdm(indx, :) = [];

            self.gr_ss = {};
            for rvec_ii = size(ss_gvdm, 1):-1:1
                self.gr_ss{rvec_ii}.R_vector = ss_gvdm(rvec_ii, :);
            end

            self.plot_supersampling()

            self.produce_geometry_supersampling()
        end

        function valid = is_valid(self, type, edge_points)
            valid = strcmpi(self.stats.sampling.type, type) ...
                && all(size(self.lattice(1).gr) == edge_points([1 1 1]));
        end

        function [or, or_ss] = get_orientations(self)
            or = {};
            for ii = 1:numel(self.lattice)
                or = [or(:); self.lattice(ii).gr(:)];
            end
            or_ss = self.gr_ss(:);
        end

        function coeffs_ss = get_ss_coeffs(self)
            coeffs_ss = self.coeffs_gr_ss;
        end

        function r_vecs = get_R_vectors(self)
            r_vecs = self.gvdm;
        end

        function tot_orient = get_total_num_orientations(self)
            tot_orient = sum(arrayfun(@(x)numel(x.gr), self.lattice));
        end

        function [avg_R_vecs, avg_R_vecs_int] = getAverageOrientations(self, vols, selected)
            [avg_R_vecs, avg_R_vecs_int] = GtOrientationSampling.computeAverageOrientations(self.get_orientations(), vols, selected);
        end

        function avg_R_vec = getAverageOrientation(self, vols, selected)
            avg_R_vec = GtOrientationSampling.computeAverageOrientation(self.get_orientations(), vols, selected);
        end

        function odf = getODF(self, vols, selected)
            odf = GtOrientationSampling.computeODF(self.get_orientations(), vols, selected);
        end
    end

    methods (Access = protected)
        function gvdm_tmp = guess_ODF_BB(self)
            omega_step = 180 / self.parameters.acq.nproj;

            g = gtMathsRod2OriMat(self.ref_gr.R_vector);
            selected = self.ref_gr.included(self.ref_gr.selected);
            num_refl = numel(selected);

            Gl = self.ref_gr.allblobs.plorig(selected, :)';
            hhs = gtVectorLab2Cryst(Gl', g)';
            % Taking theoretical omegas instead of experimental ones!
            w = self.ref_gr.allblobs.omega(self.ref_gr.ondet(selected));
            bls = self.bl(self.ref_gr.selected);

            extreme_points = zeros(2*num_refl, 3);

            for ii_r = 1:num_refl
                bl_r = bls(ii_r);
                ints_w = squeeze(sum(sum(bl_r.intm, 1), 2));
                real_bounds = [max(find(ints_w, 1, 'first'), 1), ...
                    min(find(ints_w, 1, 'last'), numel(ints_w))];
                sub_w = (bl_r.bbwim(1):bl_r.bbwim(2)) * omega_step;

                % Only the edges for the moment!
                sub_w = sub_w(real_bounds(1):real_bounds(2)) - w(ii_r);
%                 sub_w = sub_w(real_bounds) - w(ii_r);
                num_omegas = numel(sub_w);
                sub_w = reshape(sub_w, 1, 1, []);

                Ws = [cosd(sub_w), -sind(sub_w), zeros(1, 1, num_omegas); ...
                    sind(sub_w), cosd(sub_w), zeros(1, 1, num_omegas); ...
                    zeros(1, 1, num_omegas), zeros(1, 1, num_omegas), ones(1, 1, num_omegas)];

                hh = hhs(:, ii_r);

                r0 = zeros(num_omegas, 3);
                r_dir = zeros(num_omegas, 3);

                for ii_w = 1:num_omegas

                    plnorm = Ws(:, :, ii_w) * Gl(:, ii_r);

                    r0(ii_w, :) = [hh(2) * plnorm(3) - hh(3) * plnorm(2), ...
                          hh(3) * plnorm(1) - hh(1) * plnorm(3), ...
                          hh(1) * plnorm(2) - hh(2) * plnorm(1)] / (1 + hh' * plnorm);

                    r_dir(ii_w, :) = (hh' + plnorm') / (1 + hh' * plnorm);
                    r_dir(ii_w, :) = r_dir(ii_w, :) / norm(r_dir(ii_w, :));
                end

                extreme_points(ii_r*2-1:ii_r*2, :) = [r0(1, :); r0(end, :)];
                ex_p_dirs(ii_r*2-1:ii_r*2, :) = [r_dir(1, :); r_dir(end, :)];
            end

            trasl_ex_points = self.ref_gr.R_vector(ones(size(extreme_points, 1), 1), :) - extreme_points;
            trasl_ex_points = sum(trasl_ex_points .* ex_p_dirs, 2) ./ sqrt(sum(ex_p_dirs .^ 2, 2));
            trasl_ex_points = extreme_points + trasl_ex_points(:, [1 1 1]) .* ex_p_dirs;

            bbox_r0s = [min(trasl_ex_points, [], 1), max(trasl_ex_points, [], 1)];
            bbox_r0s = [bbox_r0s(1:3), bbox_r0s(4:6) - bbox_r0s(1:3)];

            fprintf('Estimated BB in orientation space: [(%f %f %f), (%f %f %f)]\n', bbox_r0s)

            gvdm_tmp = [];
            for cz = 0:1
                for cy = 0:1
                    for cx = 0:1
                        gvdm_tmp = [gvdm_tmp, (bbox_r0s(1:3)' + [cx; cy; cz] .* bbox_r0s(4:6)')];
                    end
                end
            end
        end

        function produce_R_vectors_simple_grid(self, sampling_range, dcenters, dranges, o_ii)
            if (~exist('o_ii', 'var'))
                o_ii = 1;
            end
            num_comps = numel(sampling_range);
            try
                self.gvdm{o_ii} = zeros(num_comps ^ 3, 3);
            catch mexc
                fprintf('Trying to allocate a [%d, 3] array.. but without enough memory!\n', ...
                    num_comps ^ 3)
                rethrow(mexc)
            end

            points = zeros(num_comps, 3);
            dcenters = dcenters(ones(1, num_comps), :);
            dranges = dranges(ones(1, num_comps), :);

            self.lattice(o_ii).gr = cell(num_comps, num_comps, num_comps);

            l_ii = 0;
            for z_ii = 1:num_comps
                for y_ii = 1:num_comps
                    points(:, 1) = sampling_range(:);
                    points(:, 2) = sampling_range(y_ii);
                    points(:, 3) = sampling_range(z_ii);

                    R_vectors = dcenters + points .* dranges;
                    self.gvdm{o_ii}(l_ii+1:l_ii+num_comps, :) = R_vectors;
                    l_ii = l_ii + num_comps;
                end
            end
        end

        function produce_R_vectors_even_simple_grid(self, sampling_ranges, dcenters, dranges, o_ii)
            if (~exist('o_ii', 'var'))
                o_ii = 1;
            end
            num_z_comps = numel(sampling_ranges{3});
            num_y_comps = numel(sampling_ranges{2});
            num_x_comps = numel(sampling_ranges{1});
            try
                self.gvdm{o_ii} = zeros(num_z_comps * num_y_comps * num_x_comps, 3);
            catch mexc
                fprintf('Trying to allocate a [%d, 3] array.. but without enough memory!\n', ...
                    num_z_comps * num_y_comps * num_x_comps)
                rethrow(mexc)
            end

            points = zeros(num_x_comps, 3);
            dcenters = dcenters(ones(1, num_x_comps), :);
            dranges = dranges(ones(1, num_x_comps), :);

            self.lattice(o_ii).gr = cell(num_x_comps, num_y_comps, num_z_comps);

            l_ii = 0;
            for z_ii = 1:num_z_comps
                for y_ii = 1:num_y_comps
                    points(:, 1) = sampling_ranges{1}(:);
                    points(:, 2) = sampling_ranges{2}(y_ii);
                    points(:, 3) = sampling_ranges{3}(z_ii);

                    R_vectors = dcenters + points .* dranges;
                    self.gvdm{o_ii}(l_ii+1:l_ii+num_x_comps, :) = R_vectors;
                    l_ii = l_ii + num_x_comps;
                end
            end
        end

        function produce_geometry(self)
            grain_center = self.get_grain_center();

            % Bounding box positions
            for ii = length(self.bl):-1:1
                bbpos(ii,:) = [self.bl(ii).bbuim(1) self.bl(ii).bbvim(1) self.bl(ii).bbsize(1:2)];
            end

            fprintf('Computing geometries: ')
            c = tic();
            for o_ii = 1:numel(self.lattice)
                num_orients = numel(self.lattice(o_ii).gr);
                num_chars_o = fprintf('order %d/%d: (%d) ', o_ii-1, numel(self.lattice), num_orients);

                order_gvdm = self.gvdm{o_ii};

                num_chars_gr = fprintf('creating initial structures..');
                for ii = 1:num_orients
                    self.lattice(o_ii).gr{ii} = struct('R_vector', order_gvdm(ii, :), 'center', grain_center, 'phaseid', 1);
                end
                fprintf(repmat('\b', [1 num_chars_gr]));

                switch(2)
                    case 3 % All in once
                        num_chars_gr = fprintf('gtCalculateGrain in parallel..');
                        self.lattice(o_ii).gr = gtCalculateGrain_p(self.lattice(o_ii).gr, self.parameters);
                    case 2 % In chunks
                        num_chars_gr = fprintf('gtCalculateGrain in chunks: ');
                        chunk_size = 1000;
                        for ii = 1:chunk_size:num_orients
                            end_chunk = min(ii+chunk_size, num_orients);
                            num_chars_gr_ii = fprintf('%05d/%05d', ii, num_orients);
                            self.lattice(o_ii).gr(ii:end_chunk) = ...
                                gtCalculateGrain_p(self.lattice(o_ii).gr(ii:end_chunk), self.parameters);
                            fprintf(repmat('\b', [1 num_chars_gr_ii]));
                        end
                    case 1 % One by One
                        num_chars_gr = fprintf('gtCalculateGrain: ');
                        for ii = 1:num_orients
                            num_chars_gr_ii = fprintf('%05d/%05d', ii, num_orients);
                            self.lattice(o_ii).gr{ii} = gtCalculateGrain(self.lattice(o_ii).gr{ii}, self.parameters);
                            fprintf(repmat('\b', [1 num_chars_gr_ii]));
                        end
                end
                fprintf(repmat('\b', [1 num_chars_gr]));
                fprintf(repmat(' ', [1 num_chars_gr]));
                fprintf(repmat('\b', [1 num_chars_gr]));

                num_chars_gr = fprintf('filering unneeded "allblobs" fields..');
                if (isfield(self.ref_gr, 'ondet') && isfield(self.ref_gr, 'included'))
                    self.lattice(o_ii).gr = GtOrientationSampling.selectBlobsFromGrains( ...
                        self.lattice(o_ii).gr, self.ref_gr.ondet(self.ref_gr.included));
                end
                fprintf(repmat('\b', [1 num_chars_gr]));
                fprintf(repmat(' ', [1 num_chars_gr]));
                fprintf(repmat('\b', [1 num_chars_gr]));

                num_chars_gr = fprintf('computing projection geometries..');
                bbpos_oi = repmat(bbpos, [num_orients, 1]);

                allblobs = cellfun(@(x){x.allblobs}, self.lattice(o_ii).gr);
                allblobs = [allblobs{:}];
                dvecs = cat(1, allblobs(:).dvecsam);
                omegas = cat(1, allblobs(:).omega);

                proj_geom = gtGeoProjForReconstruction(...
                    dvecs, omegas, grain_center, bbpos_oi, [], ...
                    self.parameters.labgeo, ...
                    self.parameters.samgeo, ...
                    self.parameters.recgeo, ...
                    'ASTRA_grain');

                count = 1;
                for ii = 1:num_orients
                    num_vecs = size(self.lattice(o_ii).gr{ii}.allblobs.dvecsam, 1);
                    self.lattice(o_ii).gr{ii}.allblobs.proj_geom = proj_geom(count:(count+num_vecs-1), :);
                    count = count + num_vecs;
                end
                fprintf(repmat('\b', [1 num_chars_gr]));
                fprintf(repmat(' ', [1 num_chars_gr]));
                fprintf(repmat('\b', [1 num_chars_gr]));

                fprintf(repmat('\b', [1 num_chars_o]));
            end
            total_orient = sum(arrayfun(@(x)numel(x.gr), self.lattice));
            fprintf(' (orders: %d, total orientations: %04d, in %3.1f s) Done.\n', ...
                numel(self.lattice), total_orient, toc(c));
        end

        function produce_geometry_supersampling(self)
            grain_center = self.get_grain_center();

            % Bounding box positions
            for ii = length(self.bl):-1:1
                bbpos(ii,:) = [self.bl(ii).bbuim(1) self.bl(ii).bbvim(1) self.bl(ii).bbsize(1:2)];
            end

            num_gr_os = numel(self.gr_ss);
            fprintf('Computing super-sampling geometries: ')
            c = tic();
            for ii = num_gr_os:-1:1
                if (mod(num_gr_os - ii, 10) == 0)
                    numChars = fprintf('%03d/%03d', num_gr_os - ii, num_gr_os);
                end

                self.gr_ss{ii}.center  = grain_center;
                self.gr_ss{ii}.phaseid = 1;

                % All blob data;
                % don't use proj_geom from gtCalculateGrain !
                self.gr_ss{ii} = gtCalculateGrain(self.gr_ss{ii}, self.parameters);

                if (isfield(self.ref_gr, 'ondet') && isfield(self.ref_gr, 'included'))
                    self.gr_ss{ii}.allblobs = GtOrientationSampling.selectBlobsFromGrain( ...
                        self.gr_ss{ii}.allblobs, self.ref_gr.ondet(self.ref_gr.included));
                end

                self.gr_ss{ii}.allblobs.proj_geom = gtGeoProjForReconstruction(...
                    self.gr_ss{ii}.allblobs.dvecsam, ...
                    self.gr_ss{ii}.allblobs.omega, ...
                    grain_center, ...
                    bbpos, ...
                    [], ...
                    self.parameters.labgeo, ...
                    self.parameters.samgeo, ...
                    self.parameters.recgeo, ...
                    'ASTRA_grain');

                if (mod(num_gr_os - ii, 10) == 0)
                    fprintf(repmat('\b', [1 numChars]));
                end
            end
            fprintf(' (%04d in %3.1f s) Done.\n', num_gr_os, toc(c))

            ones_num_gr_os = ones(num_gr_os, 1);
            fprintf('Computing interpolation coefficients: ')
            c = tic();
            points_ss = [self.gr_ss{:}];
            points_ss = cat(1, points_ss.R_vector);

            shifts = self.stats.sampling.bbox_shift(ones_num_gr_os, :);
            points_ss = points_ss - shifts;

            gaps = self.stats.sampling.gaps(ones_num_gr_os, :);
            points_ss = points_ss ./ gaps;

            indx_min = floor(points_ss) + 1;
            indx_coeffs(:, :, 1) = 1 - (points_ss - indx_min + 1);
            indx_coeffs(:, :, 2) = 1 - indx_coeffs(:, :, 1);

            size(self.lattice.gr)

            progressive = 1;
            for z_ii = 0:1
                for y_ii = 0:1
                    for x_ii = 0:1
                        indx(:, :, progressive) = indx_min + [ ...
                            x_ii * ones_num_gr_os, ...
                            y_ii * ones_num_gr_os, ...
                            z_ii * ones_num_gr_os];

                        coeffs(:, progressive) = ...
                            indx_coeffs(:, 1, x_ii+1) .* ...
                            indx_coeffs(:, 2, y_ii+1) .* ...
                            indx_coeffs(:, 3, z_ii+1);

                        progressive = progressive +1;
                    end
                end
            end

            indx_so = reshape(repmat(permute(1:size(indx, 1), [1 3 2]), [8 1 1]), [], 1);
            indx = reshape(permute(indx, [3 1 2]), [], 3);
            coeffs = reshape(permute(coeffs, [2 3 1]), [], 1);

            good_coeffs = abs(coeffs) > 1e-4;

            indx_so = indx_so(good_coeffs);
            indx = indx(good_coeffs, :);
            coeffs = coeffs(good_coeffs);

%             [ indx_so, indx, coeffs ]

            indx = sub2ind(size(self.lattice.gr), indx(:, 1), indx(:, 2), indx(:, 3));

            self.coeffs_gr_ss = struct( ...
                'indx', cell(numel(points_ss), 1), ...
                'coeff', cell(numel(points_ss), 1) );

            for l_ii = numel(indx):-1:1
                numChars = fprintf('%03d/%03d', numel(self.gr_ss) - ii, numel(self.gr_ss));

                self.coeffs_gr_ss(indx_so(l_ii)).indx  = [ self.coeffs_gr_ss(indx_so(l_ii)).indx  indx(l_ii)   ];
                self.coeffs_gr_ss(indx_so(l_ii)).coeff = [ self.coeffs_gr_ss(indx_so(l_ii)).coeff coeffs(l_ii) ];

                fprintf(repmat('\b', [1 numChars]));
            end
            fprintf(' (%04d in %3.1f s) Done.\n', numel(self.gr_ss), toc(c))
        end

        function produce_stats(self, type)
            self.stats.sampling.type = type;
            switch(type)
                case 'cubic'
                    self.compute_pixel_distances_simple_grid();
                    self.compute_blob_coverage_simple_grid();
                    self.compute_sampling_distances_cubic();
                case 'fcc'
                case 'bcc'
                    self.compute_pixel_distances_bcc();
                    self.compute_blob_coverage_simple_grid();
                    self.compute_sampling_distances_bcc();
            end
        end

        function compute_sampling_distances_cubic(self)
            self.stats.sampling.gaps(1) = self.lattice.gr{2, 1, 1}.R_vector(1) - self.lattice.gr{1, 1, 1}.R_vector(1);
            self.stats.sampling.gaps(2) = self.lattice.gr{1, 2, 1}.R_vector(2) - self.lattice.gr{1, 1, 1}.R_vector(2);
            self.stats.sampling.gaps(3) = self.lattice.gr{1, 1, 2}.R_vector(3) - self.lattice.gr{1, 1, 1}.R_vector(3);
            self.stats.sampling.bbox_size = self.lattice.gr{end}.R_vector - self.lattice.gr{1}.R_vector;
            self.stats.sampling.bbox_shift = self.lattice.gr{1}.R_vector;
        end

        function compute_sampling_distances_bcc(self)
            self.stats.sampling.gaps = self.lattice(2).gr{1}.R_vector - self.lattice(1).gr{1}.R_vector;
            self.stats.sampling.bbox_size = self.lattice(1).gr{end}.R_vector - self.lattice(1).gr{1}.R_vector;
            self.stats.sampling.bbox_shift = self.lattice(1).gr{1}.R_vector;
        end

        function compute_pixel_distances_simple_grid(self)
            max_dist_x = 0;
            max_dist_y = 0;
            max_angle_x = 0;
            max_angle_y = 0;
 
            for z_ii = 1:size(self.lattice(1).gr, 3)
                for y_ii = 1:size(self.lattice(1).gr, 2)
                    for x_ii = 1:size(self.lattice(1).gr, 1)-1
                        gr1 = self.lattice(1).gr{x_ii, y_ii, z_ii};
                        gr2 = self.lattice(1).gr{x_ii+1, y_ii, z_ii};
                        [gr12_dist, gr12_angle] = self.compute_max_proj_pixel_dist(gr1, gr2);
                        max_dist_x = max(max_dist_x, gr12_dist);
                        max_angle_x = max(max_angle_x, gr12_angle);
                    end
                end

                for y_ii = 1:size(self.lattice(1).gr, 2)-1
                    for x_ii = 1:size(self.lattice(1).gr, 1)
                        gr1 = self.lattice(1).gr{x_ii, y_ii, z_ii};
                        gr2 = self.lattice(1).gr{x_ii, y_ii+1, z_ii};
                        [gr12_dist, gr12_angle] = self.compute_max_proj_pixel_dist(gr1, gr2);
                        max_dist_y = max(max_dist_y, gr12_dist);
                        max_angle_y = max(max_angle_y, gr12_angle);
                    end
                end
            end

            self.stats.proj.max_pixel_dist_x = max_dist_x;
            self.stats.proj.max_pixel_dist_y = max_dist_y;

            self.stats.proj.max_angle_x = max_angle_x;
            self.stats.proj.max_angle_y = max_angle_y;

            self.stats.sample.avg_dist = self.compute_average_sample_distance();
        end

        function compute_pixel_distances_bcc(self)
            max_dist_x = 0;
            max_dist_y = 0;
            max_angle_x = 0;
            max_angle_y = 0;

            for z_ii = 1:size(self.lattice(2).gr, 3)
                for y_ii = 1:size(self.lattice(2).gr, 2)
                    for x_ii = 1:size(self.lattice(2).gr, 1)
                        gr1 = self.lattice(1).gr{x_ii, y_ii, z_ii};
                        gr2 = self.lattice(2).gr{x_ii, y_ii, z_ii};
                        [gr12_dist, gr12_angle] = self.compute_max_proj_pixel_dist(gr1, gr2);
                        max_dist_x = max(max_dist_x, gr12_dist);
                        max_angle_x = max(max_angle_x, gr12_angle);
                    end
                end

                for y_ii = 1:size(self.lattice(2).gr, 2)
                    for x_ii = 1:size(self.lattice(2).gr, 1)
                        gr1 = self.lattice(1).gr{x_ii+1, y_ii, z_ii};
                        gr2 = self.lattice(2).gr{x_ii, y_ii, z_ii};
                        [gr12_dist, gr12_angle] = self.compute_max_proj_pixel_dist(gr1, gr2);
                        max_dist_x = max(max_dist_x, gr12_dist);
                        max_angle_x = max(max_angle_x, gr12_angle);
                    end
                end

                for y_ii = 1:size(self.lattice(2).gr, 2)
                    for x_ii = 1:size(self.lattice(2).gr, 1)
                        gr1 = self.lattice(1).gr{x_ii, y_ii+1, z_ii};
                        gr2 = self.lattice(2).gr{x_ii, y_ii, z_ii};
                        [gr12_dist, gr12_angle] = self.compute_max_proj_pixel_dist(gr1, gr2);
                        max_dist_x = max(max_dist_x, gr12_dist);
                        max_angle_x = max(max_angle_x, gr12_angle);
                    end
                end

                for y_ii = 1:size(self.lattice(2).gr, 2)
                    for x_ii = 1:size(self.lattice(2).gr, 1)
                        gr1 = self.lattice(1).gr{x_ii+1, y_ii+1, z_ii};
                        gr2 = self.lattice(2).gr{x_ii, y_ii, z_ii};
                        [gr12_dist, gr12_angle] = self.compute_max_proj_pixel_dist(gr1, gr2);
                        max_dist_x = max(max_dist_x, gr12_dist);
                        max_angle_x = max(max_angle_x, gr12_angle);
                    end
                end
            end

            self.stats.proj.max_pixel_dist_x = max_dist_x;
            self.stats.proj.max_pixel_dist_y = max_dist_y;

            self.stats.proj.max_angle_x = max_angle_x;
            self.stats.proj.max_angle_y = max_angle_y;

            self.stats.sample.avg_dist = self.compute_average_sample_distance();
        end

        function dist = compute_average_sample_distance(self)
            dist = 0;
            tot_orient = get_total_num_orientations(self);
            for o_ii = 1:numel(self.lattice)
                for z_ii = 1:size(self.lattice(o_ii).gr, 3)
                    for y_ii = 1:size(self.lattice(o_ii).gr, 2)
                        for x_ii = 1:size(self.lattice(o_ii).gr, 1)
                            positions = self.lattice(o_ii).gr{x_ii, y_ii, z_ii}.allblobs.proj_geom(:, 4:6);
                            dist = dist + sum(sqrt(sum(positions .^ 2, 2))) / size(positions, 1);
                        end
                    end
                end
            end
            dist = dist / tot_orient;
        end

        function [dist, angle] = compute_max_proj_pixel_dist(self, gr1, gr2)
            omega_step = self.get_omega_step_deg();

            same_omegas = abs(gr1.allblobs.omega - gr2.allblobs.omega) < omega_step;

            geom1 = gr1.allblobs.proj_geom(same_omegas, :);
            geom2 = gr2.allblobs.proj_geom(same_omegas, :);

            dirs1 = geom1(:, 1:3);
            dirs2 = geom2(:, 1:3);

            det_pos1 = geom1(:, 4:6);
            det_pos2 = geom2(:, 4:6);

            lengths1 = sqrt(sum(det_pos1 .^ 2, 2));
            lengths2 = sqrt(sum(det_pos2 .^ 2, 2));

            disps1 = dirs1 .* lengths1(:, [1 1 1]) - det_pos1;
            disps2 = dirs2 .* lengths2(:, [1 1 1]) - det_pos2;

            % Approximately
            disps = disps1 - disps2;
            [dist, ~] = max(sqrt(sum(disps .^ 2, 2)));

%             rot_dirs = cross(det_pos2, det_pos1);
%             norm_rot_dirs = sqrt(sum(rot_dirs .^ 2, 2));
%             rot_dirs = rot_dirs ./ norm_rot_dirs(:, [1 1 1]);
% 
%             renorm_det_pos = lengths1 .* lengths2;
%             dot_det_pos = det_pos1 .* det_pos2;
%             rot_angles = abs(acos(sum(dot_det_pos, 2) ./ renorm_det_pos));
% 
%             rotated_dirs2 = zeros(size(dirs2));
% 
%             for r_ii = 1:numel(rot_angles)
%                 rot_comp = gtMathsRotationMatrixComp(rot_dirs(r_ii, :), 'row');
%                 rot_tensor = gtMathsRotationTensor(rot_angles(r_ii), rot_comp);
% 
%                 rotated_dirs2(r_ii, :) = dirs2(r_ii, :) * rot_tensor;
%             end
% 
%             angles = acos(sum(dirs1 .* rotated_dirs2, 2));

            angles = abs(acos(sum(dirs1 .* dirs2, 2)));
            [angle, ~] = max(angles);
        end

        function compute_blob_coverage_simple_grid(self)
            omega_step = self.get_omega_step_deg();

            num_blobs = numel(self.bl);
            max_depth_blobs = 0;
            for b_ii = 1:num_blobs
                max_depth_blobs = max(max_depth_blobs, size(self.bl(b_ii).intm, 3));
            end

            switch(self.stats.sampling.type)
                case 'cubic'
                    num_z_comps = size(self.lattice.gr, 3);

                    % Coverage of the slices in blobs (by Z layer of the sampling)
                    self.stats.pics.blob_coverage_map = cell(num_z_comps);
            end
            % Total intensity of each slice in a blob
            self.stats.pics.blob_intensity = zeros(num_blobs, max_depth_blobs);
            % Coverage of the slices in blobs (generic)
            self.stats.pics.blob_coverage = zeros(num_blobs, max_depth_blobs);

            min_blob_pos = zeros(num_blobs, 1);
            max_blob_pos = zeros(num_blobs, 1);

            for b_ii = 1:num_blobs
                depth_blob = size(self.bl(b_ii).intm, 3);
                min_blob_pos(b_ii) = floor((max_depth_blobs - depth_blob)/2) + 1;
                max_blob_pos(b_ii) = min_blob_pos(b_ii) + depth_blob - 1;

                self.stats.pics.blob_intensity(b_ii, min_blob_pos(b_ii):max_blob_pos(b_ii)) ...
                    = squeeze(sum(sum(self.bl(b_ii).intm, 2), 1));
            end

            limits = cat(1, self.bl(:).bbwim);
            for o_ii = 1:numel(self.lattice)
                for z_ii = 1:size(self.lattice(o_ii).gr, 3)
                    switch(self.stats.sampling.type)
                        case 'cubic'
                            self.stats.pics.blob_coverage_map{z_ii} = zeros(num_blobs, max_depth_blobs);
                    end

                    for y_ii = 1:size(self.lattice(o_ii).gr, 2)
                        for x_ii = 1:size(self.lattice(o_ii).gr, 1)
                            omegas = self.lattice(o_ii).gr{x_ii, y_ii, z_ii}.allblobs.omega / omega_step;
                            within_lims = find(omegas >= limits(:, 1) & omegas <= limits(:, 2));
                            pos_in_blobs = round(omegas(within_lims)) - limits(within_lims, 1) + 1;

                            pos_in_coverage = pos_in_blobs + min_blob_pos(within_lims) - 1;

                            indx = sub2ind(size(self.stats.pics.blob_coverage), within_lims, pos_in_coverage);

                            switch(self.stats.sampling.type)
                                case 'cubic'
                                    self.stats.pics.blob_coverage_map{z_ii}(indx) = ...
                                        self.stats.pics.blob_coverage_map{z_ii}(indx) + 1;
                            end

                            self.stats.pics.blob_coverage(indx) = 1;
                        end
                    end
                end
            end

            % Intensity reached by at least one orientation
            self.stats.pics.blob_covered_intensity = ...
                self.stats.pics.blob_intensity .* self.stats.pics.blob_coverage;
            % Intensity NOT reached by any orientation
            self.stats.pics.blob_uncovered_intensity = ...
                self.stats.pics.blob_intensity .* (1 - self.stats.pics.blob_coverage);
        end

        function plot_blob_coverage(self)
            f1 = figure();
            lims = [min(self.stats.pics.blob_intensity(:)) max(self.stats.pics.blob_intensity(:))];
            ax1 = subplot(1, 4, 1, 'parent', f1);
            imagesc(self.stats.pics.blob_covered_intensity, 'parent', ax1, lims);
            title(ax1, 'Covered Intensity');
            ax2 = subplot(1, 4, 2, 'parent', f1);
            imagesc(self.stats.pics.blob_uncovered_intensity, 'parent', ax2, lims);
            title(ax2, 'Uncovered Intensity');
            ax2 = subplot(1, 4, 3, 'parent', f1);
            imagesc(self.stats.pics.blob_coverage, 'parent', ax2);
            title(ax2, 'Coverage');
            ax2 = subplot(1, 4, 4, 'parent', f1);
            imagesc(self.stats.pics.blob_intensity, 'parent', ax2);
            title(ax2, 'Intensity');

            switch(self.stats.sampling.type)
                case 'cubic'
                    num_z_comps = size(self.lattice.gr, 3);

                    % Let's visualize the maps separately
                    grid_rows = floor(sqrt(num_z_comps));
                    grid_cols = ceil(num_z_comps / grid_rows);
                    f2 = figure();
                    for ii = 1:num_z_comps
                        ax = subplot(grid_rows, grid_cols, ii, 'parent', f2);
                        imagesc(self.stats.pics.blob_coverage_map{ii}, 'parent', ax);
                        title(ax, sprintf('Coverage of Z Layer %d', ii));
                    end
                otherwise
            end

            drawnow();
        end

        function plot_sampling(self, gvdm, dmean)
            points = self.gvdm;
            points = cat(1, points{:});

            figure %, scatter3(points(:, 1), points(:, 2), points(:, 3))
            hold on
            scatter3(gvdm(1, :)', gvdm(2, :)', gvdm(3, :)', 20, 'c')
            scatter3(points(:, 1), points(:, 2), points(:, 3), 20, 'r')
            scatter3(mean(points(:, 1)), mean(points(:, 2)), mean(points(:, 3)), 30, 'm', 'filled')
            scatter3(dmean(1), dmean(2), dmean(3), 30, 'g', 'filled')
            hold off

            drawnow();
        end

        function plot_supersampling(self)
            points = [self.lattice.gr{:}];
            points = cat(1, points.R_vector);
            points_os = [self.gr_ss{:}];
            points_os = cat(1, points_os.R_vector);

            figure %, scatter3(points(:, 1), points(:, 2), points(:, 3))
            hold on
            scatter3(points(:, 1), points(:, 2), points(:, 3), 20, 'c')
            scatter3(points_os(:, 1), points_os(:, 2), points_os(:, 3), 20, 'r')
            hold off

            drawnow();
        end

        function print_stats(self)
            fprintf('Sampling:\n - lattice type: "%s" (size: %d, %d, %d)\n', ...
                self.stats.sampling.type, size(self.lattice(1).gr));
            fprintf(' - Gaps: x %e, y %e, z %e\n', ...
                self.stats.sampling.gaps(1), ...
                self.stats.sampling.gaps(2), ...
                self.stats.sampling.gaps(3) );
            fprintf(' - BBox: size (%e, %e, %e), shift (%e, %e, %e)\n', ...
                self.stats.sampling.bbox_size, ...
                self.stats.sampling.bbox_shift )
            fprintf('Average smple distance: %g (pixels)\n', ...
                self.stats.sample.avg_dist)
            fprintf('Max projection pixel distances: x %f, y %f\n', ...
                self.stats.proj.max_pixel_dist_x, ...
                self.stats.proj.max_pixel_dist_y );
            fprintf('Max projection angle distances: x %e (deg %f), y %e (deg %f)\n', ...
                self.stats.proj.max_angle_x, ...
                (180/pi*self.stats.proj.max_angle_x), ...
                self.stats.proj.max_angle_y, ...
                (180/pi*self.stats.proj.max_angle_y) );
        end

        function omega_step = get_omega_step_deg(self)
            omega_step = 180 / self.parameters.acq.nproj;
        end

        function omega_step = get_omega_step_rad(self)
            omega_step = pi / self.parameters.acq.nproj;
        end

        function gc = get_grain_center(self)
            if (~isempty(self.fedpars))
                gc = self.fedpars.vox000sam' + ...
                    (self.fedpars.grainsize' ./ 2 + [0.5; 0.5; 0.5]) .* self.parameters.recgeo.voxsize';
                gc = gc';
            else
                gc = self.ref_gr.center; % .* self.parameters.recgeo.voxsize;
            end
        end

        function [raw_depths, weighted_depths] = get_blob_depths(self)
            % Try to account for Forentz factor, at least to avoid that max
            % blows up. Computation is only approximated, so doesn't really
            % count
            indx = self.ref_gr.ondet(self.ref_gr.included(self.ref_gr.selected));
            etas = self.ref_gr.allblobs.eta(indx);
            factors = abs(sind(etas));

            raw_depths = cellfun(@(x){sum(sum(abs(x), 1), 2)}, {self.bl(self.ref_gr.selected).intm}');
            raw_depths = cellfun(@(x)(find(x >= 0, 1, 'last') - find(x >= 0, 1, 'first') + 1), raw_depths);
%             raw_depths = cellfun(@(x)size(x, 3), {self.bl(self.ref_gr.selected).intm}') - 2;
            weighted_depths = raw_depths .* factors;
        end

        function [max_blob_depth, min_blob_depth] = get_range_blob_depth(self)
            max_blob_depth = 0;
            min_blob_depth = +Inf;
            num_blobs = numel(self.bl);

            etas = self.ref_gr.allblobs.eta(self.ref_gr.ondet(self.ref_gr.included(self.ref_gr.selected)));
            for b_ii = 1:num_blobs
                % 2 is the pading we apply in forward simulation
                current_blob_depth = (size(self.bl(b_ii).intm, 3) - 2);

                lorentz_factored_depth = current_blob_depth * abs(sind(etas(b_ii)));

                max_blob_depth = max(max_blob_depth, lorentz_factored_depth);
                min_blob_depth = min(min_blob_depth, current_blob_depth);
            end
        end
    end

    methods (Access = protected, Static)
        function range = get_cubic_range(edge_num, oversize)
            if (~exist('oversize', 'var'))
                oversize = 1.1;
            end
            limit = (edge_num - 1) / 2;
            range = (-limit:limit) / (limit * 2) * oversize;
        end

        function range = get_bcc_center_range(edge_num, oversize)
            if (~exist('oversize', 'var'))
                oversize = 1.1;
            end
            limit_bcc = (edge_num - 1) / 2;
            limit = (edge_num - 2) / 2;
            range = (-limit:limit) / (limit_bcc * 2) * oversize;
        end

        function grains = selectBlobsFromGrains(grains, included)
            names = fieldnames(grains{1}.allblobs);
            try
                for ii_n = 1:numel(names)
                    name = names{ii_n};
                    if (size(grains{1}.allblobs.(name), 3) > 1)
                        for ii_gr = 1:numel(grains)
                            grains{ii_gr}.allblobs.(name) = grains{ii_gr}.allblobs.(name)(:, :, included);
                        end
                    else
                        for ii_gr = 1:numel(grains)
                            grains{ii_gr}.allblobs.(name) = grains{ii_gr}.allblobs.(name)(included, :);
                        end
                    end
                end
            catch mexc
                fprintf('Error when selecting field: %s\n', names{ii_n})
                fprintf(' Size: [%s ]', sprintf(' %d', size(allblobs.(names{ii_n}))));
                included
                ii_n
                ii_gr
                rethrow(mexc);
            end
        end

        function allblobs = selectBlobsFromGrain(allblobs, included)
            names = fieldnames(allblobs);
            try
                for ii = 1:numel(names)
                    if (size(allblobs.(names{ii}), 3) > 1)
                        allblobs.(names{ii}) = allblobs.(names{ii})(:, :, included);
                    else
                        allblobs.(names{ii}) = allblobs.(names{ii})(included, :);
                    end
                end
            catch mexc
                fprintf('Error when selecting field: %s\n', names{ii})
                fprintf(' Size: [%s ]', sprintf(' %d', size(allblobs.(names{ii}))));
                included
                rethrow(mexc);
            end
        end
    end

    methods (Access = public, Static)
        function [avg_R_vecs, avg_R_vecs_int] = computeAverageOrientations(gr, vols, selected)
            final_vol_size = [size(vols{1}, 1), size(vols{1}, 2), size(vols{1}, 3)];

            avg_R_vecs = zeros([final_vol_size, 3]);
            avg_R_vecs_int = zeros(final_vol_size);

            gr_sel = [gr{selected}];

            %fprintf('Processing: ')
            %c = tic();
            for ii = 1:numel(vols)
                %num_chars = fprintf('%05d/%05d (%f s)', ii, numel(vols), toc(c));

                weights = vols{ii};
                r_vec = gr_sel(ii).R_vector;
                r_vec = cat(4, ...
                    r_vec(1) * ones(final_vol_size), ...
                    r_vec(2) * ones(final_vol_size), ...
                    r_vec(3) * ones(final_vol_size) );

                avg_R_vecs_int = avg_R_vecs_int + weights;
                avg_R_vecs = avg_R_vecs + weights(:, :, :, [1 1 1]) .* r_vec;

                %fprintf(repmat('\b', 1, num_chars))
            end
            %fprintf('%05d Done in %f seconds.\n', numel(vols), toc(c));

            expand_avg_ins = avg_R_vecs_int(:, :, :, [1 1 1]);
            avg_R_vecs = avg_R_vecs ./ (expand_avg_ins + (expand_avg_ins == 0));
        end

        function avg_R_vec = computeAverageOrientation(gr, vols, selected)
            gr_sel = [gr{selected}];
            R_vecs = cat(1, gr_sel(:).R_vector);

            vol_ints = GtOrientationSampling.computeODF(gr, vols, selected);
            vol_ints = reshape(vol_ints, [], 1);
            vol_ints = vol_ints(selected, [1 1 1]);

            avg_R_vec = sum(R_vecs .* vol_ints, 1) ./ sum(vol_ints, 1);
        end

        function odf = computeODF(gr, vols, selected)
            odf = zeros(size(gr));
            odf(selected) = cellfun(@(x)sum(x(:)), vols);
        end
    end
end