function edfheader = edf_info(fname)
% EDF_INFO.M
% provides a structure containing all elements found in the header of the
% specified .edf file.
% Usage:
% edfheader = edf_info(filename)
%   where filename is the name of the EDF file
%
% GJ May 2006
% modified July 2006 to force all parameters to lower case (PyHST generates
% a number of different variations!
%
% Modified by Nicola Vigano, 07/12/2011, nicola.vigano@esrf.fr

  function field_str = sfCleanField(field_str)
    % if header field has weird characters (spaces, parentheses) - this removes
    % them
    field_str( find(field_str == ' ') ) = '_';
    field_str( find(field_str == '-') ) = '_';
    field_str( find(field_str == ',') ) = '_';
    field_str( find(field_str == '(') ) = [];
    field_str( find(field_str == ')') ) = [];
  end

  %% Fields pattern
  pattern = '(?<parameter>.*?)=(?<value>.*);';

  %% Acquire file information
  edfFile = dir(fname);

  %% File Error Checking
  if isempty(edfFile)
    Mexc = MException('EDF:fileNotFound', ['Could not find ' fname]);
    throw(Mexc);
  end

  if edfFile.isdir == true
    Mexc = MException('EDF:fileIsDirectory', [ fname ' is a directory instead of a file.']);
    throw(Mexc);
  end

  if edfFile.bytes == 0
    Mexc = MException('EDF:fileIsEmpty', [ fname ' is an empty file.']);
    throw(Mexc);
  end

  %% Read the header
  fid = fopen(fname, 'rt');

  if fid == -1
    Mexc = MException('EDF:fileNotFound', ['Could not open ' fname]);
    throw(Mexc);
  end

  edfheader = [];
  endofheader = [];
  edfheader.('headerlength') = 0;
  while isempty(endofheader)
    % Split lines
    htxt = fgetl(fid);

    % Check for inconsistencies in lines
    if isnumeric(htxt)
      % Possible empty file! or at least malformed ;-)
      % Close Resource
      close(fid)

      Mexc = MException('EDF:emptyLineInHeader', 'Found an empty line in header.');
      throw(Mexc)
    end

    tmp = regexp(htxt, pattern, 'names');
    if ~isempty(tmp)
      field = sfCleanField(lower(strtrim(tmp.parameter)));

      try
        edfheader.(field) = strtrim(tmp.value);
      catch Mexc
        gtPrintException(Mexc)
        warning('EDF:malformed_field', 'Didn''t understand this header');
        disp(field);
      end
    end

    % Look for end of header reached
    endofheader = find( htxt == '}' );
    % Plus 1 is for the newline
    edfheader.headerlength = edfheader.headerlength + length(htxt) +1;
  end

  % Count the bytes of the header, and check if it is a standard header
  if (mod(edfheader.headerlength, 512) ~= 0)
    warning('EDF:wrong_header_size', ...
            'Header is not multiple of 512 bytes in %s (%d instead)!\n', ...
            fname, edfheader.headerlength);
%     disp('Setting correctly!')
%     % XXX - may be cause of errors in edf_read. Check it!
%     filePosition = round(filePosition/512) * 512;
  end

  % Release Resource
  fclose(fid);

  edfheader = sfCleanupHeaderInfo(edfheader);
end

function edfheader = sfCleanupHeaderInfo(edfheader)

  % convert to numbers any fields that I know should not be strings
  numericFields = { ...
    'dim_1', 'dim_2', 'dim_3', 'size', 'count_time', 'timestamp_ms', ...
    'col_beg', 'col_end', 'row_beg', 'row_end', 'col_bin', 'row_bin', ...
    'energy', 'optic_used', 'acq_frame_nb', 'point_no', 'scan_no', ...
    'run', 'image' ...
    };

  for ii = 1:length(numericFields)
    if isfield(edfheader, numericFields{ii})
      value = edfheader.(numericFields{ii});
      value = str2double(value);
      edfheader.(numericFields{ii}) = value;
    end
  end

  % convert all the motors and counters to individual fields
  coupledFields = { 'motor', 'counter' };

  for ii = 1:length(coupledFields)
    field_pos = [coupledFields{ii} '_pos'];
    field_mne = [coupledFields{ii} '_mne'];

    if isfield(edfheader, field_pos)
      field_tmp = [];
      token_pos = textscan(edfheader.(field_pos), '%f');
      token_name = textscan(edfheader.(field_mne), '%s');
      for jj = 1:length(token_pos{1})
        field_tmp.(token_name{1}{jj}) = token_pos{1}(jj);
      end
      edfheader.(coupledFields{ii}) = field_tmp;
      edfheader = rmfield(edfheader, {field_pos, field_mne} );
    end
  end

  % convert endianness to standard representation
  if strcmpi(edfheader.byteorder,'HighByteFirst')
    edfheader.byteorder='b';
  elseif strcmpi(edfheader.byteorder,'LowByteFirst')
    edfheader.byteorder='l';
  else
    disp('No endian-ness specified');
  end

  % convert datatype to matlab standard representation
  switch lower(edfheader.datatype)
    case {'doublevalue'}
      edfheader.datatype='float64';
    case {'float','floatvalue','real'}
      edfheader.datatype='float32';
    case {'unsigned64'}
      edfheader.datatype='uint64';
    case {'signed64'}
      edfheader.datatype='int64';
    case {'unsignedinteger','unsignedlong'}
      edfheader.datatype='uint32';
    case {'signedinteger','signedlong'}
      edfheader.datatype='int32';
    case {'unsignedshort'}
      edfheader.datatype='uint16';
    case {'signedshort'}
      edfheader.datatype='int16';
    case {'unsignedbyte'}
      edfheader.datatype='uint8';
    case {'signedbyte'}
      edfheader.datatype='int8';

    otherwise
      edfheader.datatype=[];
      warning('EDF:unknown_type', ['Type ' edfheader.datatype ' not known']);
  end
end

