function opt = gtFITParameters(spot, par, grain, parameters)
% GTFITPARAMETERS Fitting (optimisation) of average grain properties and
% global setup parameters.
%
% opt = gtFITParameters(spot, grain, parameters, varargin)
%
% INPUT
%   spot       - diffraction spots data (structure from gtINDEXIndividualSpots)
%   par        - fitting parameters (as from gtFitDefaultPars)
%
% OPTIONAL INPUT
%   grain      - grain data structure (cell array); if undefined, it will
%                be loaded from the index.mat file
%   parameters - as in parameters.mat; if undefined, it will be loaded from
%                the parameters file 
%
% OUTPUT
%   opt  = updated optimised setup, grain and drift parameters
%    .labgeo   - updated labgeo
%    .energy   - updated energy (keV)
%	 .grainnew - updated grain structure
%	 .drift    - sample and energy drifts
%	 .resini   - initial residuals
%	 .resend   - optimised residuals 
%	 .resnorm  - optimised norm of all residuals
%	 .jacobian - Jacobian matrix (sparse matrix)
%
%
% Currently only handles a single phase. Hydrostatic strain (lattice 
% expansion) is not listed as an option, but can be fitted via the beam
% energy.
%
% In case of multiphase materials, fit global parameters first on the 
% phase of which lattice parameters are known most precisely (e.g.
% a stochiometric phase or has low solubility of alloying elements).
%
% What cannot be fitted simultaneously:
%   1. detposZ  &  grain position
% 	   (Or another detpos variable parallel with the rotation axis.)
% 	   Translation along the rotation axis would be a free, arbitrary
%      variable.
% 
%   2. mean pixel size  &  grain position
%      The size of the entire setup would be scalable whilst still 
%      satisfying all constraints.
%      
%   3. beam energy  &  grain strain
% 	   The beam energy and hydrostatic strain have equivalent effects via
%      Bragg's law.
%
%   4. drifts, if scan was interlaced
%      In an interlaced each diffraction spot is combined from images 
%      recorded at several different omega rotational positions, thus
%      the spot centroid may have errors from drifts at various times.
% 
%
% Default parameters: par = gtFitDefaultPars();
%
% Which spots the different parameter types affect:
%   global parameters: affect all spot positions
%   grain parameters:  affect only the grain's spot positions
%   drift parameters:  affect spot positions in a limited omega range
%
%
% Version 001 18-04-2014 by P.Reischig
%


% Notes on nomenclature:
%   fitted parameters: optimum setup parameters and grain properties to be found      
%   target parameters: components of the optimisation target function     


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Prepare input
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% if (nargin>3) && strcmp(varargin{1},'par')
%     par = varargin{2};
%     varargin(1:2) = [];
%     par.fit_drift_allpos = false;
%     par.fit_drift_allrot = false;
% else
%     par = gtFitDefaultPars();
%     par.fit_drift_allpos = false;
%     par.fit_drift_allrot = false;
% end
% par = parse_pv_pairs(par, varargin);
% disp('Fitting parameters used:')
% disp(par)


if isempty(grain)
    fname = sprintf('4_grains/phase_%02d/index.mat', par.phase);
    fprintf('Loading grain info from file %s\n', fname)
    index = load(fname);
    grain = index.grain;
end

if isempty(par.cryst)
    fname = sprintf('4_grains/phase_%02d/index.mat', par.phase);
    fprintf('Loading crystallographic info from file %s\n', fname)
    index = load(fname);
    par.cryst = index.cryst;
end

par.latticepar = par.cryst.latticepar;
par.Bmat       = gtCrystHKL2CartesianMatrix(par.latticepar);


if isempty(parameters)
    disp('Loading parameters file...')
    parameters = load('parameters.mat');
    parameters = parameters.parameters;
end

if isempty(par.energy)
    disp('Using initial beam energy from the parameters file.')
    par.energy = parameters.acq.energy;
end


if isempty(par.labgeoini)
    disp('Using initial labgeo parameters from the parameters file.')
    par.labgeoini = parameters.labgeo;
end

par.labgeoini.detorig = gtGeoDetOrig(par.labgeoini);
par.labgeoini.detnorm = cross(par.labgeoini.detdiru, par.labgeoini.detdirv);

par.wavelength = gtConvEnergyToWavelength(par.energy);
par.samgeo     = parameters.samgeo;
par.omstep     = 180/parameters.acq.nproj;
par.numimages  = 2*parameters.acq.nproj;


% Check available Optimization Toolbox License
ok = license('checkout','Optimization_Toolbox');
if ~ok
    disp('Could not check out Optimization Toolbox license.')
    return
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check fitting options
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (par.fit_global_detposZ & par.fit_grain_position)
	disp('If the rotation axis is not parallel with Z, you can ignore the following warning.')
	disp('Note: translation along the rotation axis would be a free, arbitrary variable.')
	error('Cannot fit both the detector position along the rotation axis and the grain positions.')
end

if (par.fit_global_pixmean & par.fit_grain_position)
	disp('Note: with the current settings the size of the entire setup would be scalable.') 
	error('Cannot fit both the mean pixel size and the grain positions.')
end

if (par.fit_global_energy & par.fit_grain_strain)
	disp('Note: the beam energy and hydrostatic strain have equivalent effects.') 
	error('Cannot fit both the beam energy and grain strain.')
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Global variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% old variables: all geometry variables that could be fitted and their
%                initial value
% new variables: global variables that are actually being fitted

% Order of fitted parameters in newvars and oldvars:
fitname{1}  = 'Position X';
fitname{2}  = 'Position Y';
fitname{3}  = 'Position Z';
fitname{4}  = 'Tilt in plane';    % rad
fitname{5}  = 'Angle U-V';        % rad
fitname{6}  = 'Tilt around U';    % rad
fitname{7}  = 'Tilt around V';    % rad
fitname{8}  = 'Pixel size mean';
fitname{9}  = 'Pixel size ratio';
fitname{10} = 'Energy';
fitname{11} = 'Rot wedge angle';



% Indices of fitted labgeo parameters
activeglobal = false(11,1);
if par.fit_global_detposX;  activeglobal(1)  = true; end
if par.fit_global_detposY;  activeglobal(2)  = true; end
if par.fit_global_detposZ;  activeglobal(3)  = true; end
if par.fit_global_detroll;  activeglobal(4)  = true; end
if par.fit_global_detangUV; activeglobal(5)  = true; end
if par.fit_global_dettiltU; activeglobal(6)  = true; end
if par.fit_global_dettiltV; activeglobal(7)  = true; end
if par.fit_global_pixmean;  activeglobal(8)  = true; end
if par.fit_global_pixratio; activeglobal(9)  = true; end
if par.fit_global_energy;   activeglobal(10) = true; end
if par.fit_global_rotwedge; activeglobal(11) = true; end

fitnameused = fitname(activeglobal);

disp('Fitted acquisition parameters:')
disp(fitnameused')
                   
% Number of global fitted parameters
par.nvarglobalall = length(activeglobal);
par.nvarglobal    = sum(activeglobal);

tStart = tic;

%h_fitpars = figure('name','Fitted parameters','Units','normalized','Position',[0.6 0.6 0.4 0.4]);

strainT     = [];
strainTmean = [];
residual{2} = [];

%figure('Name','Residuals after final fit (excl. outliers)','Units','normalized','Position',[0.6 0.6 0.4 0.4])

disp('Optimised parameters:')
disp(fitnameused)
%disp(varsopt(:,1:sum(newind)))
disp(' ')

%figure(h_fitpars);
%plot(varsopt,'.-')
%legend(fitnameused,'Location','best')



% Old global variables: the starting point for fitting
% Geometry
iniglobal    = zeros(11,1);
iniglobal(1) = par.labgeoini.detrefpos(1);
iniglobal(2) = par.labgeoini.detrefpos(2);
iniglobal(3) = par.labgeoini.detrefpos(3);
iniglobal(4) = 0;
iniglobal(5) = acos(par.labgeoini.detdiru * par.labgeoini.detdirv');
iniglobal(6) = 0;
iniglobal(7) = 0;
iniglobal(8) = 1000*0.5*(par.labgeoini.pixelsizeu + par.labgeoini.pixelsizev);
iniglobal(9) = par.labgeoini.pixelsizeu/par.labgeoini.pixelsizev;
iniglobal(10)= par.energy;
iniglobal(11)= 0;

% Initial guesses for active fitted global variables
guess = iniglobal(activeglobal);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Drift variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Drift data is stored in 'drift'. Active Those variables that are not 
%   drift.om   - omega rotation angle; input points for interpolation 
%   drift.ipol - interpolation method
%   drift.nfit - no. of fitted points; normally length(omega)-1 because 
%                the initial point is the reference which is not fitted
% 

% Parameters for interpolation
drift.omega  = par.drift_omega;
drift.method = par.drift_method;
drift.nfit   = max(length(par.drift_omega) - 1, 0);
numom        = length(par.drift_omega);


% Setting active variables

% if par.fit_drift_allpos
%     par.fit_drift_posX = true;
%     par.fit_drift_posY = true;
%     par.fit_drift_posZ = true;
% end
% 
% if par.fit_drift_allrot
%     par.fit_drift_rotX = true;
%     par.fit_drift_rotY = true;
%     par.fit_drift_rotZ = true;
% end


% Active variable indices
nind = par.nvarglobal + 1;

if par.fit_drift_posX 
    drift.posx             = zeros(1,numom);
    par.ind.fit_drift_posX = nind : (nind + drift.nfit - 1);
    nind                   = nind + drift.nfit;
else
    drift.posx             = [];
    par.ind.fit_drift_posX = [];
end

if par.fit_drift_posY
    drift.posy             = zeros(1,numom);
    par.ind.fit_drift_posY = nind : (nind + drift.nfit - 1);
    nind                   = nind + drift.nfit;
else
    drift.posy             = [];
    par.ind.fit_drift_posY = [];
end

if par.fit_drift_posZ
    drift.posz             = zeros(1,numom);
    par.ind.fit_drift_posZ = nind : (nind + drift.nfit - 1);
    nind                   = nind + drift.nfit;
else
    drift.posz             = [];
    par.ind.fit_drift_posZ = [];
end

if par.fit_drift_rotX
    drift.rotx             = zeros(1,numom);
    par.ind.fit_drift_rotX = nind : (nind + drift.nfit - 1);
    nind                   = nind + drift.nfit;
else
    drift.rotx             = [];
    par.ind.fit_drift_rotX = [];
end

if par.fit_drift_rotY
    drift.roty             = zeros(1,numom);
    par.ind.fit_drift_rotY = nind : (nind + drift.nfit - 1);
    nind                   = nind + drift.nfit;
else
    drift.roty             = [];
    par.ind.fit_drift_rotY = [];
end

if par.fit_drift_rotZ
    drift.rotz             = zeros(1,numom);
    par.ind.fit_drift_rotZ = nind : (nind + drift.nfit - 1);
    nind                   = nind + drift.nfit;
else
    drift.rotz             = [];
    par.ind.fit_drift_rotZ = [];
end

if par.fit_drift_energy
    drift.energy             = par.energy(1, ones(1,numom));
    par.ind.fit_drift_energy = nind : (nind + drift.nfit - 1);
    nind                     = nind + drift.nfit;
else
    drift.energy             = [];
    par.ind.fit_drift_energy = [];
end


par.nvardrift = drift.nfit * (par.fit_drift_energy + ...
    par.fit_drift_posX + par.fit_drift_posY + par.fit_drift_posZ + ...
    par.fit_drift_rotX + par.fit_drift_rotY + par.fit_drift_rotZ);


% Add initial guesses for fitted drift variables
iniglobal = [iniglobal; drift.posx(2:end)'];
iniglobal = [iniglobal; drift.posy(2:end)'];
iniglobal = [iniglobal; drift.posz(2:end)'];
iniglobal = [iniglobal; drift.rotx(2:end)'];
iniglobal = [iniglobal; drift.roty(2:end)'];
iniglobal = [iniglobal; drift.rotz(2:end)'];
iniglobal = [iniglobal; drift.energy(2:end)'];

guess = [guess; iniglobal((par.nvarglobalall+1):end)];

par.iniglobal = iniglobal;
par.drift     = drift;


%%%%%%%%%%%%%%%%%%%%%%%%
% Grain variables
%%%%%%%%%%%%%%%%%%%%%%%%

% Reorganised and checked grain data in a structure array 'gr';
% grain parameters added to guess
[gr, spotgr, guess] = sfGrainInfo(grain, spot, guess, par);
%guess = [guess; gr.center; gr.R_vector; gr.strcmp];
par.nvargrain = length(guess) - par.nvarglobal - par.nvardrift;

% Jacobian pattern: a sparse matrix showing dependencies of target parameters
% from fitted parameters
fprintf('No. of spots input for fitting; %d\n', length(spot.omind))
disp('Creating Jacobian pattern...')
tic
JacPatt = gtFitJacobianPattern(guess, spotgr, gr, par);
toc

% Initial value of target function
duvw   = gtFitSpotPosDeviations(guess, activeglobal, spotgr, gr, par);
resini = duvw;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Run fitting
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Matlab otimisation settings
optset = optimset(...
   'Algorithm', par.algorithm, 'FunValCheck','off', ...
   'Display','iter', 'Diagnostics','off', 'Jacobian','off', ...
   'DiffMinChange',par.finitediff, 'DiffMaxChange',par.finitediff, 'MaxIter',par.maxiter, ...
    'TolFun', par.tolfun, 'TolX',par.tolvar, 'MaxFunEvals',par.maxfunevals, ...
   'FinDiffType','forward', 'PlotFcns',par.plotfun);
   
% Jacobian pattern
optset.JacobPattern = JacPatt;

% Using Matlab's built-in non-linear least square solver lsqnonlin 
tic
[varsopt, resnorm, resend, exitflag, optimout, lagrange, jacobian] = ...
    lsqnonlin(@(newvars) gtFitSpotPosDeviations(newvars, activeglobal, spotgr, gr, par),...
    guess, [], [], optset);
toc


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Update fitted variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Grains
grainvec.center  = vertcat(gr(:).center)';
grainvec.Rvec    = vertcat(gr(:).Rvec)';
grainvec.strain  = zeros(6, length(gr));  % vertcat(gr(:).strain)';
grainvec.origind = zeros(1, length(gr));

for ii = 1:length(gr)
    if par.fit_grain_position
        grainvec.center(:,ii) = varsopt(gr(ii).indcenter);
    end
    if par.fit_grain_orientation
        grainvec.Rvec(:,ii)   = varsopt(gr(ii).indRvec);
    end
    if par.fit_grain_strain
		if ~isempty(gr(ii).indstrain)
			grainvec.strain(:,ii) = varsopt(gr(ii).indstrain);
		end
	end
	grainvec.origind(ii) = gr(ii).origind;
end


for ii = 1:length(grain)
    grain{ii}.opt.center   = grain{ii}.center;
	grain{ii}.opt.Rvec     = grain{ii}.R_vector;
    grain{ii}.opt.strain   = zeros(1,6);
	grain{ii}.opt.strainT  = zeros(3,3);
	grain{ii}.opt.coplanar = false;
end

for ii = 1:length(gr)
    if par.fit_grain_position
        grain{gr(ii).origind}.opt.center = varsopt(gr(ii).indcenter)';
    end
    if par.fit_grain_orientation
        grain{gr(ii).origind}.opt.Rvec = varsopt(gr(ii).indRvec)';
    end
    if par.fit_grain_strain
		if isempty(gr(ii).indstrain)
			grain{gr(ii).origind}.opt.coplanar = true;
		else
			st = varsopt(gr(ii).indstrain)';
			grain{gr(ii).origind}.opt.strain  = st;
			grain{gr(ii).origind}.opt.strainT = [st(1) st(6) st(5); ...
				                                 st(6) st(2) st(4); ...
				                                 st(5) st(4) st(3)];
		end
	end
end


% Global
if any(activeglobal)
    [labgeo, energy] = gtFitUpdateLabgeo(varsopt(1:sum(activeglobal)), ...
                       activeglobal, par.iniglobal, par.labgeoini);
    labgeo = rmfield(labgeo,'detnorm');
    labgeo = rmfield(labgeo,'detorig');
else
    labgeo = par.labgeoini;
    energy = par.iniglobal(10);
end


% Drifts
drift = par.drift;
if ~isempty(par.ind.fit_drift_posX); drift.posx = [0, varsopt(par.ind.fit_drift_posX)']; end
if ~isempty(par.ind.fit_drift_posY); drift.posy = [0, varsopt(par.ind.fit_drift_posY)']; end
if ~isempty(par.ind.fit_drift_posZ); drift.posz = [0, varsopt(par.ind.fit_drift_posZ)']; end
if ~isempty(par.ind.fit_drift_rotX); drift.rotx = [0, varsopt(par.ind.fit_drift_rotX)']; end
if ~isempty(par.ind.fit_drift_rotY); drift.roty = [0, varsopt(par.ind.fit_drift_rotY)']; end
if ~isempty(par.ind.fit_drift_rotZ); drift.rotz = [0, varsopt(par.ind.fit_drift_rotZ)']; end
if ~isempty(par.ind.fit_drift_energy); drift.energy = [0, varsopt(par.ind.fit_drift_energy)']; end



% Plot strain
meanstrain = mean(grainvec.strain, 2);

if ~isempty(strainTmean)
    figure('name','Average normal strain components','Units','normalized','Position',[0 0.6 0.4 0.4])
    hold on
    %plot(squeeze(strainTmean(1,1,:)),'-r.')
    %plot(squeeze(strainTmean(2,2,:)),'-g.')
    %plot(squeeze(strainTmean(3,3,:)),'-b.')
    plot(meanstrain(1),'-r.')
    plot(meanstrain(2),'-g.')
    plot(meanstrain(3),'-b.')
    
    figure('name','Average shear strain components','Units','normalized','Position',[0 0.6 0.4 0.4])
    hold on
    %plot(squeeze(strainTmean(2,3,:)),'-r.')
    %plot(squeeze(strainTmean(1,3,:)),'-g.')
    %plot(squeeze(strainTmean(1,2,:)),'-b.')
    plot(meanstrain(4),'-r.')
    plot(meanstrain(5),'-g.')
    plot(meanstrain(6),'-b.') 
end


% figure('name','Jacobian components')
% for ii = 1:size(varsopt,2)
%     subplot(2,size(varsopt,2),ii)
%     hist(jacobian{1}(:,ii),100)
%     subplot(2,size(varsopt,2),size(varsopt,2)+ii)
%     hist(jacobian{end}(:,ii),100)    
% end

opt.labgeo      = labgeo;
opt.energy      = energy;
opt.grain       = grain;
opt.grainvec    = grainvec;
opt.drift       = drift;
opt.duvwini     = reshape(resini, [], 3);
opt.duvwend     = reshape(resend, [], 3);
opt.duvwendnorm = resnorm;
opt.jacobian    = jacobian;
opt.par         = par;


disp(' ')
disp('Initial std of all spot coordinates (u,v,w, all):')
disp([std(opt.duvwini) std(opt.duvwini(:))])
disp('Final std of all spot coordinates (u,v,w, all):')
disp([std(opt.duvwend) std(opt.duvwend(:))])

sfPlotRes(opt)

fprintf('\nTotal elapsed time (sec): %g\n', toc(tStart))
beep, pause(0.5), beep, pause(0.5), beep


end % of function




%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SUB-FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [gr, spotgr, guess] = sfGrainInfo(grain, spot, guess, par)
% Recreate grain info of active grains in concise form    
    
    gr = struct('center',{},'Rvec',{},'strain',{},'numspots',{},'coplanar',{},'plref',{},...
                'plsam',{},'dspref',{},'indcenter',{},'indRvec',{},...
                'indstrain',{},'mesu',{},'mesv',{},'mesw',{},...
                'simu',{},'simv',{},'simw',{});

    % Loop through all grains, gather data of active grains
    gi    = 1;
    spind = false(length(spot.grainind),1); 
    
    for ii = 1:length(grain)
        
        if ~grain{ii}.active
            continue
        end
        
        ind0 = length(guess);
        
        gr(gi).origind = ii;
        
        gr(gi).numspots = length(grain{ii}.spot.difspotid);
        gr(gi).center = grain{ii}.center;
        gr(gi).Rvec   = grain{ii}.R_vector;        
        gr(gi).plref  = grain{ii}.spot.plref;
        gr(gi).dspref = grain{ii}.spot.dspref;
        gr(gi).plsam  = grain{ii}.spot.plsam0;

        gr(gi).simu   = grain{ii}.spot.simu;
        gr(gi).simv   = grain{ii}.spot.simv;
        gr(gi).simw   = grain{ii}.spot.simw;

        gr(gi).mesu   = grain{ii}.spot.mesu;
        gr(gi).mesv   = grain{ii}.spot.mesv;
        gr(gi).mesw   = grain{ii}.spot.mesw;

        gr(gi).indspot = grain{ii}.spot.spotind; 
        
        spind(grain{ii}.spot.spotind) = true;
        
        if par.fit_grain_position
            gr(gi).indcenter = (ind0+1) : (ind0+3);
            guess = [guess; gr(gi).center'];
            ind0  = length(guess);
        end
        
        if par.fit_grain_orientation
            gr(gi).indRvec = (ind0+1) : (ind0+3);
            guess = [guess; gr(gi).Rvec'];
            ind0  = length(guess);      
        end
        
        if par.fit_grain_strain       
            % Check that plane normals are not all in one plane (determinant is
            % significantly larger than 0 or some tolerance)
            detlim = 1;
            pl     = grain{ii}.spot.plref;
            detpl  = det(pl*pl');
            
            if (detpl >= detlim)
                % Plane normals are not coplanar
                gr(gi).strain    = [0, 0, 0, 0, 0, 0];
                gr(gi).coplanar  = false;
                gr(gi).indstrain = (ind0+1) : (ind0+6);
                guess            = [guess; zeros(6,1)];
            else
                % Plane normals are coplanar
                gr(gi).strain    = [];
                gr(gi).coplanar  = true;
                gr(gi).indstrain = [];
            end
        end
        
        gi = gi + 1;
    end
    
    % Keep only those spots which belong to one of the active grains
    fs = fieldnames(spot);
    for ii = 1:length(fs)
        spotgr.(fs{ii}) = spot.(fs{ii})(spind,:);     
    end

end


function sfPlotRes(opt)

hf = figure('name', 'Histograms of initial and optimised errors of spot positions');

h1 = subplot(1,3,1);
[ne, be] = hist(opt.duvwend(:,1), 100);
ni = hist(opt.duvwini(:,1), be);
bar(be,ne)
hold on
plot(be, ni, 'r-', 'LineWidth',2)
title(h1, 'U coordinates', 'FontWeight','bold', 'FontSize',14)

h2 = subplot(1,3,2);
[ne, be] = hist(opt.duvwend(:,2), 100);
ni = hist(opt.duvwini(:,2), be);
bar(be, ne)
hold on
plot(be, ni, 'r-', 'LineWidth',2)
title(h2, 'V coordinates', 'FontWeight','bold', 'FontSize',14)

h3 = subplot(1,3,3);
[ne, be] = hist(opt.duvwend(:,3), 100);
ni = hist(opt.duvwini(:,3), be);
bar(be, ne)
hold on
plot(be, ni, 'r-', 'LineWidth',2)
title(h3, 'W coordinates', 'FontWeight','bold', 'FontSize',14)

legend({'Optimised','Initial'})

set(hf, 'Units','normalized', 'Position',[0 0.3 1 0.6])

end % sfPlotRes


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Unused
%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Typical magnitudes to scale finite differences for computing derivatives
% typ    = zeros(11,1);
% typ(1) = par.typ_detrefpos;
% typ(2) = par.typ_detrefpos;
% typ(3) = par.typ_detrefpos;
% typ(4) = par.typ_dettilt;
% typ(5) = par.typ_detangUV;
% typ(6) = par.typ_dettilt;
% typ(7) = par.typ_dettilt;
% typ(8) = par.typ_pixmean;
% typ(9) = par.typ_pixratio;
% typ(10)= par.typ_energy;
% typ(11)= par.typ_rotwedge;
% typ    = typ(activeglobal);

