function out = gtForwardSimulate_v2(first, last, workingdirectory, phaseID, varargin)
% GTFORWARDSIMULATE_V2  Finds missing diffraction spots and updates grain structure
%     out = gtForwardSimulate_v2(first, last, workingdirectory, [phaseID], varargin)
%     ------------------------------------------------------------------------------
%
%     INPUT:
%       first            = <int>     first grainid
%       last             = <int>     last grainid
%       workingdirectory = <string>  working directory
%       phaseID          = <int>     phase ID
%       varargin         =           parameters.fsim plus some additional
%                                    options as in gtAppFsimExtraPars():
%
%     OPTIONAL INPUT (p/v pairs):
%       updateSample     = <logical> true/false to update sample.mat and
%                                    spot2grain.mat {true}
%       updateDBtables   = <logical> true/false to update DB table with
%                                    'grainID', 'phaseID', 'flag' {false}
%       area_factor      = <double>  Weight for area criterion {1}
%       position_factor  = <double>  Weight for UV position criterion {1}
%       image_factor     = <double>  Weight for image criterion {1}
%       int_factor       = <double>  Weight for intensity criterion {0}
%
%     OPTIONAL INPUT for twin variants calculation (p/v pairs):
%       variants         = <logical> Use fsim relative to twin variants
%                                    {false}
%       plot_variants    = <logical> Plot the variants orientation wrt to
%                                    the parent orientation as XYZ crystal
%                                    axes {false}
%       num_variant      = <double>  Variant number to forward simulate
%                                    {[]}
%       twin_angle       = <double>  twin angle {[]}
%       twin_axis        = <double>  twin axis {[]}
%       twin_center      = <double>  alternative center for the twin (mm)
%                                    or the twinID if available from the
%                                    same indexed dataset {[]}
%       twin_grain       = <struct>  twin grain structure for common spot
%                                    properties and full image {[]}
%
%     OUTPUT:
%       out  = <struct>   updated grain structure. Fields:
%          .bl             = <struct N> stack of normalized difblobs, along with metadata
%          .proj.geom      = <double>   input for grain reconstruction
%                                        / forward projection using ROI
%                                        (hsize, vsize) (Nx12)
%          .proj.geom_full = <double>   input for grain reconstruction
%                                        / forward projection (full images) (Nx12)
%          .proj.geom_abs  = <double>   input for grain reconstruction
%                                        / forward projection of extinction spots (Nx12)
%          .proj.difstack  = <double>   stack of normalized difspots (XxYxN)
%          .ondet          = <double>   indices of those spots (from the list in allblobs)
%                                       which are expected on the detector
%          .uvw            = <double>   spot positions on the detector (Nx3)
%          .hklsp          = <double>   signed hk(i)l of reflection (Nx3(4))
%          .hkl            = <double>   hk(i)l family for each reflection (Nx3(4))
%          .flag           = <int>      status flag indicating result of forward simulation
%          .allblobs       = <struct>   forward simulation output produced by gtIndexFwdSimGrains
%          .used_fam       = <double>   List of unique HKL families used
%                                       (theoretical)
%          .used_famsp     = <double>   List of all the HKL families used
%                                       (theoretical)
%
%       We will classify spots with a flag  [1...6]:
%        - 1 : No intensity at expected spot position
%        - 2 : Found intensity in FULL images at predicted position
%        - 3 : Found segmented difspot at expected position - but position
%              / size do not match strict criteria
%        - 4 : Found segmented difspot in database matching the strict
%              search criteria - we will use theoretically predicted
%              angles for reconstruction
%        - 5 : Difspot is part of a pair identified by INDEXTER - use
%              Friedelpair information for reconstruction
%        - 6 : Conflict difspot (if existing)
%
%     Version 010 21-07-2014 by LNervo
%       Extracted the ASTRA geometry and the gtFwdSimFindCandidates routines

%     Version 009 14-05-2014 by LNervo
%       Added check of used hkl families
%
%     Version 008 25-04-2014 by LNervo
%       Adapted to use twin forward simulation
%
%     Version 007 Beginning-2014 by LNervo
%       Modulized, simplified, added application parameters to update
%       sample.mat and DB difspot table
%
%     Version 006 18-12-2013 by LNervo
%       Moduling a bit more...
%
%     Version 005 05/03/2013 by Nicola Vigano (nicola.vigano@esrf.fr)
%       Re-structured completely, to be more clear and easy to debug.
%
%     Version 004 21/06/2012 by AKing
%       Trivial change to make segbb zeros(1, 6)
%
%     Version 003 14-06-2012 by LNervo
%       Use gtFsimDefaultParameters


if (isdeployed)
    global GT_DB
    global GT_MATLAB_HOME
    load('workspaceGlobal.mat');
end

cd(workingdirectory);
gtDBConnect();

% extra parameters
[par, vars_list] = gtAppFsimExtraPars();

parameters = [];
load('parameters.mat');
labgeo = parameters.labgeo;
samgeo = parameters.samgeo;
recgeo = parameters.recgeo;

if ~isfield(parameters.fsim, 'save_grain')
    parameters.fsim.save_grain = true;
end
if ~exist('phaseID', 'var') || isempty(phaseID)
    phaseID = 1;
end

if (isdeployed)
    first   = str2double(first);
    last    = str2double(last);
    phaseID = str2double(phaseID);
    parameters.fsim.display_figure = false;
elseif ~isempty(varargin)
    % in interactive mode we can modify parameters
    [parameters.fsim, rej_pars] = parse_pv_pairs(parameters.fsim, varargin);
end
if ~isempty(varargin)
    [par, rej_pars] = parse_pv_pairs(par, varargin);
elseif (~exist('rej_pars', 'var'))
    rej_pars = {};
end

if (~parameters.fsim.save_grain)
    par.updateSample   = false;
    par.updateDBtables = false;
end

indexgrainModel = fullfile(parameters.acq.dir, '4_grains', 'phase_%02d', 'index.mat');
fprintf('\nLoading indexing results...\n');
grain = [];
load(sprintf(indexgrainModel, phaseID), 'grain');
totGrains     = length(grain);

if (last > totGrains)
    warning('gtForwardSimulate_v2:wrong_parameter', ...
        'Last was bigger than the total number of grains, setting it to max')
    last = totGrains;
end

difspotfile = fullfile(parameters.acq.dir, '2_difspot', '00000', 'difspot00001.edf');
if exist(difspotfile, 'file');
    info = edf_info(difspotfile);
else
    info = [];
end

%%% this section should be replaced with a function dealing with different
% versions of parameters names:  fed, v1, v2...

% !!! using a mean pixel size is uncertain; ideally fwd simulation should
% use mm-s

% detector reference position in pixels
detrefpos_p = gtGeoLab2Sam(labgeo.detrefpos, eye(3), labgeo, recgeo, true);

fprintf('\nLoading reflections from database..');
% Initialization of spot table and info about pairs
segmentedSpots = gtDBLoadTable([parameters.acq.name 'difspot'], 'difspotID');
pairsTable     = gtDBLoadTable(parameters.acq.pair_tablename, 'pairID');
fprintf('(Done).\n')

difspot_num = numel(segmentedSpots.difspotID); % number of difspots in database
spot2grain  = cell(difspot_num, 1); % cell array linking difspots with grainids
spot2phase  = cell(difspot_num, 1);

filetable   = [parameters.acq.name '_filetable'];
filename    = fullfile(parameters.acq.dir, '4_grains', 'spot2grain.mat');
spotTable   = GtMergeTable(filetable, filename, {'spot2grain', 'spot2phase'});

samplefile  = fullfile(parameters.acq.dir, '4_grains', 'sample.mat');
sample      = GtSample.loadFromLockedFile(filetable, samplefile);

% initialise sample structure
phase = GtPhase(parameters.cryst(phaseID).name, totGrains);
sample.phases{phaseID} = phase;

% initialize output
out = [];

spotID = cell(1, length(first:last));

if (parameters.fsim.verbose)
    disp('Extra parameters:')
    disp(par)
    print_structure(parameters.fsim, 'parameters.fsim', false, true)
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for n = first : last
    fprintf('\nRunning forward simulation for grain %d: \n - Preparing pair tables..', n);

    gr = grain{n}; % output from INDEXTER

    %%% Predict spot positions
    gr = gtCalculateGrain(gr, parameters);
    uvw = gr.allblobs.uvw;

    %%% Get:
    % - average difspot X and Y sizes for the current spots in the grain
    % - theshold values used during the database search
    % - offset in possible lateral grain orientation
    % - diffstack Vertical and Horizontal sizes
    spotsCommProps = gtFwdSimGetDiffspotsCommonProperties(gr, detrefpos_p, parameters);
    % radial distances of spot center from detector center
    % !!! to be corrected: detrefu and detrefv can be anywhere, not
    % necessarily in the center
    radiuses = sqrt((uvw(:, 1) - labgeo.detrefu).^2 + (uvw(:, 2) - labgeo.detrefv).^2);

    %%% Get the indices of all spots which fully intersect the active
    % region of the detector
    onDet = gtFwdSimFindSpotsOnDetector(spotsCommProps, uvw(:, 1), uvw(:, 2), radiuses, parameters.acq, parameters.seg.bbox);
    uvw = uvw(onDet, :);

    % number of exspected diffraction spots
    numspots = numel(onDet);

    flag      = ones(numspots, 1); % see above
    % stack of zero padded diffraction spots (projections used for 3D reconstruction)
    bl = struct( ...
        'intm', cell(numspots, 1), ...
        'mask', cell(numspots, 1), ...
        'bbuim', cell(numspots, 1), ...
        'bbvim', cell(numspots, 1), ...
        'bbwim', cell(numspots, 1), ...
        'bbsize', cell(numspots, 1) );
    checkSpots = struct('info', cell(numspots, 1));

    bb        = zeros(numspots, 4); % difspot BoundingBox from database
    cent      = zeros(numspots, 1); % centroid distance from database added 06-02-2014 LNervo
    spotid    = zeros(numspots, 1); % Mysql difspottable difspotID
    check     = false(numspots, 2);
    cands     = struct('info', cell(numspots, 1));
    intensity = zeros(numspots, 1); % Integrated blob intensity

    % Diffraction vectors directions
    diff_beam_dirs  = zeros(numspots, 3);

    U = zeros(numspots, 1); % COM u positions of difspots from difspottable
    V = zeros(numspots, 1); % COM v positions of difspots from difspottable
    W = zeros(numspots, 1); % COM w positions of difspots from difspottable

    full = zeros(parameters.acq.ydet, parameters.acq.xdet);

    tempIsSpotA = false(numspots, 1);

    fprintf(' (Done).\n - Matching reflections and spots..');

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Forward simulation:
    % We try to match the reflections with the segmented spots, so we
    % examine all hklsp and search corresponding spots / intensity on the
    % detector..
    for ii = 1:numspots
        % Look for difspots appying loose search criteria - is there any intensity at the predicted spot
        % position ?  Choose the one which is closest to the expected size
        [candidateIDs, spotsCands] = gtFwdSimFindCandidates(spotsCommProps, uvw(ii, :), segmentedSpots);

        if isempty(candidateIDs)
            % NO CANDIDATES FOUND:
            %   segmentation / overlap / intensity / ... problem ?
            % If requested, we will check into the raw images later.
            continue;
        end

        Us    = spotsCands.CentroidX;
        Vs    = spotsCands.CentroidY;
        Ws    = spotsCands.CentroidImage;
        BBs   = [...
            spotsCands.BoundingBoxXorigin, spotsCands.BoundingBoxYorigin, ...
            spotsCands.BoundingBoxXsize, spotsCands.BoundingBoxYsize ];
        CENTs = sqrt( (Us -  uvw(ii, 1)).^2 + (Vs - uvw(ii, 2)).^2 );
        INTs  = abs(spotsCands.Integral) ./ gr.stat.intmean;

        % in case there are several candidates, take the one which fits best a goodness criterion
        % with different weight fractions for position, size and rotation
        % angle (image_number).  We could add intensity as another criterium
        area_diff  = abs(BBs(:, 3) .* BBs(:, 4) - spotsCommProps.Xsize * spotsCommProps.Ysize)...
            ./ (spotsCommProps.Xsize * spotsCommProps.Ysize);
        image_diff = abs(Ws - uvw(ii, 3))...
            ./ (parameters.fsim.omegarange / gtGetOmegaStepDeg(parameters));
        pos_diff   = CENTs...
            ./ (spotsCommProps.Offset); % / (parameters.fsim.bbsize_factor));
        int_diff   = INTs ./ gr.stat.intrat;

        [~, area_index]  = sort(area_diff);
        [~, image_index] = sort(image_diff);
        [~, pos_index]   = sort(pos_diff);
        [~, int_index]   = sort(int_diff);

        % in case there are several candidates, take the one which fits best a goodness criterion
        % with different weight fractions for position, size and rotation
        % angle (image_number).  We could add intensity as another criterium
        goodness = par.area_factor .* area_diff + ...
            par.image_factor .* image_diff + ...
            par.position_factor .* pos_diff + ...
            par.int_factor .* int_diff;
        [~, index] = sort(goodness);

        if (length(candidateIDs) > 1)
            all_index = [area_index, image_index, pos_index, int_index];
            [a, ~] = hist(all_index(1, :), length(area_index));
            cand_index2 = all_index(find(a == max(a), 1));

            cand_index = index(1);
        else
            % only one candidate
            cand_index = 1;
        end

        cands(ii).info = sfGetTableIndexes(...
            rmfield(spotsCands, {'thr', 'diff'}), index);
        cands(ii).thr  = spotsCands.thr;
        cands(ii).diff = spotsCands.diff;
        cands(ii).diff.area     = area_diff;
        cands(ii).diff.image    = image_diff;
        cands(ii).diff.pos      = pos_diff;
        cands(ii).diff.int      = int_diff;
        cands(ii).goodness      = goodness;
        cands(ii).factor.area   = par.area_factor;
        cands(ii).factor.image  = par.image_factor;
        cands(ii).factor.pos    = par.position_factor;
        cands(ii).factor.int    = par.int_factor;
        cands(ii).cand_index    = cand_index;
        if length(candidateIDs) > 1
            cands(ii).cand_index2 = cand_index2;
        end

        % take the one which satifies more criteria
        U(ii)       = Us(cand_index);
        V(ii)       = Vs(cand_index);
        W(ii)       = Ws(cand_index);
        bb(ii, :)   = BBs(cand_index, :);
        cent(ii, :) = CENTs(cand_index, :);
        intensity(ii) = INTs(cand_index);

        matchedID   = candidateIDs(cand_index);

        % Found a segmented difspot at expected position - we will check if it does match strict criteria
        uv_ok = cent(ii, :) <= spotsCommProps.Offset;
        bb_ok = all( [...
            between(bb(ii, 3), spotsCommProps.XsizeMin, spotsCommProps.XsizeMax), ...
            between(bb(ii, 4), spotsCommProps.YsizeMin, spotsCommProps.YsizeMax) ] );
        check(ii, :)  = [uv_ok, bb_ok];

        isSpotA = ~isempty(find(gr.difspotidA == matchedID, 1));
        isSpotB = ~isempty(find(gr.difspotidB == matchedID, 1));

        if (isSpotA || isSpotB)
            % This spot is part of a pair and has been used by INDEXTER
            flag(ii) = 5;
            tempIsSpotA(ii) = isSpotA;
        elseif all(check(ii, :))
            flag(ii) = 4;
        else
            flag(ii) = 3;
        end
        spotid(ii) = matchedID;
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Re-check flag == 1, to see if there's intensity at the predicted
    % position, if requested.
    if (parameters.fsim.check_spot)
        fprintf(' (Done).\n -  Checking spots in Raw images..');

        toBeChecked = find(flag == 1);
        for ii = reshape(toBeChecked, [1 numel(toBeChecked)]);
            [found_intensity, spotInfo] = gtFwdSimCheckSpotInRawImages(...
                uvw(ii, :), spotsCommProps, parameters);

            if (found_intensity)
                % Found intensity in FULL images at predicted position
                flag(ii) = 2;

                if (parameters.fsim.assemble_figure)
                    full = gtPlaceSubImage2(spotInfo.spot, full, spotInfo.bb(1), spotInfo.bb(2), 'summed');
                end
            end
            checkSpots(ii).info = spotInfo;
        end % end for ii
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Reflections that have been matched to segmented spots
    matched = find(flag >= 3);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% We now build the diffraction stack
    fprintf(' (Done).\n - Building stack..');

    difstack = gtFwdSimBuildDifstackSpots(spotid, matched, parameters, ...
        spotsCommProps.stackHSize, spotsCommProps.stackVSize, ...
        spotsCommProps.Xsize, spotsCommProps.Ysize, ...
        bb, info);

    if (parameters.fsim.assemble_figure)
        % no padded spots
        full = sfBuildFull(spotid(reshape(matched, [1 numel(matched)])), ...
            parameters, spotsCommProps, bb(reshape(matched, [1 numel(matched)]), :));
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% We now build the diffraction blob stack
    if (isfield(parameters.seg, 'writehdf5') && parameters.seg.writehdf5)
        fprintf(' (Done).\n - Building blobs..');

        bl = gtFwdSimBuildDifstackBlobs(spotid, matched, parameters, ...
            spotsCommProps.stackHSize, spotsCommProps.stackVSize, ...
            spotsCommProps.Xsize, spotsCommProps.Ysize, ...
            bb, info);
    end

    fprintf(' (Done).\n - Building geometry..');

    %%% Diffraction geometry
    for ii = reshape(matched, [1 numel(matched)])
        % If available, we want to use the angles as determined from
        % the Friedel pairs - interrogate spotpair table
        %
        % diff_beam_dirs: direction of diffracted beam used in ASTRA
        if (flag(ii) == 5)
            [diff_beam_dirs(ii, :), W(ii)] = sfGetFriedelPair(pairsTable, spotid(ii), tempIsSpotA(ii));
        else
            % we found a segmented spot - can use the theoretically
            % predicted direction or apparent direction (based on COM of
            % grain and spot) for backprojection

            % theoretically prediceted omega angle
            if (parameters.fsim.use_th)
                currentReflection = onDet(ii);
                % use theoretically predited directions (default)
                W(ii) = gr.allblobs.omega(currentReflection);
                % theoretically predicted beam direction (in sample
                % coordinate system)
                diff_beam_dirs(ii, :) = gr.allblobs.dvecsam(currentReflection, :);
            else
                % use experimentally determined omega angles
                W(ii) = segmentedSpots.CentroidImage(spotid(ii)) * gtGetOmegaStepDeg(parameters);
                % projection direction as inferred from grain / spot center
                % positions
                diff_beam_dirs(ii, :) = gtGeoDiffVecInSample([U(ii), V(ii)], omega(ii), gr.center, labgeo, samgeo);
            end
        end
    end

    included = find(flag >= parameters.fsim.included);
    % indices of the selected spots in list of included spots
    selected = flag(included) >= parameters.fsim.selected;

    %%% We now Build ASTRA's geometry
    proj = gtFwdSimBuildProjGeometry(...
        diff_beam_dirs(included, :), gr.center, W(included), ...
        bb(included, :), parameters, ...
        spotsCommProps.stackHSize, spotsCommProps.stackVSize, ...
        selected);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Now fill in the information for the output
    %  remove reflections for which no segmented spot has been found

    out.id        = gr.id;
    out.phaseid   = phaseID;
    out.R_vector  = gr.R_vector;
    out.center    = gr.center;

    out.selected  = selected;  % vector of logicals with pre-selected spots active
    out.included  = included; % This is needed by Orientation Reconstruction
    out.ondet     = onDet;
    out.allblobs  = gr.allblobs;

    out.cands     = cands;
    out.bb        = bb;
    out.cent      = cent;
    out.check     = check;
    out.flag      = flag;
    out.pairid    = gr.pairid;
    out.spotid    = spotid;

    % only included spots info
    out.difspotID = spotid(included);
    out.uv_exp    = [U(included), V(included)];
    out.om_exp    = W(included);
    out.bb_exp    = bb(included, :);
    out.intensity = intensity(included);

    % save full image
    out.full      = full;

    % check spot info
    out.checkSpots = checkSpots;

    % diffraction geometry
    out.proj       = proj;
    % blob/spot stacks
    out.bl         = bl(included);
    out.proj.stack = difstack(:, included, :);

    % recontruction and assemble stuff
    out.seg       = [];
    out.vol       = [];
    out.segbb     = zeros(1, 6);
    out.threshold = 0;

    found_reflections = length(find(flag > 1));
    out.completeness  = found_reflections / numspots;

    if (parameters.fsim.save_grain)
        fprintf(' (Done).\n - Saving ouput..');

        grain_file_name = fullfile(parameters.acq.dir, '4_grains', 'phase_%02d', 'grain_%04d.mat');
        save(sprintf(grain_file_name, phaseID, gr.id), '-struct', 'out');
    end

    if (parameters.fsim.display_figure)
        fprintf(' (Done).\n - Displaying figure..');

        displayShowFsim(phaseID, out, rej_pars{:})
    end

    fprintf(' (Done).\n\n');

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Now update spotTables and sample
    spot2grain(out.difspotID) = {n};
    spot2phase(out.difspotID) = {phaseID};

    sample.phases{phaseID}.setR_vector(n, gr.R_vector);
    sample.phases{phaseID}.setCompleteness(n, out.completeness);
    sample.phases{phaseID}.setCenter(n, out.center); % center in pixels
    sample.phases{phaseID}.setSelectedDiffspots(n, out.selected);

    if (isempty(out.selected))
        warning('gtForwardSimulate_v2:bad_value', ...
            'No spot was selected! This will make reconstruction impossible')
    elseif (numel(out.selected) < 10)
        warning('gtForwardSimulate_v2:bad_value', ...
            'Too few selected spots! This will give a very bad reconstruction')
    end

    if (~parameters.fsim.check_spot)
        fprintf(['Completeness estimates may be too pessimistic - activate ' ...
                 'option ''check_spot'' for more reliable results\n'])
    end

    fprintf('\n Grain number %d:\n', n);

    newspots = length(find(flag == 4));
    fprintf(['Forward simulation found intensity at %d out of %d expected ' ...
             'spot positions \n'], found_reflections, numspots);
    fprintf('Completeness estimate:  %f\n', out.completeness);
    fprintf('%d spots will be used for reconstruction\n', length(out.selected));
    fprintf(['%d new spots have been detected in addition to the %d ' ...
             'spotpairs identified by INDEXTER\n'], newspots, gr.nof_pairs);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    if (par.updateDBtables)
        % update difspot table in the DB
        % keep list of difspots
        spotID{n} = out.spotid;
        spotID{n}(out.spotid == 0) = [];
        flags{n} = out.flag;
        flags{n}(out.spotid == 0) = [];

        gtDBFillUpdateColumn([parameters.acq.name 'difspot'], ...
            'grainID', 'int', 'integral', 'difspotID', {out.difspotID}, n)
        gtDBFillUpdateColumn([parameters.acq.name 'difspot'], ...
            'phaseID', 'int', 'integral', 'difspotID', {out.difspotID}, phaseID)
        gtDBFillUpdateColumn([parameters.acq.name 'difspot'], ...
            'flag', 'int', 'integral', 'difspotID', spotID(n), flags{n});
    end
end

if (par.updateDBtables)
    % fill spotTables
    spotTable.fillTable(spot2grain, 'spot2grain');
    spotTable.fillTable(spot2phase, 'spot2phase');
    % save sample.mat
    sample.mergeToFile(filetable, samplefile, 'sample');
end

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sub-functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function full = sfBuildFull(spotid, parameters, spotsCommProps, bb)
% full = sfBuildFull(spotid, parameters, spotsCommProps, bb)
%
% spotid         : spotid from grain
% parameters     : parameters.mat
% spotsCommProps : spots common properties
% bb             : bounding boxes from DB

    full = zeros(parameters.acq.ydet, parameters.acq.xdet);

    for ii = 1:numel(spotid)
        [spot, ~] = gtFwdSimBuildSpot(spotid(ii), parameters, spotsCommProps.Xsize, spotsCommProps.Ysize, bb(ii, :));

        full = gtPlaceSubImage2(spot, full, bb(ii, 1), bb(ii, 2), 'summed');
    end
end

function [diff_beam_dir, omega] = sfGetFriedelPair(pairsTable, spotID, isSpotA)
% [diff_beam_dirs, omega] = sfGetFriedelPair(pairsTable, spotID, isSpotA)

    if (isSpotA)
        idx = find(pairsTable.difAID == spotID);
        omega = pairsTable.omega(idx);
        diff_beam_dir = [pairsTable.ldirX(idx), pairsTable.ldirY(idx), pairsTable.ldirZ(idx)];
    else
        idx = find(pairsTable.difBID == spotID);
        omega = pairsTable.omegaB(idx);
        % in this case we have to reverse the line direction
        diff_beam_dir = -[pairsTable.ldirX(idx), pairsTable.ldirY(idx), pairsTable.ldirZ(idx)];
    end
end

function spotsCands = sfGetTableIndexes(segmentedSpots, candidateIDs)
% spotsCands = sfGetTableIndexes(segmentedSpots, candidateIDs)

    if (~isempty(candidateIDs))
        spotsCands = structfun(@(x)x(candidateIDs, :), segmentedSpots, 'UniformOutput', false);
    else
        spotsCands = [];
    end
end

function displayShowFsim(phaseID, grain, varargin)
% displayShowFsim(phaseID, grain, varargin)

    if isempty(varargin)
        [vars_pars, vars_list] = gtAppFsimDefaultPars();
        ind = findValueIntoCell(fieldnames(vars_pars), {'fsimtype', 'f_title'});
        vars_list(ind(:, 1), 4) = {2};

        vars_pars = gtModifyStructure(vars_pars, vars_list, 1, 'Options for displaying the Forward Sim:');
        varargin = struct2pv(vars_pars);
    end
    gtShowFsim(grain, phaseID, varargin{:});
end
