function [samXYZ, lab2sam] = gtGeoLab2Sam(labXYZ, omega, labgeo, samgeo, ...
                             freevec, rot_flag)
% GTGEOLAB2SAM  LAB -> SAMPLE reference coordinate transformation
%
% [samXYZ, lab2sam] = gtGeoLab2Sam(labXYZ, omega, labgeo, ...
%                                  samgeo, freevec, rot_flag)
%
% ----------------------------------------------------------------------
%
% Given the LAB coordinates of a set of points, transforms them 
% into SAMPLE coordinates according to an omega rotation. Omega is usually
% positive.
% It optionally creates the transformation matrix.
%
% INPUT  
%   labXYZ         - coordinates in the Lab reference in pixels (n,3)
%   omega          - omega angles in degrees (n,1)
%   labgeo.rotpos  - rotation axis position in Lab coordinates (1,3)
%   labgeo.rotdir  - rotation axis direction (right-handed rotation)
%   samgeo.orig    - origin of the sample reference in Lab coordinates (1,3)
%   samgeo.voxsize - [x,y,z] voxel sizes in lab units
%   freevec        - 0 for fixed vectors (e.g. location);
%                    1 for free vectors (e.g. orientation/direction)
%   rot_flag       - if true, only applies rotation so the norm of the 
%                    input vectors will be unchanged; 
%                    i.e. no scaling applied according to the 
%                    voxel sizes (only works if 'samgeo.lab2sam' is undefined)
%
% OUTPUT 
%   samXYZ         - coordinates in the Sample reference in no. of voxels (n,3)
%   lab2sam        - coordinate transformation matrix
%
% OPTIONAL, INPUT:
%   omega          - can be rotation tensors ('col') from omega (3,3,n) or
%                    simply (3,3,1) which would then be used for all the
%                    lab vectors.
%

n = size(labXYZ, 1);

if (~exist('rot_flag', 'var'))
    rot_flag = false;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Change of reference from Lab to Sample
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Transformation matrix from Lab to Sample coordinates:
%   Note that Lab is an orthonormal basis by definition, but Sample may not be !!
%   Therefore needs calculation of the inverse, and not only transpose.
if (~isfield(samgeo,'lab2sam') || (isfield(samgeo,'lab2sam') && isempty(samgeo.lab2sam)))
    
    if (rot_flag)
        % Keep norm of input vectors -> keep unit lengths in the rows of 
        % the transformation matrix 
        sam2lab = [samgeo.dirx; samgeo.diry; samgeo.dirz];
    else
        sam2lab = [samgeo.dirx * samgeo.voxsize(1);...
                   samgeo.diry * samgeo.voxsize(2);...
                   samgeo.dirz * samgeo.voxsize(3)];
    end

    lab2sam = inv(sam2lab);

elseif (rot_flag)
    error('Transformation matrix already defined. Set ''rot_flag'' to false.')
else
    lab2sam = samgeo.lab2sam;   
end

%%%%%%%%%%%%%%%%%%%%%
%% Rotation in Lab
%%%%%%%%%%%%%%%%%%%%%

[size_omega(1), size_omega(2), size_omega(3)] = size(omega);
% Tensor which rotates with omega from Lab back to Sample:
if (all(size_omega(1:2) == [3 3]))
    % This means that the rotation tensors were passed already!
    if (size_omega(3) == 1)
        rottensors = omega(:, :, ones(1, n));
    else
        rottensors = omega;
    end
else
    rotcomp    = gtMathsRotationMatrixComp(labgeo.rotdir', 'col');
    rottensors = gtMathsRotationTensor(omega, rotcomp);
end

if (freevec)  % free vectors

    % Multiply element-wise
    pro = reshape([labXYZ, labXYZ, labXYZ]',3,n*3).*reshape(rottensors,3,n*3);

    % Sum, reshape, transpose to get rotated vector
    labXYZ_rot = reshape(sum(pro,1),3,n)';

    % Change ref. base
    samXYZ = labXYZ_rot*lab2sam;
else % offset vectors

    % Vectors relative to rotation axis position
    labXYZrel = labXYZ - labgeo.rotpos(ones(n,1), :);

    % Apply rottensor element-wise
    labXYZrel_t = [labXYZrel, labXYZrel, labXYZrel]';
    labXYZrel_t = reshape(labXYZrel_t, 3, n*3);
    rottensors_t = reshape(rottensors, 3, n*3);
    pro = labXYZrel_t .* rottensors_t;

    % Sum, reshape, transpose, set offset back to get rotated vector
    labXYZ_rot = reshape(sum(pro,1),3,n)' + labgeo.rotpos(ones(n,1), :);
 
    % Change ref. base
    samXYZ = (labXYZ_rot - samgeo.orig(ones(n,1),:))*lab2sam;
end

end % of function
