function [g_twins] = gtTwinOrientations(R_vector, twin_angle, twin_axis, spacegroup, lp, plot_variants)
% GTTWINORIENTATIONS
%     [g_twins] = gtTwinOrientations(R_vector, twin_angle, twin_axis, spacegroup, lp, [plot_variants])
%     ------------------------------------------------------------------------------------------------
%     By passing in the parent R_vector, return the g (sample to crystal transformation) plus
%     all the twin variants
%
%     INPUT:
%       R_vector      = <double>   Rodrigues vector (1x3)
%       twin_angle    = <double>   twin rotation angle
%       twin_axis     = <double>   hkl of twin direction (Miller/Miller
%                                  Bravais indexes)
%       spacegroup    = <double>   current crystal spacegroup
%       lp            = <double>   current crystal lattice parameters
%       plot_variants = <logical>  plot figures with parent/twin crystal axes
%                                  for each variant {false}
%
%     OUTPUT:
%       g_twins       = <struct>   twin variants info
%
%     Version 002 10-01-2013 by LNervo

if ~exist('plot_variants','var') || isempty(plot_variants)
    plot_variants = false;
end

% come back to Miller notation for direction
[~,crystal_system] = gtReadSpaceGroup(spacegroup);
if strcmpi(crystal_system, 'hexagonal')
    twin_axis_hex = twin_axis;
    if size(twin_axis,2) == 4
     %   twin_axis = gtCrystFourIndexes2Miller(twin_axis, 'direction');
    end
end

% Compute the orientation matrix g = sam2cry
sam2cry = gtMathsRod2OriMat(R_vector.');

% Normalize twin_axis
%twin_axis = twin_axis/norm(twin_axis);
% Symmetry equivalents
twin_axis = gtCrystSignedHKLs(twin_axis, gtCrystGetSymmetryOperators(crystal_system, spacegroup));

% TODO: check whether it is g or gT
switch lower(crystal_system)
    case 'cubic'
        % in crystal system, parent to twin transformation
        for ii = 1:size(twin_axis, 1)
            twin_axis(ii, :) = twin_axis(ii, :)./norm(twin_axis(ii, :));

            rotcomp = gtMathsRotationMatrixComp(twin_axis(ii, :)','col');
            cry2twin = gtMathsRotationTensor(twin_angle, rotcomp);
            orimat = gtMathsAngleAxis2OriMat(twin_angle, twin_axis(ii, :)');

            g_twins(ii).axis     = twin_axis(ii, :);
            g_twins(ii).angle    = twin_angle;
            g_twins(ii).g        = sam2cry * cry2twin;
            g_twins(ii).orimat   = orimat;
            g_twins(ii).R_vector = gtMathsOriMat2Rod( g_twins(ii).g );
            g_twins(ii).eulers   = gtMathsOriMat2Euler( g_twins(ii).g );
            g_twins(ii).cry2twin = cry2twin;
            g_twins(ii).sam2cry  = sam2cry;

            %    i.e.  normally,  [h k l] = g * pl'    where g comes from gtMathsRod2OriMat
            %    here, [h k l] = g * (twin_operation * pl')
            %    so our "new" g is g*twin_operation
        end

    case 'hexagonal'
        % convert from hex to cart
        for ii = 1:size(twin_axis, 1)
            if size(twin_axis,2) == 4
                axis = gtCrystFourIndexes2Miller(twin_axis(ii, :), 'direction');
            else
                axis = twin_axis(ii, :);
            end
            twin_axis_cart(ii, :) = gtHex2Cart(axis, lp);
            twin_axis_cart(ii, :) = twin_axis_cart(ii, :)/norm(twin_axis_cart(ii, :));

            rotcomp = gtMathsRotationMatrixComp(twin_axis_cart(ii, :)','col');
            cry2twin = gtMathsRotationTensor(twin_angle, rotcomp);
            orimat = gtMathsAngleAxis2OriMat(twin_angle, twin_axis_cart(ii, :)');

            g_twins(ii).axis_hex  = twin_axis(ii, :);
            g_twins(ii).axis      = twin_axis_cart(ii, :);
            g_twins(ii).angle     = twin_angle;
            g_twins(ii).g         = sam2cry * cry2twin;
            g_twins(ii).orimat    = orimat;
            g_twins(ii).R_vector  = gtMathsOriMat2Rod( g_twins(ii).g );
            g_twins(ii).eulers    = gtMathsOriMat2Euler( g_twins(ii).g );
            g_twins(ii).cry2twin  = cry2twin;
            g_twins(ii).sam2cry   = sam2cry;
        end
end

xc = gtVectorCryst2Lab([1 0 0], sam2cry); % sam2cry = g
yc = gtVectorCryst2Lab([0 1 0], sam2cry); % sam2cry = g
zc = gtVectorCryst2Lab([0 0 1], sam2cry); % sam2cry = g
gs = reshape([g_twins.g],3,3,[]);

xt = gtVectorCryst2Lab([1 0 0], gs);
yt = gtVectorCryst2Lab([0 1 0], gs);
zt = gtVectorCryst2Lab([0 0 1], gs);

for ii = 1:length(g_twins)
    x = xt(ii, :);
    y = yt(ii, :);
    z = zt(ii, :);
    g_twins(ii).xyz = [x; y; z];
    g_twins(ii).abc = [xc; yc; zc];
end

if (plot_variants)
    hf = figure();
    ha = [];
    for ii = 1:length(g_twins)
        ha(ii) = subplot(3, 4, ii);
        hold(ha(ii),'on');

        plot3([0 xc(1)], [0 xc(2)], [0 xc(3)], 'r:');
        plot3([0 yc(1)], [0 yc(2)], [0 yc(3)], 'g:');
        plot3([0 zc(1)], [0 zc(2)], [0 zc(3)], 'b:');

        x = xt(ii, :);
        y = yt(ii, :);
        z = zt(ii, :);
        plot3([0 x(1)], [0 x(2)], [0 x(3)], 'r');
        plot3([0 y(1)], [0 y(2)], [0 y(3)], 'g');
        plot3([0 z(1)], [0 z(2)], [0 z(3)], 'b');

        set(ha(ii), 'XLim', [-1 1]);
        set(ha(ii), 'YLim', [-1 1]);
        set(ha(ii), 'ZLim', [-1 1]);
        set(get(ha(ii),'Title'), 'String', num2str(ii), 'FontWeight', 'bold');
        set(ha(ii), 'UserData', g_twins(ii));
    end

    h = rotate3d(hf);
    set(h,'Enable','on','RotateStyle','Box');

    h_link = linkprop(ha, 'View');
    setappdata(hf, 'linkprop', h_link);
    setappdata(hf, 'g_twins', g_twins);
end

end % end of function
