#!/usr/bin/env python
'''
vtkShowCrackMeshPropagationInteractive.py
Program to show interactively the propagation of a crack through vtk GUI.
The VTK input mesh file should contains the cycle number field as cell data.

Usage: vtkShowCrackMeshPropagationInteractive.py  <meshFile>

Last update by Yoann Guilhem 2013-07-31
'''

import vtk
import sys

# Show vtk version
print "Using vtk version", vtk.vtkVersion.GetVTKVersion()

# Parameters
initAngle = -45.

# Set mesh input filename
try :
    meshFile = sys.argv[1]
except :
    print 'Missing input mesh file'
    print 'Usage: vtkShowCrackMeshPropagationInteractive.py  <meshFile>'

############### FUNCTIONS ################

# When interaction starts, the requested frame rate is increased.
def StartInteraction(obj, event):
    global renderWindow
    renderWindow.SetDesiredUpdateRate(10)

# When interaction ends, the requested frame rate is decreased to
# normal levels. This causes a full resolution render to occur.
def EndInteraction(obj, event):
    global renderWindow
    renderWindow.SetDesiredUpdateRate(0.001)

# When moving slider, changing threshold
def sliderCallback(obj, event):
    global meshThreshold
    widgetValue = obj.GetRepresentation().GetValue()
    meshThreshold.ThresholdByLower(widgetValue)
    meshThreshold.Update()

############### MESH ######################

# Mesh Reader
meshReader = vtk.vtkDataSetReader()
meshReader.SetFileName(meshFile)
meshReader.ReadAllScalarsOn() 
meshReader.Update()

# Polydata
if meshReader.IsFilePolyData():
    polydata = meshReader.GetOutput()
else:
    raise RuntimeError('Input mesh file does not contain a mesh (vtkPolyData)!')
#polydata.ReadAllScalarsOn()

# Bounds
meshBounds = polydata.GetBounds()
meshCenter = polydata.GetCenter()
meshLength = polydata.GetLength()
print "mesh bounds:", meshBounds
print "mesh center:", meshCenter
print "mesh length:", meshLength

# Range
meshRange = polydata.GetScalarRange()
print "mesh range:", meshRange

# Mesh threshold
meshThreshold = vtk.vtkThreshold()
meshThreshold.SetInputConnection(meshReader.GetOutputPort())
if vtk.vtkVersion.GetVTKMajorVersion() <= 5:
    meshThreshold.SetInput(polydata)
else:
    meshThreshold.SetInputData(polydata)
meshThreshold.ThresholdByLower(meshRange[1])
meshThreshold.Update()
viewMesh = meshThreshold.GetOutput()

# Mesh mapper
meshMapper = vtk.vtkDataSetMapper()
if vtk.vtkVersion.GetVTKMajorVersion() <= 5:
    meshMapper.SetInput(viewMesh)
else:
    meshMapper.SetInputData(viewMesh)
meshMapper.SetScalarRange(meshRange)
meshMapper.SetColorModeToMapScalars()

# Mesh actor
meshActor = vtk.vtkActor()
meshActor.SetMapper(meshMapper)

############# DISPLAY ######################

# Renderer
renderer = vtk.vtkRenderer()

# Render window
renderWindow = vtk.vtkRenderWindow()
renderWindow.AddRenderer(renderer)

# Window interactor
renderWindowInteractor = vtk.vtkRenderWindowInteractor()
renderWindowInteractor.SetRenderWindow(renderWindow)
renderer.AddActor(meshActor)

# Trackball
style = vtk.vtkInteractorStyleTrackballCamera()
renderWindowInteractor.SetInteractorStyle(style)

# First rendering
renderWindowInteractor.Initialize()
renderWindow.Render()

# Slider representation
sliderRep = vtk.vtkSliderRepresentation2D()
sliderRep.GetPoint1Coordinate().SetCoordinateSystemToNormalizedDisplay()
sliderRep.GetPoint2Coordinate().SetCoordinateSystemToNormalizedDisplay()
sliderRep.GetPoint1Coordinate().SetValue(0.1, 0.1)
sliderRep.GetPoint2Coordinate().SetValue(0.9, 0.1)
sliderRep.SetSliderLength(0.025)
sliderRep.SetSliderWidth(0.025)
sliderRep.SetEndCapLength(0.005)
sliderRep.SetEndCapWidth(0.05)
sliderRep.SetTubeWidth(0.005)
sliderRep.SetTitleText("Cycle number")
sliderRep.GetSliderProperty().SetColor(0, 1, 0)    # knob in gree
sliderRep.GetSelectedProperty().SetColor(1, 0 ,0) # slected knob in red
sliderRep.SetMinimumValue(meshRange[0])
sliderRep.SetMaximumValue(meshRange[1])
sliderRep.SetValue(meshRange[1])
sliderRep.SetLabelFormat("%6.0f")

# Slider
#sliderWidget = vtk.vtkCenteredSliderWidget()
sliderWidget = vtk.vtkSliderWidget()
sliderWidget.SetInteractor(renderWindowInteractor)
sliderWidget.SetRepresentation(sliderRep)
#sliderWidget.SetNumberOfAnimationSteps(5)
#sliderWidget.SetAnimationModeToAnimate()
sliderWidget.SetAnimationModeToJump()
sliderWidget.EnabledOn()
sliderWidget.AddObserver("EndInteractionEvent", StartInteraction)
sliderWidget.AddObserver("InteractionEvent", sliderCallback)
sliderWidget.AddObserver("EndInteractionEvent", EndInteraction)

# Camera
camera = renderer.GetActiveCamera()
camera.Elevation(initAngle)
camera.SetViewUp(0, 0, 1)
renderWindow.Render()

# Recorder
recorder = vtk.vtkInteractorEventRecorder()
recorder.SetInteractor(renderWindowInteractor)
recorder.SetFileName("vtk.log")

renderWindowInteractor.Start()
